local framework = battleplus.framework

-- functions for the damage hooks
framework.Damage_PlayerSpin = function(p, inflictor, source, gamemode)
	if (gamemode.playerspin) then return gamemode.playerspin(p, inflictor, source) end
end
framework.Damage_PlayerSquish = function(p, inflictor, source, gamemode)
	if (gamemode.playersquish) then return gamemode.playersquish(p, inflictor, source) end
end
framework.Damage_PlayerExplode = function(p, inflictor, source, gamemode)
	if (gamemode.playerexplode) then return gamemode.playerexplode(p, inflictor, source) end
end
framework.Damage_MobjDamage = function(mo, inflictor, source, gamemode)
	if (gamemode.mobjdamage) then return gamemode.mobjdamage(mo, inflictor, source) end
end

-- easy friendlyfire function
framework.ShouldDamage_FriendlyFire = function(p, inflictor, source)
	if not (server.bp_server.teamgamemode) then return nil end
	if CV_FindVar("friendlyfire").value then return nil end
	if (inflictor and inflictor.type == MT_SINK) then return nil end	-- :ahaha:
	
	local p2
    if (inflictor and inflictor.valid) and inflictor.player
        p2 = inflictor.player
    end
    if (source and source.valid) and source.player
        p2 = source.player
    end
    if not (p2) return nil end
	
	if (p.bp.team == p2.bp.team)	-- same team?
	and (p != p2)	-- we can damage ourselves
		K_DoInstashield(p)
        return false
    end
end

-- functions for the shoulddamage hooks
framework.Damage_ShouldSpin = function(p, inflictor, source, gamemode)
	-- friendlyfire first
	local friendlyfire = framework.ShouldDamage_FriendlyFire(p, inflictor, source)
	if (friendlyfire != nil) then return friendlyfire end
	
	-- now the gamemode function
	if (gamemode.shouldspin) then return gamemode.shouldspin(p, inflictor, source) end
end
framework.Damage_ShouldSquish = function(p, inflictor, source, gamemode)
	-- friendlyfire first
	local friendlyfire = framework.ShouldDamage_FriendlyFire(p, inflictor, source)
	if (friendlyfire != nil) then return friendlyfire end
	
	-- now the gamemode function
	if (gamemode.shouldsquish) then return gamemode.shouldsquish(p, inflictor, source) end
end
framework.Damage_ShouldExplode = function(p, inflictor, source, gamemode)
	-- friendlyfire first
	local friendlyfire = framework.ShouldDamage_FriendlyFire(p, inflictor, source)
	if (friendlyfire != nil) then return friendlyfire end
	
	-- now the gamemode function
	if (gamemode.shouldexplode) then return gamemode.shouldexplode(p, inflictor, source) end
end
framework.Damage_ShouldDamage = function(mo, inflictor, source, gamemode)
	-- friendlyfire first
	local friendlyfire = framework.ShouldDamage_FriendlyFire(mo.player, inflictor, source)
	if (friendlyfire != nil) then return friendlyfire end
	
	-- now the gamemode function
	if (gamemode.shoulddamage) then return gamemode.shoulddamage(mo, inflictor, source) end
end

-- friendlyfire collision hook
framework.Damage_FriendlyFireCollision = function(mo, mo2)
	-- ...unless we want to collide!
	if not G_BattleGametype() then return end
	if (leveltime < 106) then return end
	if not (server.bp_server.teamgamemode) then return end
	if CV_FindVar("friendlyfire").value then return end

	if mo and mo.valid and mo.type == MT_PLAYER and mo2 and mo2.valid
		local accept_types = {	
			MT_ORBINAUT, MT_ORBINAUT_SHIELD,
			MT_JAWZ, MT_JAWZ_DUD, MT_JAWZ_SHIELD,
			MT_BANANA, MT_BANANA_SHIELD,
			MT_EGGMANITEM, MT_EGGMANITEM_SHIELD,
			MT_SPBEXPLOSION, MT_MINEEXPLOSION,
			MT_BALLHOG
		}

		-- check if mo2 is 1 of these types, return if not
		local accepted
		for i = 1, #accept_types
			if mo2.type == accept_types[i]
				accepted = true
				break
			end
		end
		if not (accepted) return end

		if ((mo2.target and mo2.target.valid) and mo2.target.player)	-- they exist
		and (mo2.target.player.bp.team == mo.player.bp.team)	-- same team
		and (mo.z + mo.height >= mo2.z and mo2.z <= mo2.z + mo2.height)	-- z check
		and (mo.player ~= mo2.target.player) -- we can still damage ourself, though
			return false	-- this collision never happened.
		end
	end
end
addHook("MobjCollide", framework.Damage_FriendlyFireCollision, MT_PLAYER)
addHook("MobjMoveCollide", framework.Damage_FriendlyFireCollision, MT_PLAYER)