-- put all the framework functions to work here

-- unload
addHook("MapLoad", battleplus.framework.Unload)
addHook("VoteThinker", battleplus.framework.Unload_TeamSkincolors)

-- main thinker
addHook("ThinkFrame", do
	if not G_BattleGametype() then return end
	
	if (leveltime < 105)
		-- have a roulette for the selection
		if not ((leveltime+1)%3) then S_StartSound(nil, sfx_itrol1+((leveltime+1)%24/3)) end
	elseif (leveltime == 105)	-- select the gamemode
		battleplus.framework.Setup()
	else -- leveltime > 105
		battleplus.framework.LateAdd()	-- why would you add battle plus late
		battleplus.framework.MidJoin()	-- mid joining
		
		local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
		-- CRASH (actually a fallback but it's pretty funny so lmao)
		if not (gamemode)
			if (leveltime == 106) then S_StartSound(nil, sfx_bpcrsh) end
			if (leveltime == 156) then G_ExitLevel(gamemap, true) end
		end
		
		battleplus.framework.Main(gamemode)
	end
end)

-- hud
hud.add(function(v, p, c)
	if not G_BattleGametype() then return end
	
	if (leveltime < 106)
		battleplus.framework.HUD_Selection(v, p)
	else
		if not (server.bp_server) then return end
		
		local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
		battleplus.framework.HUD_Main(v, p, c, gamemode)
	end
end)

-- damage hooks
addHook("PlayerSpin", function(p, inflictor, source)
	if not G_BattleGametype() then return end
	if (leveltime < 106) then return end
	
	local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
	return battleplus.framework.Damage_PlayerSpin(p, inflictor, source, gamemode)
end)
addHook("PlayerSquish", function(p, inflictor, source)
	if not G_BattleGametype() then return end
	if (leveltime < 106) then return end
	
	local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
	return battleplus.framework.Damage_PlayerSquish(p, inflictor, source, gamemode)
end)
addHook("PlayerExplode", function(p, inflictor, source)
	if not G_BattleGametype() then return end
	if (leveltime < 106) then return end
	
	local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
	return battleplus.framework.Damage_PlayerExplode(p, inflictor, source, gamemode)
end)
addHook("MobjDamage", function(mo, inflictor, source)
	if not G_BattleGametype() then return end
	if (leveltime < 106) then return end
	
	local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
	return battleplus.framework.Damage_MobjDamage(mo, inflictor, source, gamemode)
end, MT_PLAYER)

-- shoulddamage hooks
addHook("ShouldSpin", function(p, inflictor, source)
	if not G_BattleGametype() then return end
	if (leveltime < 106) then return end
	
	local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
	return battleplus.framework.Damage_ShouldSpin(p, inflictor, source, gamemode)
end)
addHook("ShouldSquish", function(p, inflictor, source)
	if not G_BattleGametype() then return end
	if (leveltime < 106) then return end
	
	local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
	return battleplus.framework.Damage_ShouldSquish(p, inflictor, source, gamemode)
end)
addHook("ShouldExplode", function(p, inflictor, source)
	if not G_BattleGametype() then return end
	if (leveltime < 106) then return end
	
	local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
	return battleplus.framework.Damage_ShouldExplode(p, inflictor, source, gamemode)
end)
addHook("ShouldDamage", function(mo, inflictor, source)
	if not G_BattleGametype() then return end
	if (leveltime < 106) then return end
	
	local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
	return battleplus.framework.Damage_ShouldDamage(mo, inflictor, source, gamemode)
end, MT_PLAYER)
addHook("MobjCollide", battleplus.framework.Damage_FriendlyFireCollision, MT_PLAYER)
addHook("MobjMoveCollide", battleplus.framework.Damage_FriendlyFireCollision, MT_PLAYER)

-- handle bumpers
addHook("MobjThinker", function(mo)
	if not G_BattleGametype() then return end
	
	if (mo and mo.valid)
		if (leveltime < 106)
			-- for SOME reason bumpers use flags, not flags2???
			mo.flags = $|MF_NOSECTOR	-- yeah I'm completely baffled but whatever
		else
			if not (server.bp_server) then return end	-- hasnt had fallback yet?
			local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
			if not (gamemode.rules & BP_GMR_BUMPERS) then P_KillMobj(mo) return end	-- no need for bumpers
			
			-- fancy pop in animation for when we need them
			if (leveltime == 130)
				mo.flags = $ & ~(MF_NOSECTOR)
				local pop = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
				pop.state = S_XPLD1
				pop.scale = mo.scale/2
				if not S_SoundPlaying(mo.target, sfx_shield) -- make sure it only plays once, no earrape
					S_StartSound(mo.target, sfx_shield)
				end
			end
		end
	end
end, MT_BATTLEBUMPER)

-- Removing Battle Points in gamemodes that don't need them (removes confusion)
addHook("MobjThinker", function(mo)
	if not G_BattleGametype() then return end
	if (leveltime < 106) then return end
	
	if (mo and mo.valid)
		if not (server.bp_server) then return end	-- hasnt had fallback yet?
		local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
		if not (gamemode.rules & BP_GMR_REMOVEPOINTS) then return end	-- we dont want to remove them
		if (mo.bp_override) then return end	-- dont remove these
		
		P_KillMobj(mo)
	end
end, MT_BATTLEPOINT)