local framework = battleplus.framework

framework.HUD_Main_TimeLimit = function(v, p, gamemode)
	hud.enable("time")
	-- dont draw conditions
	if not (gamemode.rules & BP_GMR_TIMELIMIT) then return end	-- we dont need this, just keep it normal
	if not (server.bp_server.timelimit) then return end	-- table somehow wasnt setup yet?
	if (battleplus.IsDisplayPlayer(p) ~= 0) then return end	-- reduce rendering as much as possible
	if (p.exiting) then return end
	if not (splitscreen) and (p.kartstuff[k_comebacktimer] and not p.kartstuff[k_bumper]) then return end
	
	-- set drawing vars
	local hx, hy, flags = 172, (splitscreen) and 3 or 9, V_SNAPTOTOP|V_SNAPTORIGHT|V_HUDTRANS
	if (splitscreen) > 1
		hx, hy, flags = 90, 3, V_SNAPTOTOP|V_HUDTRANS
	end
	
	-- set timelimit
	local timelimit = gamemode.timelimit.value or 3	-- default to 3 minutes if .timelimit isnt set
	if (gamemode.timelimit and not gamemode.timelimit.value) then timelimit = 0 end
	if not (timelimit) then return end	-- we dont need custom rendering
	hud.disable("time")
	
	-- get vars 
	local time, freeze, overtime = server.bp_server.timelimit.time, server.bp_server.timelimit.roundfreeze, server.bp_server.timelimit.overtime
	-- redfont conditions
	local freeze_cond = (freeze and (freeze > 1*TICRATE and not (freeze%10 >= 5) or freeze <= 1*TICRATE))
	local time_cond = (time <= 10*TICRATE and not (time%10 >= 5) and not overtime)
	local overtime_cond = (overtime and not (overtime%10 >= 5) and not server.bp_server.preventovertime)
	local redfont = (freeze_cond or time_cond or overtime_cond)
	-- drawing conditions
	local freeze_dontdraw = (freeze and freeze > 1*TICRATE and (freeze%10 >= 5))
	local time_dontdraw = (time <= 10*TICRATE and (time%10 >= 5))
	local overtime_dontdraw = (overtime and (overtime%10 >= 5) and not server.bp_server.preventovertime)
	local dontdraw = (freeze_dontdraw or time_dontdraw or overtime_dontdraw)
					
	v.draw(hx, hy, v.cachePatch((redfont) and "K_STTIMR" or "K_STTIMW"), flags)
	if not (dontdraw)
		local m = G_TicsToMinutes(time, true)
		local s = G_TicsToSeconds(time)
		local c = G_TicsToCentiseconds(time)
		
		local time_string = (overtime and not server.bp_server.preventovertime) and "OVERTIME" or (m < 10 and "0"..m or m).."'"..(s < 10 and "0"..s or s).."\""..(c < 10 and "0"..c or c)
		battleplus.DrawKartText(v, hx+129, hy+3, time_string, flags, "right", (redfont))
	end
end

framework.HUD_Main_TeamScores = function(v, p, gamemode)
	-- dont draw conditions
	if (gamemode.rules & BP_GMR_HIDETEAMSCORES) then return end	-- we dont need this.
	if not (server.bp_server and server.bp_server.teamgamemode) then return end
	if (battleplus.IsDisplayPlayer(p) ~= 0) then return end	-- reduce rendering as much as possible
	if (p.exiting) then return end
	if not (splitscreen) and (p.kartstuff[k_comebacktimer] and not p.kartstuff[k_bumper]) then return end
	
	-- set drawing vars
	local hx, hy, flags = 237, (splitscreen) and 3 or 9, V_SNAPTOTOP|V_SNAPTORIGHT|V_HUDTRANS
	if (splitscreen > 1)
		hx, hy, flags = 154, 3, V_SNAPTOTOP|V_HUDTRANS
	end
	
	-- offset hy depending on if timelimit is drawn
	if (not (splitscreen) and hud.enabled("time"))	-- timelimit is drawn (unless in splitscreen, timelimit isnt drawn there)
	or ((gamemode.rules & BP_GMR_TIMELIMIT) and server.bp_server.timelimit)
		hy = $+21
	end
	
	-- get team score numbers!
	local sums, truesums = {0, 0}, {0, 0}
	local nums = {{}, {}}
	for team = 1, 2
		-- sum player scores
		for k, v in ipairs(server.bp_server.teams[team])
			if not (v and v.valid) then continue end
			sums[team], truesums[team] = $+v.marescore, $+v.marescore
			-- additional points that wont be displayed, but go into the winning team calculations
			truesums[team] = $+(v.bp.hud_hiddenpoints or 0)
		end
		local s = tostring(sums[team])	-- convert to string for next use
		
		-- separate sums into individual numbers
		local t = nums[team]
		for i = 1, s:len()
			t[#t+1] = tonumber(s:sub(i, i))
		end
	end
	
	-- get colors
	local colors = {battleplus.cv_t1c.value, battleplus.cv_t2c.value}
	-- last minute colours
	if (gamemode.rules & BP_GMR_TIMELIMIT)
	and not (gamemode.rules & BP_GMR_OVERTIME)
		local timelimit = gamemode.timelimit.value or 3
		if (gamemode.timelimit and not gamemode.timelimit.value) then timelimit = 0 end
		
		if (server.bp_server.timelimit) -- timelimit is set up
		and (timelimit > 0)	-- timelimit var is > 0
		and (server.bp_server.timelimit.time <= 60*TICRATE)	-- last minute?
			-- get winning team
			local winning_team = 0
			if (truesums[1] > truesums[2])
				winning_team = 1
			elseif (truesums[2] > truesums[1])
				winning_team = 2
			end
			
			for team = 1, 2
				if (winning_team ~= team) then continue end -- not the winning team
				
				local colormap_sets = {
					{SKINCOLOR_GREY, SKINCOLOR_SILVER, SKINCOLOR_SILVER, SKINCOLOR_WHITE, SKINCOLOR_SILVER, SKINCOLOR_SILVER}, -- 1: No colour mapping (light white/black shades are so limited....)
					{SKINCOLOR_BLACK, SKINCOLOR_NICKEL, SKINCOLOR_SILVER, SKINCOLOR_WHITE, SKINCOLOR_SILVER, SKINCOLOR_NICKEL}, -- 2: V_GREYMAP
					{SKINCOLOR_LEATHER, SKINCOLOR_BROWN, SKINCOLOR_BEIGE, SKINCOLOR_WHITE, SKINCOLOR_BEIGE, SKINCOLOR_BROWN}, -- 3: V_BROWNMAP
					{SKINCOLOR_ROSE, SKINCOLOR_PINK, SKINCOLOR_LILAC, SKINCOLOR_WHITE, SKINCOLOR_LILAC, SKINCOLOR_PINK}, -- 4: V_PINKMAP
					{SKINCOLOR_RED, SKINCOLOR_RASPBERRY, SKINCOLOR_SALMON, SKINCOLOR_WHITE, SKINCOLOR_SALMON, SKINCOLOR_RASPBERRY}, -- 5: V_REDMAP
					{SKINCOLOR_PUMPKIN, SKINCOLOR_ORANGE, SKINCOLOR_TANGERINE, SKINCOLOR_WHITE, SKINCOLOR_TANGERINE, SKINCOLOR_ORANGE}, -- 6: V_ORANGEMAP
					{SKINCOLOR_BRICK, SKINCOLOR_PINK, SKINCOLOR_CARAMEL, SKINCOLOR_WHITE, SKINCOLOR_CARAMEL, SKINCOLOR_PINK}, -- 7: V_PEACHMAP
					{SKINCOLOR_BRONZE, SKINCOLOR_ROYAL, SKINCOLOR_GOLD, SKINCOLOR_WHITE, SKINCOLOR_GOLD, SKINCOLOR_ROYAL}, -- 8: V_GOLDMAP
					{SKINCOLOR_OLIVE, SKINCOLOR_MUSTARD, SKINCOLOR_YELLOW, SKINCOLOR_WHITE, SKINCOLOR_YELLOW, SKINCOLOR_MUSTARD}, -- 9: V_YELLOWMAP
					{SKINCOLOR_MOSS, SKINCOLOR_MINT, SKINCOLOR_PISTACHIO, SKINCOLOR_WHITE, SKINCOLOR_PISTACHIO, SKINCOLOR_MINT}, -- 10: V_TEAMAP
					{SKINCOLOR_PINETREE, SKINCOLOR_GREEN, SKINCOLOR_MINT, SKINCOLOR_WHITE, SKINCOLOR_MINT, SKINCOLOR_GREEN}, -- 11: V_GREENMAP
					{SKINCOLOR_NAVY, SKINCOLOR_CERULEAN, SKINCOLOR_JAWZ, SKINCOLOR_WHITE, SKINCOLOR_JAWZ, SKINCOLOR_CERULEAN}, -- 12: V_SKYMAP
					{SKINCOLOR_PLATINUM, SKINCOLOR_STEEL, SKINCOLOR_SLATE, SKINCOLOR_WHITE, SKINCOLOR_SLATE, SKINCOLOR_STEEL}, -- 13: V_STEELMAP
					{SKINCOLOR_BLUE, SKINCOLOR_PERIWINKLE, SKINCOLOR_JAWZ, SKINCOLOR_WHITE, SKINCOLOR_JAWZ, SKINCOLOR_PERIWINKLE}, -- 14: V_BLUEMAP
					{SKINCOLOR_PURPLE, SKINCOLOR_LAVENDER, SKINCOLOR_BUBBLEGUM, SKINCOLOR_WHITE, SKINCOLOR_BUBBLEGUM, SKINCOLOR_LAVENDER}, -- 15: V_PURPLEMAP
					{SKINCOLOR_LAVENDER, SKINCOLOR_PASTEL, SKINCOLOR_TOXIC, SKINCOLOR_WHITE, SKINCOLOR_TOXIC, SKINCOLOR_PASTEL} -- 16: V_LAVENDERMAP
				}
				
				local set = colormap_sets[battleplus.ColorToMap(colors[team], false, true)]
				colors[team] = set[((leveltime%12)/2)+1]
			end
		end
	end
	
	-- draw team scores
	for team = 1, 2
		local hx2 = {hx-13 - (12*(#nums[team]-1)), hx+13}
		
		-- divider
		v.draw(hx, hy, v.cachePatch("TMFNTDT"..team), flags, v.getColormap(-1, colors[team]))
		-- numbers
		for i = 1, #nums[team]
			local n = nums[team][i]
			v.draw(hx2[team], hy, v.cachePatch("TMFNT"..n), flags, v.getColormap(-1, colors[team]))
			hx2[team] = $+12
		end
	end
end

framework.HUD_Main_Bumpers = function(v, p, gamemode)
	hud.disable("gametypeinfo") 
	if not (gamemode.rules & BP_GMR_BUMPERS) then hud.disable("battlerankingsbumpers") return end	-- no need for rendering
	hud.enable("battlerankingsbumpers")
	
	-- conditions when we wouldnt draw it normally
	if (p.exiting or p.spectator) then return end
	if (p.kartstuff[k_comebacktimer] and not p.kartstuff[k_bumper]) then return end
	-- additional condition for perma elimination
	if ((gamemode.rules & BP_GMR_ELIMINATION) and not p.kartstuff[k_bumper]) then return end
	
	local dp = battleplus.IsDisplayPlayer(p)
	local critical = ((leveltime%10 >= 5) and p.kartstuff[k_bumper] == 1)	-- critical!
	local max_bumpers = gamemode.bumpers.value or 3
	
	local patch = (p.kartstuff[k_bumper] <= 0) and "K_STKARM" or "K_STBALN"
	local string = (p.kartstuff[k_bumper] <= 0) and p.kartstuff[k_comebackpoints].."/2" or p.kartstuff[k_bumper].."/"..max_bumpers
	
	if (splitscreen) > 1
		local hx = (dp == 1 or dp == 3) and 120 or 3
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_SNAPTORIGHT|V_HUDTRANS, dp)
		patch = (p.kartstuff[k_bumper] <= 0) and "K_SPTKRM" or "K_BLNICO"
		
		v.draw(hx, 86, v.cachePatch(patch), flags, v.getColormap(-1, p.skincolor))
		v.drawString(hx+37, 88, string, flags|(critical and V_REDMAP or 0), "right")
	else
		local hy = (splitscreen) and 76 or 171
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_HUDTRANS, dp)
		
		v.draw(9, hy, v.cachePatch(patch), flags, v.getColormap(-1, p.skincolor))
		battleplus.DrawKartText(v, 92, hy+3, string, flags, "right", critical)
	end
end

framework.HUD_Main_DisabledHUDs = function(gamemode)
	if not (gamemode.disabledhuds) then return end
	
	for k, v in ipairs(gamemode.disabledhuds)
		hud.disable(v)
	end
end

framework.HUD_Main_GamemodeFadeout = function(v, p, gamemode)
	if (battleplus.IsDisplayPlayer(p) ~= 0) then return end	-- reduce rendering as much as possible
	
	-- fadeout starts at 142...
	local startfade = v.localTransFlag(v)>>V_ALPHASHIFT
	local fade = leveltime >= 142 and (min(9, startfade + (leveltime-142))) or startfade
	
	if (fade < 9)
		v.draw(135, 11, v.cachePatch("K_ITBG"), fade<<V_ALPHASHIFT)
		local flash = ((leveltime-106 < TICRATE) and (leveltime%2))	-- flash icon
		
		local patch = gamemode.icon
		local skin = (flash) and -6 or -1
		local colour = (flash) and (server.bp_server.gamemodehell and SKINCOLOR_CRIMSON or SKINCOLOR_WHITE) or SKINCOLOR_NONE
		local string, string2 = gamemode.name, gamemode.objective
		
		if (server.bp_server.teamgamemode)	-- we're in teams
			-- set colours
			local c = {battleplus.cv_t1c.value, battleplus.cv_t2c.value}
			-- set colour
			if (flash) 
				if not (server.bp_server.gamemodehell)
					colour = c[(splitscreen) and 1 or p.bp.team]	-- first get from table
					c[2] = c[(splitscreen) and 2 or p.bp.team]	-- now set table since we dont need it no more
				end
			else
				colour = c[1]
			end
			
			if (gamemode.teamicons)	-- need to be valid
				-- set team icon
				patch = gamemode.teamicons[1]
				v.draw(135, 11, v.cachePatch(gamemode.teamicons[2]), fade<<V_ALPHASHIFT, v.getColormap(skin, c[2]))
			end
			string = gamemode.teamname or gamemode.name
			string2 = gamemode.teamobjective or gamemode.objective
			
			-- the gamemode might have objectives for EACH team?
			if (gamemode.teamobjectives)
				-- {team 1 objective, team 2 objective}
				string2 = {gamemode.teamobjectives[1], gamemode.teamobjectives[2]}
			end
		end
		patch = $ or "K_ITFREE"	-- failsafe (custom gamemodes?)
		
		v.draw(135, 11, v.cachePatch(patch), fade<<V_ALPHASHIFT, v.getColormap(skin, colour))
		if (string) then v.drawString(160, 58, "*"..string.."*", fade<<V_ALPHASHIFT, "center") end
		-- draw gamemode objective
		if not (string2) then return end	-- no objective? (custom gamemodes?)
		
		-- this'll make this easier
		local function drawObjective(stringFade)
			if (string2[2])	-- we got 2 objectives for each team
				if not (p.bp.team) then return end	-- can't display anything if we have no team!
				v.drawString(160-(v.stringWidth("*"..string2[p.bp.team].."*", 0, "thin")/2), 67, "*"..string2[p.bp.team].."*", (stringFade)<<V_ALPHASHIFT, "thin")
			else
				v.drawString(160-(v.stringWidth("*"..string2.."*", 0, "thin")/2), 67, "*"..string2.."*", (stringFade)<<V_ALPHASHIFT, "thin")
			end
		end
		
		if (leveltime < 142)	-- use our own fade here
			-- fade in starts at 107
			local maxfade = v.localTransFlag(v)>>V_ALPHASHIFT
			local fade2 = max(maxfade, (9 - ((leveltime-107))))
								
			if (fade2 <= 9)
				drawObjective(fade2)
			end
		else
			drawObjective(fade)
		end
	end
end

framework.HUD_Main = function(v, p, c, gamemode)
	framework.HUD_Main_TimeLimit(v, p, gamemode)
	framework.HUD_Main_TeamScores(v, p, gamemode)
	framework.HUD_Main_Bumpers(v, p, gamemode)
	framework.HUD_Main_DisabledHUDs(gamemode)
	if (gamemode.hudfunc) then gamemode.hudfunc(v, p, c) end	-- run hud func for gamemode
	framework.HUD_Main_GamemodeFadeout(v, p, gamemode)	-- fadeout for gamemode after selection
end