local framework = battleplus.framework

-- copypaste from Setup.lua but with N_RandomRange and only updates odds every so often
local icons_funny, icons_teams, icons_hell = false, false, false

framework.HUD_GetFunnyOdds = function()
	if ((leveltime+1)%3) or (paused) then return icons_funny end
	
	icons_funny = (battleplus.N_RandomRange(1, 128) == 69)
	return icons_funny
end
framework.HUD_GetTeamOdds = function()
	if ((leveltime+1)%3) or (paused) then return icons_teams end
	
	if not (battleplus.cv_teams.value)	-- disabled
		icons_teams = false
	else
		local odds = (battleplus.GetPlayerCount()%2) and 8 or 3	-- less chance for teams with uneven player counts
		
		if (battleplus.GetPlayerCount() >= 4)
			if (battleplus.cv_teams.value == 2)	-- forced
			or (battleplus.N_RandomRange(1, odds) == odds)
				icons_teams = true	-- yeah... this is bad!
			else
				icons_teams = false
			end
		else
			icons_teams = false
		end
	end
	
	return icons_teams
end
framework.HUD_GetHellOdds = function()
	if ((leveltime+1)%3) or (paused) then return icons_hell end
	
	if not (battleplus.cv_hell.value)	-- disabled
		icons_hell = false
	else
		local odds = (mapheaderinfo[gamemap].menuflags & LF2_HIDEINMENU) and 2 or 10	-- hell maps have an increased chance
		
		if (battleplus.cv_hell.value == 2)	-- forced
		or (battleplus.N_RandomRange(1, odds) == odds)
			icons_hell = true	-- yeah... this is bad!
		else
			icons_hell = false
		end
	end
	
	return icons_hell
end

framework.HUD_Selection_GetGamemodes = function()
	local funny, teams, hell = framework.HUD_GetFunnyOdds(), framework.HUD_GetTeamOdds(), framework.HUD_GetHellOdds()
	-- go through gamemodes
	local possibleGamemodes = {}
	for k, gamemode in ipairs(battleplus.gamemodes)
		-- funny icon
		if (k == 1 and (funny))
			table.insert(possibleGamemodes, -1) -- haha funny
			break
		end
		-- check fields
		local rules = gamemode.rules or 0	-- fallback
		local hellgamemode = (rules & BP_GMR_HELL) and true or false
		
		if not (teams) and (rules & BP_GMR_FORCETEAMS) then continue end	-- this gamemode NEEDS teams
		if (teams) and not ((rules & BP_GMR_TEAMS) or (rules & BP_GMR_FORCETEAMS)) then continue end	-- we want team games
		if (hell != hellgamemode) then continue end	-- we want the right type of hell
		
		table.insert(possibleGamemodes, k)	-- gamemode checks out, add it!
	end
	
	-- return gamemode table
	return possibleGamemodes
end

framework.HUD_Selection = function(v, p)
	if (battleplus.IsDisplayPlayer(p) ~= 0) then return end	-- reduce rendering as much as possible
	
	local maxfade = v.localTransFlag(v)>>V_ALPHASHIFT    -- value of V_HUDTRANS
	local fade = max(maxfade, (9 - ((leveltime-4))/3))
	
	-- get gamemodes
	local gamemodes = framework.HUD_Selection_GetGamemodes()
	local index = 1+((leveltime+1)%(#gamemodes*3)/3)
	local gamemode = battleplus.gamemodes[gamemodes[index]]	
	
	if (fade <= 9)
		local yoffset = min(7, (leveltime/5))
		
		-- sample text
		local loadNames = {"Selecting Gamemode.", "Selecting Gamemode..", "Selecting Gamemode..."}
		if (leveltime%20 >= 10) then v.drawString(160, (-1+yoffset), loadNames[(leveltime%60/20)+1], fade<<V_ALPHASHIFT, "center") end
		-- backdrop
		v.draw(135, (4+yoffset), v.cachePatch("K_ITBG"), fade<<V_ALPHASHIFT)
		
		-- get stuff from the gamemode
		local patch, colour = gamemode.icon, SKINCOLOR_NONE
		local string = gamemode.name
		
		if (icons_teams)	-- team icons
			if (gamemode.teamicons)	-- need to be valid
				local c = {battleplus.cv_t1c.value, battleplus.cv_t2c.value}
					
				patch, colour = gamemode.teamicons[1], c[1]
				v.draw(135, (4+yoffset), v.cachePatch(gamemode.teamicons[2]), fade<<V_ALPHASHIFT, v.getColormap(-1, c[2]))
			end
			string = gamemode.teamname or gamemode.name
		end
		patch = $ or "K_ITFREE"	-- failsafe (custom gamemodes?)
		
		v.draw(135, (4+yoffset), v.cachePatch(patch), fade<<V_ALPHASHIFT, v.getColormap(-1, colour))
		if (string) then v.drawString(160, (51+yoffset), "*"..string.."*", fade<<V_ALPHASHIFT, "center") end
	end
end