local framework = battleplus.framework

-- TEAMS FUNCTION
framework.Main_Teams = function(gamemode)
	if not (server.bp_server.teamgamemode) then return end	-- we dont need to do this
	
	framework.Main_Teams_HighlightItems()
	-- team sync stuff
	local teams = {server.bp_server.teams[1], server.bp_server.teams[2]}
	for i = 1, 2
		for k, v in ipairs(teams[i])
			if not (v.valid)
			or (v.valid and v.spectator)
				if (v.valid and v.spectator)
					if (v.bp.savedskincolor)
						-- reset skincolor
						battleplus.SmartSetColor(v, false, v.bp.savedskincolor)
						v.skincolor = v.bp.savedskincolor
						v.bp.savedskincolor = nil
					end
					v.bp.team = nil
				end
				table.remove(teams[i], k)
			end
		end
	end
	-- give players a team if they dont have one
	-- otherwise colour them and a marker if they dont have one
	for p in players.iterate
		local mo = p.mo
		
		if not p.spectator
			if not (p.bp.team)
				battleplus.AssignTeam(p) 
			else
				local c = {battleplus.cv_t1c.value, battleplus.cv_t2c.value}
				
				battleplus.SmartSetColor(p, false, c[p.bp.team])
				p.skincolor = c[p.bp.team] -- just for the hud flashing
				-- team marker
				if (mo and mo.valid)
					framework.Main_Teams_Jawz(mo, gamemode)
					if not mo.bp_teammarker or not mo.bp_teammarker.valid
						mo.bp_teammarker = P_SpawnMobj(mo.x, mo.y, mo.z, MT_BP_TEAMMARKER)
						mo.bp_teammarker.target = mo
						mo.bp_teammarker.color = c[p.bp.team]
						mo.bp_teammarker.flags2 = $|MF2_DONTDRAW
					end
				end
			end
		end
	end
	-- make SURE teams are tallied right before intermission
	for p in players.iterate
		if (p.exiting == 2)	
			battleplus.TallyTeams(true, true)
			return
		end
	end
end

-- TIMELIMIT PRESET
framework.Main_TimeLimit = function(gamemode)
	if not (gamemode.rules & BP_GMR_TIMELIMIT) then return end	-- we dont need this.
	if not (server.bp_server.timelimit) then return end	-- table somehow wasnt setup yet?
	
	local starttime = 6*TICRATE + (3*TICRATE/4)
	local timelimit = gamemode.timelimit.value or 3	-- default to 3 minutes if .timelimit isnt set
	if (gamemode.timelimit and not gamemode.timelimit.value) then timelimit = 0 end
	
	if (timelimit)
		if not (server.bp_server.timelimit.roundstarted)
			if (leveltime > starttime)	-- round would start here
				server.bp_server.timelimit.roundstarted = true
			else	-- otherwise keep at the starting time
				server.bp_server.timelimit.time = ((timelimit*TICRATE)*60)
			end
		else
			if not (server.bp_server.timelimit.roundended)
				for p in players.iterate
					if (p.exiting)	-- round already ended???
						server.bp_server.timelimit.roundended = true
						return
					end
				end
				
				if (server.bp_server.timelimit.overtime)
					if (server.bp_server.timelimit.overtime == 1)
						S_StartSound(nil, sfx_strpst)
					end
					server.bp_server.timelimit.overtime = $+1
				else
					local time = server.bp_server.timeoverride or max(0, server.bp_server.timelimit.roundfreeze and 60*TICRATE or ((timelimit*TICRATE)*60) - (leveltime-starttime))
					server.bp_server.timelimit.time = time
					if not (server.bp_server.timelimit.time)
						if (gamemode.rules & BP_GMR_OVERTIME)
							if not (server.bp_server.preventovertime)
								server.bp_server.timelimit.overtime = 1
							end
						else
							-- end battle here
							server.bp_server.timelimit.roundended = true
							if (gamemode.finishhook) then gamemode.finishhook() end	-- run finish "hook"
							battleplus.TallyTeams(true, true)
							battleplus.EndBattle()
						end
						return
					else
						-- play a sfx in the last 10 seconds, hurry up!
						if (server.bp_server.timelimit.time <= 10*TICRATE)
							if not (server.bp_server.timelimit.time%TICRATE) then S_StartSound(nil, sfx_bptly) end
						end
						
						-- indicate one minute left
						if (timelimit > 1) -- we already started at 1 minute
							if (server.bp_server.timelimit.time == 60*TICRATE)
							and not (server.bp_server.timelimit.roundfreeze)
								server.bp_server.timelimit.roundfreeze = 3*TICRATE
								S_StartSound(nil, sfx_1mnlft)
								-- final lap music but not really support
								if (DynMusic) then DynMusic.flm.TriggerFLM() end
							end
						end
					end
				end
			else
				server.bp_server.timelimit.time = 0
			end
		end
	end
	server.bp_server.timelimit.roundfreeze = $ and $-1	-- freeze the timer
end

-- BUMPERS PRESET
framework.Main_Bumpers_Update = function(p, bumpers)
	if not (bumpers) then return end
	//print("Bumpers needed: "..bumpers)
	//print(p.name.."'s bumpers: "..p.kartstuff[k_bumper])
	
	if (p.kartstuff[k_bumper] > bumpers)	-- we have more than we should
		p.kartstuff[k_bumper] = bumpers
	elseif (p.kartstuff[k_bumper] < bumpers)
		while (p.kartstuff[k_bumper] < bumpers)
			//print("Gave "..p.name.." a bumper")
			local b = battleplus.SpawnBumper(p)
			if (b and b.valid)
				if (leveltime < 130)
					b.flags = $|MF2_DONTDRAW
				end
			end
			p.kartstuff[k_bumper] = $+1
			//print(p.name.."'s bumpers: "..p.kartstuff[k_bumper])
		end
	end
end
framework.Main_Bumpers = function(gamemode)
	if not (gamemode.rules & BP_GMR_BUMPERS)
		-- make sure bumpers never go into critical (can still go higher, just for wanted hacks :cory:)
		for p in players.iterate
			if ((gamemode.rules & BP_GMR_ELIMINATION) and not p.kartstuff[k_bumper])
				p.kartstuff[k_comebackpoints] = 0	-- nope
				continue 
			end
			p.kartstuff[k_bumper] = max(2, $)
		end
		return 
	end
	
	-- handle late joiners
	local bumpers = gamemode.bumpers.value or 3	-- fallback
	local starttime = 6*TICRATE + (3*TICRATE/4)
	for p in players.iterate
		if (p.spectator) then continue end
		
		if (p.bp.spectatorcheck)
			if (leveltime < starttime)
				framework.Main_Bumpers_Update(p, bumpers)
			else
				if not (p.bp.joincheck)
					framework.Main_Bumpers_Update(p, bumpers)
				end
			end
		end
	end
	-- handle team elim
	if not (server.bp_server.teamgamemode) then return end	-- only in team gamemodes (duh)
	-- dont do this if we've already finished
	local finished = false
	for p in players.iterate
		if (p.exiting)
			finished = true
			break
		end
	end
	if (finished) then return end
	-- check teams
	local teams = {server.bp_server.teams[1], server.bp_server.teams[2]}
	for t = 1, 2
		if (#teams[t] == battleplus.GetDeadCount(t))	-- one side is wiped out
			local winning_team = (t == 1) and 2 or 1
		
			battleplus.TallyTeams(true, true)
			-- additional points (we should have server.bp_teamtally stored here)
			local points = (server.bp_server.teamtally[t]/#teams[t])
			for k, v in ipairs(teams[winning_team])
				v.marescore = $+points
				server.bp_server.teamtally[winning_team] = $+points
			end
			local c = {
				{battleplus.ColorToMap(battleplus.cv_t1c.value), battleplus.ColorToString(battleplus.cv_t1c.value)}, 
				{battleplus.ColorToMap(battleplus.cv_t2c.value), battleplus.ColorToString(battleplus.cv_t2c.value)}
			}
			local points_string = (points == 1) and "point" or "points"
			print("The "..(c[winning_team][1])..(c[winning_team][2]).." Team".."\x80".." received "..points.." "..points_string.." each for winning!")
			battleplus.EndBattle()
			break
		end
	end
end
-- NO WANTED PRESET
framework.Main_RemoveWanted = function(gamemode)
	if not (gamemode.rules & BP_GMR_REMOVEWANTED) then return end
	
	for p in players.iterate
		p.kartstuff[k_wanted] = 0	-- this removes the camping wanted
		-- make sure the bumpers (of alive players) STAYS at 2 so bestplayerbumpers never triggers
		if not (gamemode.rules & BP_GMR_BUMPERS)	-- obviously it wont work here!
			if (p.kartstuff[k_bumper]) then p.kartstuff[k_bumper] = 2 end
		end
	end
end
-- SPECTATOR HACK PRESET
framework.Main_SpectatorHack = function()
	local starttime = 6*TICRATE + (3*TICRATE/4)
	for p in players.iterate
		if (p.spectator)
			p.bp.spectatorcheck = true
		else
			if (p.bp.spectatorcheck)
				p.bp.spectatorcheck = nil
			end
		end
		if (p and p.mo) and (leveltime > starttime) and not p.spectator then p.bp.joincheck = true end
	end
end

framework.Main = function(gamemode)
	if (gamemode.func) then gamemode.func() end	-- run the gamemode function
	-- rest of the functions are run after for var hack checks
	framework.Main_Teams(gamemode)	-- just for making SURE teams are tallied at the end
	framework.Main_TimeLimit(gamemode)
	framework.Main_Bumpers(gamemode)	-- handles bumper sync
	framework.Main_RemoveWanted(gamemode)	-- handles removing wanted status
	framework.Main_SpectatorHack()	-- spectator hacks
	framework.Main_ItemOdds(gamemode)
end