-- this got massive so it's in a separate script for management
local framework = battleplus.framework

-- change the colour of items
local flashItems = {
	MT_EGGMANITEM, MT_EGGMANITEM_SHIELD,
	MT_BANANA, MT_BANANA_SHIELD,
	MT_ORBINAUT, MT_ORBINAUT_SHIELD,
	MT_JAWZ, MT_JAWZ_DUD, MT_JAWZ_SHIELD,
	MT_SSMINE_SHIELD, MT_SSMINE,
	MT_BALLHOG
}
for i = 1, #flashItems
	addHook("MobjSpawn", function(item) 
		if not (item and item.valid) then return end
		if not G_BattleGametype() then return end
		if not (server.bp_server.teamgamemode) then return end

		-- add mo to the table of stuff to do
		if not (server.bp_server.teamhighlighting) then return end	-- safety
		server.bp_server.teamhighlighting[#server.bp_server.teamhighlighting+1] = item
	end, flashItems[i])
end
framework.Main_Teams_HighlightItems = function()
	server.bp_server.teamhighlighting = $ or {}
	if not (#server.bp_server.teamhighlighting) then return end
	
	local i = #server.bp_server.teamhighlighting
	while i
		local mo = server.bp_server.teamhighlighting[i]
		-- everything else is copypaste from TSR
		
		-- check if mo is still valid
		if not (mo and mo.valid) or not mo.health
			table.remove(server.bp_server.teamhighlighting, i)
			i = $-1
			continue
		end

		-- now colour the item
		if (mo.target and mo.target.player.bp.team)
			-- needs to be on a team
			if not (mo.target.player.bp.team)
				i = $-1
				continue
			end
			local c = {battleplus.cv_t1c.value, battleplus.cv_t2c.value}
			-- spawn a marker on them if they dont have one already (not for shield items)
			local noMarker = false
			local noMarkerTypes = {MT_BANANA_SHIELD, MT_ORBINAUT_SHIELD, MT_EGGMANITEM_SHIELD, MT_JAWZ_SHIELD, MT_SSMINE}
			for item = 1, #noMarkerTypes
				if (mo.type == noMarkerTypes[item]) then noMarker = true end
			end

			if not (mo.bp_itemmarker and mo.bp_itemmarker.valid) and not noMarker
				mo.bp_itemmarker = P_SpawnMobj(mo.x, mo.y, mo.z, MT_BP_ITEMMARKER)
				mo.bp_itemmarker.target = mo
				mo.bp_itemmarker.color = c[mo.target.player.bp.team]
				mo.bp_itemmarker.flags2 = $|MF2_DONTDRAW
			end

			-- make sure bananas, eggboxes, jawz and ballhogs are coloured, too
			local teamSprites = {
			-- Bananas
			[MT_BANANA] = SPR_TBNA,
			[MT_BANANA_SHIELD] = SPR_TBNA,
			-- Jawz
			[MT_JAWZ] = SPR_TJWZ,
			[MT_JAWZ_DUD] = SPR_TJWZ,
			[MT_JAWZ_SHIELD] = SPR_TJWZ,
			-- Ballhog
			[MT_BALLHOG] = SPR_THOG
			}
			mo.sprite = teamSprites[mo.type] or states[mo.state].sprite

			mo.color = c[mo.target.player.bp.team]
		else
			-- reset sprite
			mo.sprite = states[mo.state].sprite
		end

		-- go to next item in the table
		i = $-1
	end
end

-- jawz targetting
framework.Main_Teams_Jawz = function(mo, gamemode)
	if (gamemode.rules & BP_GMR_OVERRIDETEAMJAWZ) then return end
	if CV_FindVar("friendlyfire").value then return end

	if (mo.player.kartstuff[k_itemtype] == KITEM_JAWZ and mo.player.kartstuff[k_itemheld])
		-- btech jawz targeting with teams involved
		local targ
		local best = -1

		for p in players.iterate
			-- check player stuff first
			if not (p and p.mo and p.mo.health) then continue end -- needs to be alive
			if (p.spectator) then continue end -- cant be a spectator
			if (p == mo.player) then continue end -- dont target ourselves
			if (p.kartstuff[k_bumper] <= 0) then continue end	-- they're dead
			if (p.kartstuff[k_hyudorotimer]) then continue end -- dont target invisible players
			if (mo.player.bp.team == p.bp.team) then continue end -- dont target teammates

			-- get distance & angle
			local d = P_AproxDistance(p.mo.x - (mo.x + mo.momx), p.mo.y - (mo.y + mo.momy))
			local a = mo.angle - R_PointToAngle2(mo.x, mo.y, p.mo.x, p.mo.y)
			if (battleplus.AngToInt(a) > 180) then a = InvAngle(a) end

			-- now let's find a target!
			if (battleplus.AngToInt(a) > 45) then continue end -- dont target players behind you
			if (abs(p.mo.z - (mo.z + mo.momz)) > (1280<<FRACBITS)/8) then continue end -- too high/low compared to player
			if (d > 2*(1280<<FRACBITS)) then continue end -- too far
			
			local avg = (AngleFixed(a) + d)/2
			if (best == -1 or avg < best)
				targ = p
				best = avg
			end
		end

		-- now override the target
		if (targ) -- check if we have a target
			if (mo.player.kartstuff[k_lastjawztarget] ~= #targ) then S_StartSound(targ.mo, sfx_s3k89) end
			mo.player.kartstuff[k_lastjawztarget] = #targ
			mo.player.kartstuff[k_jawztargetdelay] = 5
		end
	end
end
addHook("MobjThinker", function(mo)
	if not G_BattleGametype() then return end
	if not (server.bp_server) then return end	-- uh oh
	if not (server.bp_server.teamgamemode) then return end
	
	if (mo and mo.valid) -- paranoia
		if (mo.target and mo.target.valid)
		and (mo.tracer and mo.tracer.valid)
			local p, p2 = mo.target.player, mo.tracer.player
			if not (p and p2) then return end
			
			if (p.bp.team == p2.bp.team)
			and not CV_FindVar("friendlyfire").value
				mo.tracer = nil
			end
		end
	end
end, MT_JAWZ)

-- team karma items
addHook("TouchSpecial", function(kh, mo)
	if not G_BattleGametype() then return end
	if not (server.bp_server) then return end	-- uh oh
	if not (server.bp_server.teamgamemode) then return end
	
	if not (mo and mo.player) then return end -- just be sure its a player that hit it LOL
	if not (kh.target and kh.target.player) then return end	-- bruh
	if (kh.target.player == mo.player) then return end -- make sure its actually not the player that is in karma comeback because this runs for them for some reason
	
	local p, p2 = kh.target.player, mo.player
	if not (p2.kartstuff[k_bumper]) then return end	-- they need bumpers
	
	if (p.kartstuff[k_comebackmode] == 0 or p.kartstuff[k_comebackmode] == 2)
		-- dont blow up or eggman your teammates
		if (p.bp.team == p2.bp.team)
			if not CV_FindVar("friendlyfire").value
				-- cant blow up your teammates!
				return true
			end
		else
			-- if there's only one teammate left, you got em!!!
			if ((#server.bp_server.teams[p.bp.team] - battleplus.GetDeadCount(p.bp.team)) == 1)
				p.kartstuff[k_yougotem] = 2*TICRATE
				p.kartstuff[k_comebackpoints] = 2
			end
		end
	else
		if (p.bp.team ~= p2.bp.team) and (P_CanPickupItem(p2, 1))
			-- no karma points if you give items to the enemy!
			local poof = P_SpawnMobj(kh.x, kh.y, kh.z, MT_EXPLODE)
			S_StartSound(poof, kh.info.seesound)

			for i = 0, 4
				local fw = P_SpawnMobj(kh.x, kh.y, kh.z, MT_KARMAFIREWORK)
				fw.momx = (kh.target.momx + mo.momx)/2
				fw.momy = (kh.target.momy + mo.momy)/2
				fw.momz = (kh.target.momz + mo.momz)/2
				P_Thrust(fw, FixedAngle((72*i)<<FRACBITS), P_RandomRange(1, 8)*kh.scale)
				P_SetObjectMomZ(fw, P_RandomRange(1, 8)*kh.scale, false)
				fw.color = kh.target.color
			end
				
			p.kartstuff[k_comebackmode] = 0
			p.kartstuff[k_comebacktimer] = 10*TICRATE

			p2.kartstuff[k_itemroulette] = 1
			p2.kartstuff[k_roulettetype] = 1
			return true
		end
	end
end, MT_KARMAHITBOX)