local framework = battleplus.framework
local dprint = battleplus.dprint

-- functions for getting gamemode odds
framework.Setup_GetTeamOdds = function()
	if not (battleplus.cv_teams.value) then return false end	-- disabled teams
	if (battleplus.GetPlayerCount() < 4) then return false end	-- needs at least 4 players
	
	if (battleplus.cv_teams.value == 2) then return true end	-- forced teams
	
	local odds = (battleplus.GetPlayerCount()%2) and 8 or 3	-- less chance for teams with uneven player counts
	if (P_RandomRange(1, odds) == odds) then return true end	-- yeah, all together now!
	
	return false
end
framework.Setup_GetHellOdds = function()
	if not (battleplus.cv_hell.value) then return false end	-- disabled hell
	if (battleplus.cv_hell.value == 2) then return true end	-- forced hell
	
	local odds = (mapheaderinfo[gamemap].menuflags & LF2_HIDEINMENU) and 2 or 10	-- hell maps have an increased chance
	if (P_RandomRange(1, odds) == odds) then return true end	-- yeah... this is bad!
	
	return false
end

-- select our gamemode!
framework.Setup_SelectGamemode = function()
	local teams, hell = framework.Setup_GetTeamOdds(), framework.Setup_GetHellOdds()
	dprint(2, "teams: "..tostring(teams))
	dprint(2, "hell: "..tostring(hell))
	
	-- go through gamemodes
	local possibleGamemodes = {}
	local forced = false
	for k, gamemode in ipairs(battleplus.gamemodes)
		-- force gamemode?
		if (battleplus.cv_forcegamemode.value)	-- are we forcing it?
		and (battleplus.gamemodes[battleplus.cv_forcegamemode.value])	-- gamemode exists
			local rules = battleplus.gamemodes[battleplus.cv_forcegamemode.value].rules
			
			teams = (rules & BP_GMR_TEAMS) and $ or false
			if (rules & BP_GMR_FORCETEAMS) then teams = true end
			hell = (rules & BP_GMR_HELL) and true or false
			table.insert(possibleGamemodes, battleplus.cv_forcegamemode.value)
			forced = true
			break
		else
			dprint(2, "Checking "..gamemode.name)
			-- dont add if saved
			if framework.SavedGamemodes_IsGamemodeSaved(k) then continue end
			
			local rules = gamemode.rules or 0	-- fallback
			local hellgamemode = (rules & BP_GMR_HELL) and true or false	-- this seems smarter
			
			if not (teams) and (rules & BP_GMR_FORCETEAMS) then continue end	-- this gamemode NEEDS teams
			if (teams) and not ((rules & BP_GMR_TEAMS) or (rules & BP_GMR_FORCETEAMS)) then continue end	-- we want team games
			if (hell != hellgamemode) then continue end	-- we want the right type of hell
			
			dprint(2, gamemode.name.." checks out")
			table.insert(possibleGamemodes, k)	-- gamemode checks out, add it!
		end
	end
	
	-- play sound and print name
	local sound = sfx_kc48
	if (hell)
		if (battleplus.cv_hell.value == 2 or (forced))
			sound = sfx_noooo2
		else
			sound = (mapheaderinfo[gamemap].menuflags & LF2_HIDEINMENU) and sfx_noooo3 or sfx_noooo1
		end
	elseif (teams)
		sound = sfx_yeeeah
	end
	S_StartSound(nil, sound)
	
	server.bp_server.teamgamemode, server.bp_server.gamemodehell = teams, hell
	return possibleGamemodes[P_RandomRange(1, #possibleGamemodes)]
end

-- everything relating to setting up gamemodes
-- teams
framework.Setup_Teams = function(gamemode)
	if not (server.bp_server.teamgamemode) then return end
	server.bp_server.teams = {{}, {}}	-- our team table
	
	-- get players in a table
	local assigntable = {}
	for p in players.iterate do table.insert(assigntable, p) end
	
	-- assign players to a team in random order
	if (#assigntable)
		for i = #assigntable, 1, -1
			local n = P_RandomRange(1, i)	-- get a random player
			battleplus.AssignTeam(assigntable[n])
			table.remove(assigntable, n)
		end
	end
end
-- time limit
framework.Setup_TimeLimit = function(gamemode)
	if not (gamemode.rules & BP_GMR_TIMELIMIT) then return end
	-- general vars
	server.bp_server.timelimit = {}	-- namespace for the timelimit
	server.bp_server.timelimit.time = 0	-- current time, used for display
	server.bp_server.timelimit.roundstarted = false
	server.bp_server.timelimit.roundfreeze = 0
	server.bp_server.timelimit.roundended = false
	server.bp_server.timelimit.finishhook = false	-- trust me this is much more useful than it sounds
	server.bp_server.timelimit.overtime = 0
end
-- bumpers
framework.Setup_Bumpers = function(gamemode)
	if not (gamemode.rules & BP_GMR_BUMPERS)
		for p in players.iterate do p.kartstuff[k_bumper] = 2 end
		return 
	end
	
	local bumpers = gamemode.bumpers.value or 3	-- fallback
	//print("Bumpers needed: "..bumpers)
	
	for p in players.iterate
		//print(p.name.."'s bumpers: "..p.kartstuff[k_bumper])
		
		if (p.kartstuff[k_bumper] > bumpers)	-- we have more than we should
			p.kartstuff[k_bumper] = bumpers
			//print(p.name.."'s new bumpers: "..p.kartstuff[k_bumper])
		elseif (p.kartstuff[k_bumper] < bumpers)
			while (p.kartstuff[k_bumper] < bumpers)
				//print("Gave "..p.name.." a bumper")
				local b = battleplus.SpawnBumper(p)
				if (b and b.valid)
					b.flags = $|MF2_DONTDRAW
				end
				p.kartstuff[k_bumper] = $+1
				//print(p.name.."'s new bumpers: "..p.kartstuff[k_bumper])
			end
		end
	end
end

framework.Setup = function()
	server.bp_server = {}	-- server var namespace
	for p in players.iterate do p.bp = {} end	-- player namespace
	COM_BufInsertText(server, "timelimit 0")	-- no
	
	-- select the gamemode, then set everything up!
	server.bp_server.gamemode = framework.Setup_SelectGamemode()
	framework.SavedGamemodes_AddGamemode(server.bp_server.gamemode)
		
	local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
	local name = (server.bp_server.teamgamemode) and gamemode.teamname or gamemode.name
	print("The gamemode is now: "..name)
	-- setup functions
	framework.Setup_Teams(gamemode)	-- setup teams
	framework.Setup_TimeLimit(gamemode)	-- setup timelimit
	framework.Setup_Bumpers(gamemode)	-- setup bumpers
end