local chaosringrunners = battleplus.chaosringrunners

-- handle spawned chaos rings
chaosringrunners.Func_HandleChaosRings = function()
	server.bp_server.crr_chaosrings = $ or {}
	for k, v in ipairs(server.bp_server.crr_chaosrings)
		if not (v and v.valid)
			table.remove(server.bp_server.crr_chaosrings, k)
		end
	end
end

chaosringrunners.Func_ReturnSpawnpoints = function(num)
	local t = {}
	for m in thinkers.iterate("mobj")
		if (m.type ~= MT_RANDOMITEM) then continue end	-- not a random item
		if (m.bp_fuse) then continue end	-- already has a chaos ring on it
		-- dont get items too high up from the ground
		local abs_floorz = battleplus.GetAbsoluteFloorZ(m)
		if (abs(m.z - abs_floorz) > 48*(mapobjectscale)) then continue end	-- too high, we want itemboxes near the floor
			
		t[#t+1] = m
	end
	
	local spawns = {}
	for i = #t, 1, -1
		local spawn = P_RandomRange(1, i)
		table.insert(spawns, t[spawn])
		table.remove(t, spawn)
		-- check if we have enough spawns
		if (#spawns == num) then break end
	end
	return spawns
end
chaosringrunners.Func_SpawnChaosRings = function()
	-- conditions we wouldnt spawn
	local starttime = 6*TICRATE + (3*TICRATE/4)
	
	-- handle delay
	server.bp_server.crr_spawndelay = $ or 0
	if (leveltime >= starttime)	-- decrease delay only after starting
		server.bp_server.crr_spawndelay = $ and $-1
	end
	
	-- count how many chaos rings are "active"
	server.bp_server.crr_chaosrings = $ or {}
	local totalchaosrings = #server.bp_server.crr_chaosrings
	for p in players.iterate
		-- get player chaos rings
		totalchaosrings = $+(chaosringrunners.ReturnChaosRings(p))
	end
	local maxchaosrings = (battleplus.GetPlayerCount() + 1)
	
	if (not (leveltime == 106) and (leveltime < starttime))	-- we havent started yet (but spawn in rings to start out)
	or (server.bp_server.crr_spawndelay)	-- we have a delay on spawning more rings
	or (totalchaosrings >= maxchaosrings)	-- we already have enough chaos rings
		return
	end
	
	-- alright, let's spawn chaos rings
	local spawnsneeded = min(3, maxchaosrings - totalchaosrings)
	local spawnpoints = chaosringrunners.Func_ReturnSpawnpoints(spawnsneeded)
	
	if (#spawnpoints)	-- safety
		for k, v in ipairs(spawnpoints)
			-- get colours if not done so already
			if not (server.bp_server.crr_chaosringcolours)
				server.bp_server.crr_chaosringcolours = {}
				local colours = server.bp_server.crr_chaosringcolours
				
				-- add base game colours
				for i = SKINCOLOR_WHITE, SKINCOLOR_LILAC do colours[#colours+1] = i end -- first add all the colours from the player setup
				-- additional colours I found while snooping around as usual in the source
				local extracolours = {SKINCOLOR_SUPER5, SKINCOLOR_TSUPER5, SKINCOLOR_KSUPER5, SKINCOLOR_PSUPER5, SKINCOLOR_BSUPER5, SKINCOLOR_ASUPER5, SKINCOLOR_GSUPER5, SKINCOLOR_WSUPER5, SKINCOLOR_CSUPER5}
				for k, v in ipairs(extracolours) do colours[#colours+1] = v end
			end
			
			local cring = P_SpawnMobj(v.x, v.y, v.z+(24*mapobjectscale), MT_BP_CRR_CHAOSRING)
			cring.flags2 = $|MF2_DONTDRAW
			-- set colour
			local index = P_RandomRange(1, #server.bp_server.crr_chaosringcolours)
			cring.color = server.bp_server.crr_chaosringcolours[index]
			table.remove(server.bp_server.crr_chaosringcolours, index)
			-- add to stored minimap icons
			server.bp_server.crr_chaosrings[#server.bp_server.crr_chaosrings+1] = cring
			
			-- remove the itembox it spawned on, spawn a new one in
			local box = P_SpawnMobj(v.x, v.y, v.z, MT_RANDOMITEM)
			box.scale = v.scale
			box.bp_fuse = 3*TICRATE -- fake fuse
			box.bp_fusetarget = cring
			-- pop
			local pop = P_SpawnMobj(v.x, v.y, v.z, MT_THOK)
			pop.state = S_XPLD1
			pop.scale = v.scale
			S_StartSound(pop, sfx_pop)
			
			P_RemoveMobj(v)
		end
	end
	server.bp_server.crr_spawndelay = 5*TICRATE + (spawnsneeded)*TICRATE
	S_StartSoundAtVolume(nil, sfx_kc42, 128)
end

-- handle timelimit
chaosringrunners.Func_HandleTimelimit = function()
	if not (server.bp_server.timelimit) then return end	-- bruh
	server.bp_server.crr_rounds = $ or 0
	-- we'll use a special time limit for chaos ring runners, adding time to the timelimit every round
	local starttime = 6*TICRATE + (3*TICRATE/4)
	if not (leveltime > starttime) then return end
	
	-- check if round is already over
	for p in players.iterate
		if (p.exiting) then return end
	end
	
	-- set timelimit overrides
	server.bp_server.timeoverride = max(0, (60*TICRATE) + ((30*TICRATE)*server.bp_server.crr_rounds) - (leveltime-starttime))
	server.bp_server.preventovertime = (chaosringrunners.PlayersInDanger() or (battleplus.GetPlayerCount() == 1))
	
	if not (server.bp_server.timelimit.time)
		if (chaosringrunners.PlayersInDanger() or (battleplus.GetPlayerCount() == 1))
			-- get players to eliminate
			local eliminate_players = {}
			for p in players.iterate
				if not (chaosringrunners.IsPlayerInDanger(p))
					S_StartSound(nil, sfx_strpst, p)
				else
					eliminate_players[#eliminate_players+1] = p
					S_StartSound(nil, sfx_s26d, p)
				end
			end
			
			-- eliminate players
			for k, p in ipairs(eliminate_players)
				if (p.mo and p.mo.valid)	-- paranoia
					local mo = p.mo
							
					-- spawn separate origin for sounds so they dont cut out
					local soundmo = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_THOK)
					soundmo.state = S_INVISIBLE
					soundmo.tics = -1
					S_StartSound(soundmo, skins[mo.skin].soundsid[SKSKLOSE])
									
					-- blow them up
					S_StartSound(soundmo, sfx_s3k4e)
					K_SpawnMineExplosion(mo, mo.color)
					P_DamageMobj(mo, nil, nil, 10000)
				end
				chaosringrunners.ChaosRingBurst(p, chaosringrunners.ReturnChaosRings(p), 7, 24)
				-- set elim vars
				p.kartstuff[k_bumper] = 0
				p.bp.crr_eliminated = true
				print(p.name.." was eliminated!")
			end
			
			-- check if round should end
			if ((battleplus.GetPlayerCount() - battleplus.GetDeadCount()) == 1)	-- somebody won
				-- check if round is already over
				for p in players.iterate
					if (p.exiting) then return end
				end
				
				-- tally up score and add to our victor
				for p in players.iterate
					if (p.kartstuff[k_bumper] and not p.spectator)	-- here's our victor
						local tally = 0
						for s in players.iterate
							if (s ~= p) then tally = $ + s.marescore end
						end
						p.marescore = $+tally
						print(p.name.." received "..tally.." points for winning!")
						if (p.mo and p.mo.valid) then S_StartSound(p.mo, sfx_kwin) end
					end
				end
				battleplus.EndBattle()
				return
			else
				-- update rounds
				server.bp_server.crr_rounds = $+1
				-- reset timelimit
				local time = max(0, (60*TICRATE) + ((30*TICRATE)*server.bp_server.crr_rounds) - (leveltime-starttime))
				server.bp_server.timeoverride = time
				server.bp_server.timelimit.time = time
				server.bp_server.timelimit.overtime = 0
			end
		end
	end
	
	-- spawn a tiebreaker chaos ring
	-- count how many chaos rings are "active"
	server.bp_server.crr_chaosrings = $ or {}
	local totalchaosrings = #server.bp_server.crr_chaosrings
	for p in players.iterate
		-- get player chaos rings
		totalchaosrings = $+(chaosringrunners.ReturnChaosRings(p))
	end
	local maxchaosrings = (battleplus.GetPlayerCount() + 1)
	
	if (server.bp_server.timelimit.overtime >= 30*TICRATE)
	and (server.bp_server.timelimit.overtime%(TICRATE*5))
	and not (#server.bp_server.crr_chaosrings)
	and (totalchaosrings >= maxchaosrings)
		-- do we already have a tiebreaker chaos ring?
		if (server.bp_server.crr_tiebreaker and server.bp_server.crr_tiebreaker.valid) then return end
		
		local spawnpoints = chaosringrunners.Func_ReturnSpawnpoints(1)
		if (#spawnpoints)	-- safety
			for k, v in ipairs(spawnpoints)
				-- get colours if not done so already
				if not (server.bp_server.crr_chaosringcolours)
					server.bp_server.crr_chaosringcolours = {}
					local colours = server.bp_server.crr_chaosringcolours
					
					-- add base game colours
					for i = SKINCOLOR_WHITE, SKINCOLOR_LILAC do colours[#colours+1] = i end -- first add all the colours from the player setup
					-- additional colours I found while snooping around as usual in the source
					local extracolours = {SKINCOLOR_SUPER5, SKINCOLOR_TSUPER5, SKINCOLOR_KSUPER5, SKINCOLOR_PSUPER5, SKINCOLOR_BSUPER5, SKINCOLOR_ASUPER5, SKINCOLOR_GSUPER5, SKINCOLOR_WSUPER5, SKINCOLOR_CSUPER5}
					for k, v in ipairs(extracolours) do colours[#colours+1] = v end
				end
				
				local cring = P_SpawnMobj(v.x, v.y, v.z+(24*mapobjectscale), MT_BP_CRR_CHAOSRING)
				cring.flags2 = $|MF2_DONTDRAW
				-- set colour
				local index = P_RandomRange(1, #server.bp_server.crr_chaosringcolours)
				cring.color = server.bp_server.crr_chaosringcolours[index]
				table.remove(server.bp_server.crr_chaosringcolours, index)
				-- add to stored minimap icons
				server.bp_server.crr_chaosrings[#server.bp_server.crr_chaosrings+1] = cring
				-- make this our tiebreaker chaos ring
				server.bp_server.crr_tiebreaker = cring
				
				-- remove the itembox it spawned on, spawn a new one in
				local box = P_SpawnMobj(v.x, v.y, v.z, MT_RANDOMITEM)
				box.scale = v.scale
				box.bp_fuse = 3*TICRATE -- fake fuse
				box.bp_fusetarget = cring
				-- pop
				local pop = P_SpawnMobj(v.x, v.y, v.z, MT_THOK)
				pop.state = S_XPLD1
				pop.scale = v.scale
				S_StartSound(pop, sfx_pop)
				
				P_RemoveMobj(v)
			end
		end
		S_StartSoundAtVolume(nil, sfx_kc42, 128)
	end
end

-- handle player chaos rings
chaosringrunners.Func_PlayerChaosRings = function(p)
	p.bp.crr_chaosrings = $ or {}
	
	-- respawn all your rings if you spectate (I cant burst them)
	if (p.spectator and chaosringrunners.ReturnChaosRings(p))
		p.bp.crr_chaosrings = {}
	end
	
	-- lose rings if you fall off like a cunt
	local respawn_tics = 35
	if (JUICEBOX and (JUICEBOX.value or JUICEBOX_forcefastrespawn.value))
		-- Juicebox fast respawn support
		respawn_tics = 20
	end
		
	if (p.kartstuff[k_respawn] == respawn_tics)
	and (p.bp.joincheck)	-- join hack
		if (p.mo and p.mo.valid)	-- justified paranoia
			chaosringrunners.ChaosRingBurst(p, 1, 6, 15)
		end
	end
	
	if (p.mo and p.mo.valid)
		-- overlay rings above player
		if (chaosringrunners.ReturnChaosRings(p))
			local drawFlags = {MFE_DRAWONLYFORP1, MFE_DRAWONLYFORP2, MFE_DRAWONLYFORP3, MFE_DRAWONLYFORP4}
			local objects = {}
			
			if (chaosringrunners.ReturnChaosRings(p) == 1)	-- no trig 4 me
				local scale_decrease = (mapobjectscale)/4
				
				-- overlay
				local cringo = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z+(P_GetPlayerHeight(p)*3/2), MT_THOK)
				cringo.scale = $ - (scale_decrease)
				cringo.sprite = SPR_BPCR
				cringo.color = p.bp.crr_chaosrings[1]
				cringo.frame = ((leveltime%26)/2)
				cringo.tics = 1
				objects[#objects+1] = cringo
				
				-- spark
				local radius = (mobjinfo[MT_BP_CRR_CHAOSRING].radius)>>FRACBITS
				local height = (mobjinfo[MT_BP_CRR_CHAOSRING].height)>>FRACBITS
					
				local spark = P_SpawnMobj(cringo.x+P_RandomRange(-radius, radius)*cringo.scale, cringo.y+P_RandomRange(-radius, radius)*cringo.scale, cringo.z+P_RandomRange(0, height)*cringo.scale, MT_SIGNSPARKLE)
				spark.scale = $*3/4 - (scale_decrease)
				spark.color = cringo.color
				spark.colorized = true
				objects[#objects+1] = spark
			else	-- fuck
				local crings = chaosringrunners.ReturnChaosRings(p)
				for i = 1, crings
					local scale_decrease = ((mapobjectscale)/4) + ((mapobjectscale/48)*(crings-1))
				
					-- overlay
					local distance = 28 + (5*(crings-1))
					local angle = p.mo.angle + (ANG1*(leveltime%180)*2) + ((ANG1*(360/crings))*i)
					local newx = p.mo.x+distance*FixedMul(cos(angle), mapobjectscale - scale_decrease)
					local newy = p.mo.y+distance*FixedMul(sin(angle), mapobjectscale - scale_decrease)
					
					local cringo = P_SpawnMobj(newx, newy, p.mo.z+(P_GetPlayerHeight(p)*3/2), MT_THOK)
					cringo.scale = $ - (scale_decrease)
					cringo.sprite = SPR_BPCR
					cringo.color = p.bp.crr_chaosrings[i]
					cringo.frame = ((leveltime%26)/2)
					cringo.tics = 1
					objects[#objects+1] = cringo
					
					if not (P_RandomKey(crings))	-- decrease chance for sparks with more overlays
						local radius = (mobjinfo[MT_BP_CRR_CHAOSRING].radius)>>FRACBITS
						local height = (mobjinfo[MT_BP_CRR_CHAOSRING].height)>>FRACBITS
							
						local spark = P_SpawnMobj(cringo.x+P_RandomRange(-radius, radius)*cringo.scale, cringo.y+P_RandomRange(-radius, radius)*cringo.scale, cringo.z+P_RandomRange(0, height)*cringo.scale, MT_SIGNSPARKLE)
						spark.scale = $*3/4 - (scale_decrease)
						spark.color = cringo.color
						spark.colorized = true
						objects[#objects+1] = spark
					end
				end
			end
		
			for k, v in ipairs(objects)
				for i = 0, (splitscreen) -- go through displayplayers
					if not displayplayers[i] then break end

					if (displayplayers[i] ~= p)
						v.eflags = $|drawFlags[i+1]
					end
				end
				-- somehow this isnt supposed to be displayed at ALL for any displayplayers, so dont draw it
				if not (v.eflags & (MFE_DRAWONLYFORP1|MFE_DRAWONLYFORP2|MFE_DRAWONLYFORP3|MFE_DRAWONLYFORP4))
					v.flags2 = $|MF2_DONTDRAW
				end
			end
		end
	end
	
	-- set score
	if not (p.bp.crr_eliminated or p.exiting)
		p.marescore = chaosringrunners.ReturnChaosRings(p) + battleplus.GetDeadCount()
	else
		if not (p.exiting)
			if (p.bp.crr_savedmarescore == nil) then p.bp.crr_savedmarescore = p.marescore end
			p.marescore = p.bp.crr_savedmarescore
		end
	end
end

-- handle bumper related stuff (elimination, wanted, etc)
chaosringrunners.Func_PlayerBumperHacks = function(p)
	p.kartstuff[k_wanted] = 0	-- no need for camping wanted
	if (p.bp.crr_eliminated)	-- eliminated
		p.kartstuff[k_bumper] = 0	-- make sure this stays at 0
	else	-- not eliminated, set bumpers for wanted!
		if (battleplus.GetPlayerCount() > 2)
			-- the player with the most bumpers become wanted, lets exploit that
			local highest = true
			for i in players.iterate
				if (i.spectator) then continue end
				if not (i.kartstuff[k_bumper]) then continue end
				if (i.bp.crr_eliminated) then continue end

				if (chaosringrunners.ReturnChaosRings(p) < (chaosringrunners.ReturnChaosRings(i)-2))
					highest = false
					break
				end
			end
			p.kartstuff[k_bumper] = (highest) and 3 or 2
		else
			p.kartstuff[k_bumper] = 2
		end
	end
end

chaosringrunners.Func_PlayerDangerAlert = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	if (chaosringrunners.IsPlayerInDanger(p))
	and not (mo.crr_danger and mo.crr_danger.valid)
		mo.crr_danger = P_SpawnMobj(mo.x, mo.y, mo.z, MT_BP_CRR_DANGERALERT)
		mo.crr_danger.target = mo
		mo.crr_danger.color = mo.color
		mo.crr_danger.flags2 = $|MF2_DONTDRAW
	end
end

-- build the function
chaosringrunners.Func = function()
	for p in players.iterate
		chaosringrunners.Func_PlayerChaosRings(p)
		chaosringrunners.Func_PlayerBumperHacks(p)
		chaosringrunners.Func_PlayerDangerAlert(p)
		-- handle spinout hack
		if (p.bp.crr_bursthack)
			chaosringrunners.ChaosRingBurst(p, 1, 6, 15)
			p.bp.crr_bursthack, p.bp.crr_bursthackwanted = nil, nil
		end
	end
	chaosringrunners.Func_HandleChaosRings()
	chaosringrunners.Func_SpawnChaosRings()
	chaosringrunners.Func_HandleTimelimit()
end

-- damage functions
chaosringrunners.PlayerSpin = function(p, inflictor, source)
	if (p.exiting) then return end	-- dont lose rings during results
	if (p.powers[pw_flashing]) then return end	-- no combos
	
	-- take away rings
	-- were they stolen? give them to the thief
	if ((source and source.player) and (source.player.kartstuff[k_sneakertimer] or source.player.kartstuff[k_pogospring])) -- got rammed with a sneaker/stomped with a pogospring?
		if (chaosringrunners.ReturnChaosRings(p))
			S_StartSound(p.mo, sfx_8blose)
		end
		
		local amount = K_IsPlayerWanted(p) and 2 or 1
		for i = 1, amount
			if not (chaosringrunners.ReturnChaosRings(p)) then break end
			
			local index = P_RandomRange(1, #p.bp.crr_chaosrings)
			source.player.bp.crr_chaosrings[#source.player.bp.crr_chaosrings+1] = p.bp.crr_chaosrings[index]
			table.remove(p.bp.crr_chaosrings, index)
		end
	else	-- otherwise burst them out
		p.bp.crr_bursthack, p.bp.crr_bursthackwanted = true, K_IsPlayerWanted(p)
	end
end
chaosringrunners.PlayerSquish = function(p, inflictor, source)
	if (p.exiting) then return end	-- dont lose rings during results
	if (p.powers[pw_flashing]) then return end	-- no combos
	
	-- take away rings
	chaosringrunners.ChaosRingBurst(p, 1, 6, 15)
end
chaosringrunners.PlayerExplode = function(p, inflictor, source)
	if (p.exiting) then return end	-- dont lose rings during results
	if (p.powers[pw_flashing]) then return end	-- no combos
	
	-- take away rings
	chaosringrunners.ChaosRingBurst(p, 1, 5, 22)
end

-- hud function
chaosringrunners.HudFunc_MinimapChaosRings = function(v, p)
	if (battleplus.IsDisplayPlayer(p) ~= 0) then return end	-- reduce rendering as much as possible
	if not (server.bp_server.crr_chaosrings) then return end
	if (p.exiting) then return end
	if (p.kartstuff[k_comebacktimer] and not p.kartstuff[k_bumper]) then return end
	
	for k, cring in ipairs(server.bp_server.crr_chaosrings)
		if not (cring and cring.valid) then continue end	-- case 1: spawn fucking died
		
		-- case 2: spawning in
		if (cring.state == S_BP_CRR_CHAOSRING_SPAWN)
		or ((cring.state == S_BP_CRR_CHAOSRING_FLING) and cring.tics > (states[S_BP_CRR_CHAOSRING_FLING].tics - 40))
			if not (leveltime%10 >= 5) then return end
		end
		
		-- case 3: despawning
		if (cring.state == S_BP_CRR_CHAOSRING_FLING)
		and (cring.tics < 4*TICRATE)
			if not (leveltime%2) then return end
		end
		
		v.drawOnMinimap(cring.x, cring.y, FRACUNIT*3/4, v.cachePatch("CRNGMMAP"), v.getColormap(-1, cring.color))
	end
end
chaosringrunners.HudFunc_PlayerChaosRings = function(v, p)
	-- conditions when we wouldnt draw it normally
	if (p.exiting or p.spectator) then return end
	if not (p.kartstuff[k_bumper]) then return end
	if (p.bp.crr_eliminated) then return end
	
	local crings = chaosringrunners.ReturnChaosRings(p) or 0
	
	local dp = battleplus.IsDisplayPlayer(p)
	local danger = ((leveltime%10 >= 5) and chaosringrunners.IsPlayerInDanger(p))	-- critical!
	
	if (splitscreen) > 1
		local hx = (dp == 1 or dp == 3) and 120 or 3
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_SNAPTORIGHT|V_HUDTRANS, dp)
		
		v.draw(hx, 86, v.cachePatch("K_CNGICO"), flags)
		v.drawString(hx+37, 88, crings, flags|(danger and V_REDMAP or 0), "right")
	else
		local hy = (splitscreen) and 76 or 171
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_HUDTRANS, dp)
		
		v.draw(9, hy, v.cachePatch("K_STCNGN"), flags, v.getColormap(-1, p.skincolor))
		battleplus.DrawKartText(v, 92, hy+3, crings, flags, "right", danger)
	end
end
chaosringrunners.HudFunc_RankingsChaosRings = function(v, p)
	if (splitscreen) then return end
	if (battleplus.IsDisplayPlayer(p) != 0) then return end	-- reduce rendering as much as possible
	if (battleplus.GetPlayerCount() <= 1) then return end	-- not enough players
	if (p.exiting) then return end
	
	-- since battle always displays the top 5... time to cheat.......
	local rankplayers = {}
	for player in players.iterate
		if (player.spectator) then continue end
		table.insert(rankplayers, player)
	end
	table.sort(rankplayers, function(a, b) return a.kartstuff[k_position] < b.kartstuff[k_position] end)
	
	local hudy = 101 - (9 * min(5, #rankplayers))
	for i = 1, min(5, #rankplayers)	-- only 1st to 5th
		-- get the rankplayer
		local rankplayer = rankplayers[i]
		if not (rankplayer and rankplayer.valid) then continue end
		if (rankplayer.spectator) then continue end
		if not (rankplayer.mo) then continue end
		
		local hudx = 26
		if (rankplayer.mo.color)	-- this is a check in the source, I'll keep it here too for consistency???
			local crings = chaosringrunners.ReturnChaosRings(rankplayer) or 0
			if (crings > 0)
				-- draw the first ring
				v.draw(hudx, hudy, v.cachePatch("K_RNGBA"), V_HUDTRANS|V_SNAPTOLEFT, v.getColormap(-1, rankplayer.bp.crr_chaosrings[1]))
				hudx = $ + 4
				
				-- draw subsequent rings
				if (crings > 1)
					for i = 1, (crings - 1)
						local colour = rankplayer.bp.crr_chaosrings[1 + i]
						v.draw(hudx, hudy, v.cachePatch("K_RNGBB"), V_HUDTRANS|V_SNAPTOLEFT, v.getColormap(-1, colour))
						hudx = $ + 3
					end
				end
			end
		end
		
		hudy = $ + 18
	end
end
chaosringrunners.HudFunc = function(v, p, c)
	chaosringrunners.HudFunc_MinimapChaosRings(v, p)
	chaosringrunners.HudFunc_PlayerChaosRings(v, p)
	chaosringrunners.HudFunc_RankingsChaosRings(v, p)
end

local chaosringrunners_info = {
	name = "Chaos Ring Runners",
	objective = "Collect the most chaos rings!",
	icon = "K_CNGRNR",
	func = chaosringrunners.Func,
	playerspin = chaosringrunners.PlayerSpin,
	playersquish = chaosringrunners.PlayerSquish,
	playerexplode = chaosringrunners.PlayerExplode,
	hudfunc = chaosringrunners.HudFunc,
	rules = BP_GMR_TIMELIMIT|BP_GMR_OVERTIME|BP_GMR_ELIMINATION|BP_GMR_REMOVEPOINTS,
	timelimit = chaosringrunners.cv_timelimit,
}
battleplus.AddGamemode(BP_GM_CHAOSRINGRUNNERS, chaosringrunners_info)