battleplus.chaosringrunners = {}
local chaosringrunners = battleplus.chaosringrunners
local dprint = battleplus.dprint

-- chaos ring
freeslot("S_BP_CRR_CHAOSRING_SPAWN", "S_BP_CRR_CHAOSRING", "S_BP_CRR_CHAOSRING_FLING", "MT_BP_CRR_CHAOSRING", "SPR_BPCR")
states[S_BP_CRR_CHAOSRING_SPAWN] = {SPR_BPCR, A|FF_ANIMATE|FF_FULLBRIGHT, 2*TICRATE, nil, 12, 1, S_BP_CRR_CHAOSRING}
states[S_BP_CRR_CHAOSRING] = {SPR_BPCR, A|FF_ANIMATE|FF_FULLBRIGHT, -1, nil, 12, 2, S_BP_CRR_CHAOSRING}
states[S_BP_CRR_CHAOSRING_FLING] = {SPR_BPCR, A|FF_ANIMATE|FF_FULLBRIGHT, (13*TICRATE)+1, nil, 12, 1, S_NULL}
mobjinfo[MT_BP_CRR_CHAOSRING] = {
    spawnstate = S_BP_CRR_CHAOSRING_SPAWN,
	deathstate = S_NULL,
	deathsound = sfx_kc5e,
    radius = 26*FRACUNIT,
    height = 62*FRACUNIT,
	activesound = sfx_tink,
    flags = MF_BOUNCE
}
addHook("MobjThinker", function(cring)
	if not (cring and cring.valid) then return end
	if not (cring.health) then return end
	
	-- vars that'll be useful throughout
	local radius = (cring.radius)>>FRACBITS
	local height = (cring.height)>>FRACBITS
	
	-- spawn sparks 
	local newx = cring.x + (P_RandomRange(-radius, radius)<<FRACBITS)
	local newy = cring.y + (P_RandomRange(-radius, radius)<<FRACBITS)
	local newz = cring.z + (P_RandomRange(0, height)<<FRACBITS)
	
	local spark = P_SpawnMobj(newx, newy, newz, MT_SIGNSPARKLE)
	spark.scale = $*3/4
	spark.color = cring.color
	spark.colorized = true
	
	cring.threshold = $ and $-1
	-- thinkers
	if (cring.state == S_BP_CRR_CHAOSRING_SPAWN)
		cring.flags = ($ & ~(MF_SPECIAL))|(MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT)
		-- fade in
		local fadetime = (cring.tics)-1
		if (fadetime)
			cring.flags2 = $ & ~(MF2_DONTDRAW)
			cring.frame = ($ & ~(FF_TRANSMASK))|TR_TRANS10*(fadetime/7)|FF_FULLBRIGHT
		end
	elseif (cring.state == S_BP_CRR_CHAOSRING)
		-- set flags for the rings
		cring.flags = $|MF_SPECIAL|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT
	elseif (cring.state == S_BP_CRR_CHAOSRING_FLING)
		-- remove the emerald on pits
		if (P_CheckDeathPitCollide(cring)) then P_RemoveMobj(cring) return end
		
		-- burst the chaos ring when time runs out
		if (cring.tics == 1)
			for i = 1, 64
				local spark = P_SpawnMobj(cring.x+P_RandomRange(-radius, radius)*(mapobjectscale), cring.y+P_RandomRange(-radius, radius)*(mapobjectscale), cring.z+P_RandomRange(0, height)*(mapobjectscale), MT_BP_ET_EMERALDSPARK)
				spark.color = cring.color
				spark.momx = P_RandomRange(-7, 7)*spark.scale
				spark.momy = P_RandomRange(-7, 7)*spark.scale
				spark.momz = P_RandomRange(8, 14)*spark.scale
			end
			
			P_KillMobj(cring)
			return
		end
		
		-- set flags
		cring.flags = ($ & ~(MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT))|MF_SPECIAL
		
		-- visibility
		if (cring.tics < 4*TICRATE)
		and (leveltime%2)
			cring.flags2 = $|MF2_DONTDRAW
		else
			cring.flags2 = $ & ~(MF2_DONTDRAW)
		end
		
		-- ghost trail
		local speed = FixedHypot(cring.momz, FixedHypot(cring.momx, cring.momy))
		
		if not (cring.flags2 & MF2_DONTDRAW)
		and (leveltime%2)
		and (speed > FixedMul(cring.scale, FRACUNIT*2))
			local g = P_SpawnGhostMobj(cring)
			g.colorized = true
			g.tics = 5
		end
		
		-- make it bounce
		local oldmomz = cring.bp_oldmomz or cring.momz	-- fallback
		if (P_IsObjectOnGround(cring))
		and (abs(oldmomz) > 3*cring.scale)
			-- bounce off the floor
			cring.momx = FixedMul(FRACUNIT*6/9, $)
			cring.momy = FixedMul(FRACUNIT*6/9, $)
			cring.momz = FixedMul(FRACUNIT/2, -oldmomz)
			S_StartSound(cring, cring.info.activesound)
		end
		cring.bp_oldmomz = cring.momz
	end
end, MT_BP_CRR_CHAOSRING)
addHook("TouchSpecial", function(cring, mo)
	local p = mo.player
	
	if p.kartstuff[k_spinouttimer]
	or p.kartstuff[k_squishedtimer]
	or ((cring.target and cring.target == mo) and p.powers[pw_flashing])	-- anti respawn cheese 
	or not p.kartstuff[k_bumper]
	or p.bp.crr_eliminated
	or cring.threshold	-- only just burst out
		return true 
	end
	
	local radius = (mobjinfo[MT_BP_CRR_CHAOSRING].radius)>>FRACBITS
	local height = (mobjinfo[MT_BP_CRR_CHAOSRING].height)>>FRACBITS
	for i = 1, 64
		local spark = P_SpawnMobj(cring.x+P_RandomRange(-radius, radius)*(mapobjectscale), cring.y+P_RandomRange(-radius, radius)*(mapobjectscale), cring.z+P_RandomRange(0, height)*(mapobjectscale), MT_BP_ET_EMERALDSPARK)
		spark.color = cring.color
		spark.momx = P_RandomRange(-7, 7)*spark.scale
		spark.momy = P_RandomRange(-7, 7)*spark.scale
		spark.momz = P_RandomRange(8, 14)*spark.scale
	end
	
	-- give ring to player
	p.bp.crr_chaosrings[#p.bp.crr_chaosrings+1] = cring.color	-- store ring so we can burst it the same colour later
end, MT_BP_CRR_CHAOSRING)

-- elimination alert
freeslot("S_BP_CRR_DANGERALERT", "MT_BP_CRR_DANGERALERT", "SPR_BPCA", "SPR_BPCF")
states[S_BP_CRR_DANGERALERT] = {SPR_BPCA, A|FF_ANIMATE, -1, nil, 1, 5, S_BP_CRR_DANGERALERT}
mobjinfo[MT_BP_CRR_DANGERALERT] = {
	spawnstate = S_BP_CRR_DANGERALERT,
	deathstate = S_NULL,
	spawnhealth = 1000,
	radius = 36*FRACUNIT,
	height = 37*FRACUNIT,
	dispoffset = -2,
	flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY|MF_DONTENCOREMAP
}
addHook("MobjThinker", function(danger)
	if (danger and danger.valid)
		-- now for remove situations
		if not (danger.target and danger.target.player) then P_RemoveMobj(danger) return end -- they dont have a target, or it's not a player
		if not (danger.target.player.health) then P_RemoveMobj(danger) return end -- target is dead
		if not (chaosringrunners.IsPlayerInDanger(danger.target.player)) then P_RemoveMobj(danger) return end	-- remove if they're not in danger
		
		-- set color
		danger.color = danger.target.color
		
		-- set scale
		local s = 3*danger.target.scale/2 -- base scale
		if not (splitscreen)
			if (displayplayers[0] and displayplayers[0].valid) and (displayplayers[0].mo and displayplayers[0].mo.valid)
				s = min(16*danger.target.scale, danger.target.scale + FixedMul(FixedDiv(abs(P_AproxDistance(displayplayers[0].mo.x - danger.target.x, displayplayers[0].mo.y - danger.target.y)), 1280<<FRACBITS), danger.target.scale))
			end
		end
		danger.destscale = s
		
		danger.flags2 = $ & ~MF2_DONTDRAW
		-- hide on self
		local drawFlags = {MFE_DRAWONLYFORP1, MFE_DRAWONLYFORP2, MFE_DRAWONLYFORP3, MFE_DRAWONLYFORP4}
		for i = 0, (splitscreen) -- go through displayplayers
			if not(displayplayers[i]) then break end
			if (danger.target.player == displayplayers[i]) then danger.eflags = $ & ~(drawFlags[i+1]) continue end -- dont show yourself
					
			-- show on anyone but yourself
			danger.eflags = $|drawFlags[i+1]
		end
		-- somehow this isnt supposed to be displayed at ALL for any displayplayers, so dont draw it
		if not (danger.eflags & (MFE_DRAWONLYFORP1|MFE_DRAWONLYFORP2|MFE_DRAWONLYFORP3|MFE_DRAWONLYFORP4))
			danger.flags2 = $|MF2_DONTDRAW
		end

		-- set position (flip support too)
		local shiftNum = 1
		local xpos, ypos, zpos = danger.target.x+P_RandomRange(-shiftNum, shiftNum)*danger.target.scale, danger.target.y+P_RandomRange(-shiftNum, shiftNum)*danger.target.scale, danger.target.z+P_RandomRange(-shiftNum, shiftNum)*danger.target.scale
		-- flashing
		danger.sprite = (leveltime%4 <= 2) and SPR_BPCA or SPR_BPCF
		
		if (danger.target.eflags & MFE_VERTICALFLIP)
			P_TeleportMove(danger, xpos, ypos, zpos - (P_GetPlayerHeight(danger.target.player)/2) - (-21*danger.target.scale) - (64*s))
			danger.eflags = $|MFE_VERTICALFLIP
		else
			P_TeleportMove(danger, xpos, ypos, zpos + (P_GetPlayerHeight(danger.target.player)/2) + (-21*danger.target.scale) + (64*s))
			danger.eflags = $ & ~MFE_VERTICALFLIP
		end
	end
end, MT_BP_CRR_DANGERALERT)

-- functions
chaosringrunners.ReturnChaosRings = function(p)
	p.bp.crr_chaosrings = $ or {}
	return (#p.bp.crr_chaosrings)
end
chaosringrunners.ChaosRingBurst = function(p, amount, momxy, momz)
	amount = $ * ((K_IsPlayerWanted(p) or p.bp.crr_bursthackwanted) and 2 or 1)
	amount = min(chaosringrunners.ReturnChaosRings(p), $)
	if not (amount) then return end
	S_StartSound(p.mo, sfx_8blose)
	
	for i = 1, amount
		local index = P_RandomRange(1, #p.bp.crr_chaosrings)
		
		-- get angle
		local angle = ANG1*P_RandomRange(1, 360)
		if (p.speed) -- burst in a set order
			angle = R_PointToAngle2(0, 0, -p.mo.momx, -p.mo.momy)
			if (amount > 1)	-- randomize angle based on a range
				local range = 75
				angle = $ - ANG1*(range/2) + ANG1*P_RandomRange(0, range)
			end
		end
		
		-- burst ring
		local cring = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z+(p.mo.height/2), MT_BP_CRR_CHAOSRING)
		cring.state = S_BP_CRR_CHAOSRING_FLING
		P_InstaThrust(cring, angle, momxy*cring.scale)
		cring.momz = momz*cring.scale
		cring.threshold = 10
		cring.color = p.bp.crr_chaosrings[index]
		-- anti respawn cheese
		if (p.kartstuff[k_respawn]) then cring.target = p.mo end
		
		-- set info
		server.bp_server.crr_chaosrings[#server.bp_server.crr_chaosrings+1] = cring
		table.remove(p.bp.crr_chaosrings, index)
	end
end
chaosringrunners.IsPlayerInDanger = function(p)
	if (p.spectator) then return end
	if not (p.kartstuff[k_bumper]) then return end
	if (p.bp.crr_eliminated) then return end
	
	-- check if other players have less chaos rings than us
	local indanger = true
	for p2 in players.iterate
		if (p2.spectator) then continue end
		if not (p2.kartstuff[k_bumper]) then continue end
		if (p2.bp.crr_eliminated) then continue end
		
		-- do they have less than us?
		if (chaosringrunners.ReturnChaosRings(p2) < chaosringrunners.ReturnChaosRings(p))
			-- we are not in danger
			indanger = false
			break
		end
	end
	
	-- second check to make sure there isnt a universal tie
	local tied = true
	local prevchaosrings
	for p2 in players.iterate
		if (p2.spectator) then continue end
		if not (p2.kartstuff[k_bumper]) then continue end
		if (p2.bp.crr_eliminated) then continue end
		
		if (prevchaosrings ~= nil)
			if (chaosringrunners.ReturnChaosRings(p2) ~= prevchaosrings)
				tied = false
				break
			end
		else
			prevchaosrings = chaosringrunners.ReturnChaosRings(p2)
		end
	end
	
	return (indanger and not tied)
end
chaosringrunners.PlayersInDanger = function()	-- quick way to check if there are any players in danger
	local indanger = false
	for p in players.iterate
		if (p.spectator) then continue end
		if not (p.kartstuff[k_bumper]) then continue end
		if (p.bp.crr_eliminated) then continue end
		
		indanger = chaosringrunners.IsPlayerInDanger(p)
		if (indanger) then break end
	end
	
	return indanger
end