local eggfection = battleplus.eggfection

eggfection.Func_PlayerCheese = function(p)
	if (p.bp.ef_infected) then return end
	
	local mo = p.mo
	if ((mo and mo.valid) and not mo.health)	-- are we dead? we should get infected!
		eggfection.InfectPlayer(p)
	end
end

eggfection.Func_PlayerInfection = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	if (p.bp.ef_infected)
		if (P_RandomByte() < 56)
			local rad = (mo.radius)>>FRACBITS	-- player radius
			local newx = mo.x + mo.momx + (P_RandomRange(-rad, rad)<<FRACBITS)
			local newy = mo.y + mo.momy + (P_RandomRange(-rad, rad)<<FRACBITS)
			local newz = mo.z + mo.momz + (P_RandomRange(0, mo.height>>FRACBITS)<<FRACBITS)
					
			local drop = P_SpawnMobj(newx, newy, newz, MT_BP_EF_METALDROP)
			drop.momx = FixedMul(mo.scale, P_RandomRange(-2, 2)*FRACUNIT)
			drop.momy = FixedMul(mo.scale, P_RandomRange(-2, 2)*FRACUNIT)
		end
		
		mo.friction = $ - ($/256)	-- go slower
		mo.colorized = true
		mo.color = SKINCOLOR_SKUNK
		p.bp.ef_wasinfected = true
	else
		if not (p.bp.ef_wasinfected) then return end
		
		-- reset stuff
		mo.colorized = false
		mo.color = p.skincolor
		p.bp.ef_wasinfected = nil
	end
end

eggfection.Func_Gamestate = function()
	local starttime = 6*TICRATE + (3*TICRATE/4)
	if not (leveltime > (starttime + TICRATE/2)) then return end
	
	-- end the game when everyone's infected
	if (eggfection.GetInfectedPlayers() == battleplus.GetPlayerCount())
	and (battleplus.GetPlayerCount())	-- just so rounds dont instantly end when nobody is playing
		-- check if round is already over
		for p in players.iterate
			if (p.exiting) then return end
		end
		
		battleplus.EndBattle()
	end
	
	-- no infected players? start the infection!
	if not (eggfection.GetInfectedPlayers())
	and (battleplus.GetPlayerCount() > 1)	-- we need more than 1 player
		-- tag a random player
		local targets = {}
		for p in players.iterate
			if (p.spectator) then continue end
			//if (p.bp.ef_infected) then continue end	 -- redundant
			
			table.insert(targets, p)
		end
		eggfection.InfectPlayer(targets[P_RandomRange(1, #targets)], nil, true)	-- don't give score for this
	end
end

-- intoxicate items from infected players for identification
local toxicItems = {
	MT_EGGMANITEM, MT_EGGMANITEM_SHIELD,
	MT_BANANA, MT_BANANA_SHIELD,
	MT_ORBINAUT, MT_ORBINAUT_SHIELD,
	MT_JAWZ, MT_JAWZ_DUD, MT_JAWZ_SHIELD,
	MT_SSMINE_SHIELD, MT_SSMINE,
	MT_BALLHOG
}
for i = 1, #toxicItems
	addHook("MobjSpawn", function(item) 
		if not (item and item.valid) then return end
		if not G_BattleGametype() then return end
		if not battleplus.IsGamemode(BP_GM_EGGFECTION) then return end	-- not emerald thief

		-- add mo to the table of stuff to do
		if not (server.bp_server.ef_toxicitems) then return end	-- safety
		server.bp_server.ef_toxicitems[#server.bp_server.ef_toxicitems+1] = item
	end, toxicItems[i])
end
eggfection.Func_ToxicItems = function()
	if not (server.bp_server.ef_toxicitems)
		server.bp_server.ef_toxicitems = {}
		if not (#server.bp_server.ef_toxicitems) then return end
	end
	
	local i = #server.bp_server.ef_toxicitems
	while i
		local mo = server.bp_server.ef_toxicitems[i]
		-- everything else is copypaste from TSR
		
		-- check if mo is still valid
		if not (mo and mo.valid) or not mo.health
			table.remove(server.bp_server.ef_toxicitems, i)
			i = $-1
			continue
		end
		
		mo.ef_savedcolor = $ or mo.color	-- save color
		-- now make the item seethe
		if (mo.target)
			if (mo.target.player and mo.target.player.bp.ef_infected)
				if (P_RandomByte() < 56)
					local rad = (mo.radius)>>FRACBITS	-- player radius
					local newx = mo.x + mo.momx + (P_RandomRange(-rad, rad)<<FRACBITS)
					local newy = mo.y + mo.momy + (P_RandomRange(-rad, rad)<<FRACBITS)
					local newz = mo.z + mo.momz + (P_RandomRange(0, mo.height>>FRACBITS)<<FRACBITS)
					
					local drop = P_SpawnMobj(newx, newy, newz, MT_BP_EF_METALDROP)
					drop.momx = FixedMul(mo.scale, P_RandomRange(-2, 2)*FRACUNIT)
					drop.momy = FixedMul(mo.scale, P_RandomRange(-2, 2)*FRACUNIT)
				end
								
				mo.colorized = true
				mo.color = SKINCOLOR_SKUNK
			else
				mo.colorized = false
				mo.color = mo.ef_savedcolor
			end
		else
			mo.colorized = false
			mo.color = mo.ef_savedcolor
		end
		
		i = $-1
	end
end

eggfection.Func = function()
	for p in players.iterate
		eggfection.Func_PlayerCheese(p)
		eggfection.Func_PlayerInfection(p)
		
		-- score
		p.bp.ef_score = $ or 0
		p.marescore = p.bp.ef_score
	end
	eggfection.Func_Gamestate()
	eggfection.Func_ToxicItems()
end

eggfection.FinishHook = function()
	-- get infected players to tally up additional points
	local infected_tally = 0
	for p in players.iterate
		if not (p.bp.ef_infected) then continue end
		
		-- add their score
		infected_tally = $ + (p.bp.ef_score + 1)	-- get additional points for how many are infected!
	end
	
	-- add score
	local surviving_count = battleplus.GetPlayerCount() - eggfection.GetInfectedPlayers()
	if (surviving_count)	-- don't divide by 0
		local names = {}
		local points = infected_tally / surviving_count
		for p in players.iterate
			if (p.spectator) then continue end
			if (p.bp.ef_infected) then continue end
			p.bp.ef_score = $ + points
			table.insert(names, p.name)
		end
		
		-- construct the string of players who won!
		local the_winners = ""
		for i = 1, #names
			-- append player name
			the_winners = $ + names[i]
			
			-- append connecting phrase
			if (i != #names)
				local phrase = (i == #names - 1) and " and " or ", "
				the_winners = $ + phrase
			end
		end
		
		-- print who won!
		local points_string = (points == 1) and "point" or "points"
		print(the_winners.." received "..points.." "..points_string.." "..((#names > 1) and "each " or "").."for winning!")
	end
	
	-- sync surviving players
	local surviving_sum = 0
	-- get points
	for p in players.iterate
		if (p.spectator) then continue end
		if (p.bp.ef_infected) then continue end
		surviving_sum = $ + p.bp.ef_score
	end
	-- give points
	for p in players.iterate
		if (p.spectator) then continue end
		if (p.bp.ef_infected) then continue end
		p.bp.ef_score = surviving_sum
	end
end

-- only thing we need is to have reticules on infected players
eggfection.HudFunc = function(v, p, c)
	if (battleplus.IsDisplayPlayer(p) != 0) then return end	-- reduce rendering as much as possible
	if (p.exiting) then return end
	if not (eggfection.GetInfectedPlayers()) then return end	-- no tagged players to even draw
	
	for tp in players.iterate
		if not (tp.mo and tp.mo.valid) then continue end
		if not (tp.bp.ef_infected) then continue end
		v.drawOnMinimap(tp.mo.x, tp.mo.y, FRACUNIT*5/4, v.cachePatch("MMAPEGGF"))
	end
end

-- spreading the infection!
eggfection.PlayerDamage = function(p, inflictor, source)
	if not (p and p.mo) then return end	-- you need to exist
	if not (source and source.player) then return end	-- source needs to exist
	
	if not (p.bp.ef_infected)	-- not infected? get infected if they're infected!
		if not (source.player.bp.ef_infected) then return end	-- nvm, you're good
		
		eggfection.InfectPlayer(p, source.player)
	else	-- we're infected, give the source points if they're not infected
		if (source.player.bp.ef_infected) then return end	-- nvm, they good
		
		local points = 1
		source.player.bp.ef_score = $ and $ + points or points
		battleplus.SpawnBattlePoints(source.player, p, points)
	end
end
eggfection.PlayerBump = function(mo, mo2)
	if not battleplus.IsGamemode(BP_GM_EGGFECTION) then return end	-- not egg-fection
	
	if not ((mo2 and mo2.valid) and mo2.player) then return end
	local p, p2 = mo.player, mo2.player -- lazy
	
	-- a SHITTON of checks now
	if (p2.bp.ef_infected or not p.bp.ef_infected) then return end	-- p2 is already infected, or p2 isnt infected
	if (p.kartstuff[k_justbumped] or p2.kartstuff[k_justbumped]) then return end	-- just bumped
	if (p.kartstuff[k_squishedtimer] or p.kartstuff[k_spinouttimer]) then return end	-- not damaged
	if (p.powers[pw_flashing] and not p.kartstuff[k_hyudorotimer]) then return end	-- not in flashtics (excluding hyudoro)
	if (p2.kartstuff[k_invincibilitytimer] or (p2.kartstuff[k_growshrinktimer] > 0) or p2.kartstuff[k_hyudorotimer]) then return end	-- powerups
	
	if (abs(mo.z - mo2.z) <= P_GetPlayerHeight(p))	-- ladies and gentlemen, we got em
		eggfection.InfectPlayer(p2, p)
		K_SpinPlayer(p2)
		P_PlayRinglossSound(mo2)
		K_PlayHitEmSound(mo)
	end
end
addHook("MobjCollide", eggfection.PlayerBump, MT_PLAYER)
addHook("MobjMoveCollide", eggfection.PlayerBump, MT_PLAYER)

local eggfection_info = {
	name = "Egg-fection",
	objective = "Avoid the Metal Virus!",
	icon = "K_EGGFCT",
	func = eggfection.Func,
	finishhook = eggfection.FinishHook,
	playerspin = eggfection.PlayerDamage,
	playersquish = eggfection.PlayerDamage,
	playerexplode = eggfection.PlayerDamage,
	hudfunc = eggfection.HudFunc,
	rules = BP_GMR_TIMELIMIT|BP_GMR_REMOVEWANTED|BP_GMR_REMOVEPOINTS,
	timelimit = eggfection.cv_timelimit,
	itemodds = eggfection.ItemOdds,
}
battleplus.AddGamemode(BP_GM_EGGFECTION, eggfection_info)