local emeraldthief = battleplus.emeraldthief

-- handle teams
emeraldthief.TallyTeams = function(set, finish)
	if not (server.bp_server.teamgamemode) then return end	-- we dont need to tally

	if (server.bp_server.teamtally)	-- already tallied, use cached values
		for t = 1, 2
			for k, v in ipairs(server.bp_server.teams[t])
				if not (v and v.valid) then continue end
				v.marescore = server.bp_server.teamtally[t]
			end
		end
	else
		local sums = {0, 0}
		for t = 1, 2
			-- get points on the team
			for k, v in ipairs(server.bp_server.teams[t])
				if not (v and v.valid) then continue end
				
				local score = v.bp.et_score or 0
				sums[t] = $+score
			end
			-- give points
			if not (set) then continue end
			for k, v in ipairs(server.bp_server.teams[t])
				if not (v and v.valid) then continue end
				v.bp.et_score = sums[t]
			end
		end
		-- cache values if need be
		if (finish) then server.bp_server.teamtally = sums end
	end
end
emeraldthief.Func_SetupTeamScores = function()
	if not (server.bp_server and server.bp_server.teamgamemode) then return end	-- only in the team gamemode version
	
	-- make sure team scores are setup
	local default = emeraldthief.cv_countdown.value
	server.bp_server.et_teamtimers = $ or {default*3/2, default*3/2}
	local timers = server.bp_server.et_teamtimers
	
	-- make sure the timer doesnt stay below 10 if a team doesnt have the emerald
	for team = 1, 2
		if (timers[team] >= 10) then continue end	-- no need, we're above 10
		
		local has_emerald = false
		for k, v in ipairs(server.bp_server.teams[team])
			if not (emeraldthief.PlayerHasEmerald(v)) then continue end
			-- if we reach here, they have the emerald
			has_emerald = true
			break	-- no need to continue
		end
		if (has_emerald) then continue end
		
		-- we reached here, that means we need to push the timer back up
		timers[team] = 10
	end
end

-- spawn the emerald
emeraldthief.Func_ReturnSpawnpoints = function()
	local t = {}
	for m in thinkers.iterate("mobj")
		if (m.type ~= MT_RANDOMITEM) then continue end	-- not a random item
		-- dont get items too high up from the ground
		local abs_floorz = battleplus.GetAbsoluteFloorZ(m)
		if (abs(m.z - abs_floorz) > 48*(mapobjectscale)) then continue end	-- too high, we want itemboxes near the floor

		t[#t+1] = m
	end
	return t	-- return spawns
end
emeraldthief.Func_GetSpawnpoint = function()
	local spawns = emeraldthief.Func_ReturnSpawnpoints()
	//print("Total Spawnpoints: "..#spawns)
	-- find the spawn that's as far as possible from all players
	local spawn, best = 1, 0
	for k, v in ipairs(spawns)
		local best_dist = 0
		for p in players.iterate
			if not (p and p.mo) or p.spectator then continue end
			local dist = R_PointToDist2(p.mo.x, p.mo.y, v.x, v.y)

			-- is the emerald closer to this player?
			if (dist < best_dist or not best_dist)
				best_dist = dist
			end
		end

		if (best_dist > best)
			spawn = k
			best = best_dist
		end
	end

	return spawns[spawn]
end
emeraldthief.Func_SpawnEmerald = function(message, color)
	-- spawn the emerald
	local spawn = emeraldthief.Func_GetSpawnpoint()	-- spawnpoint
	local emerald = P_SpawnMobj(spawn.x, spawn.y, spawn.z + (24*(mapobjectscale)), MT_BP_ET_EMERALD)
	emerald.flags2 = $|MF2_DONTDRAW

	-- remove the itembox it spawned on, spawn a new one in
	local box = P_SpawnMobj(spawn.x, spawn.y, spawn.z, MT_RANDOMITEM)
	box.scale = spawn.scale
	box.bp_fuse = 3*TICRATE -- fake fuse
	box.bp_fusetarget = emerald

	local pop = P_SpawnMobj(spawn.x, spawn.y, spawn.z, MT_THOK)
	pop.state = S_XPLD1
	pop.scale = spawn.scale
	S_StartSound(pop, sfx_pop)
	P_RemoveMobj(spawn)

	-- random colour selection
	local colours = {}
	for i = SKINCOLOR_WHITE, SKINCOLOR_LILAC do colours[#colours+1] = i end -- first add all the colours from the player setup
	-- additional colours I found while snooping around as usual in the source
	local extracolours = {SKINCOLOR_SUPER5, SKINCOLOR_TSUPER5, SKINCOLOR_KSUPER5, SKINCOLOR_PSUPER5, SKINCOLOR_BSUPER5, SKINCOLOR_ASUPER5, SKINCOLOR_GSUPER5, SKINCOLOR_WSUPER5, SKINCOLOR_CSUPER5}
	for k, v in ipairs(extracolours) do colours[#colours+1] = v end
	emerald.color = color or colours[P_RandomRange(1, #colours)]

	-- print message
	if (message)
		print(message)
		S_StartSound(nil, sfx_strpst)
	end

	return emerald
end
emeraldthief.Func_HandleSpawning = function()
	if not (server.bp_server.et_thief)
		if not (server.bp_server.et_emerald)	-- emerald hasnt spawned in once yet
			-- set server stuff
			server.bp_server.et_emerald = emeraldthief.Func_SpawnEmerald()
			server.bp_server.et_color = server.bp_server.et_emerald.color
		elseif (server.bp_server.et_emerald and not server.bp_server.et_emerald.valid)
			server.bp_server.et_emerald = emeraldthief.Func_SpawnEmerald("The emerald respawned somewhere on the map!", server.bp_server.et_color)
			server.bp_server.et_color = server.bp_server.et_emerald.color
		end
	else
		if not (server.bp_server.et_thief.valid)	-- they left
		or (server.bp_server.et_thief.valid and server.bp_server.et_thief.spectator)	-- they spectated
			server.bp_server.et_thief = nil
			server.bp_server.et_emerald = emeraldthief.Func_SpawnEmerald("The emerald respawned somewhere on the map!", server.bp_server.et_color)
			server.bp_server.et_color = server.bp_server.et_emerald.color
		end
	end
end

-- handle the emerald thief
emeraldthief.Func_PlayerThief = function(p)
	if not (emeraldthief.PlayerHasEmerald(p)) then p.bp.et_itemboxtime = 0 return end	-- we're not the thief
	if (p.exiting) then return end	-- no need to do this after finishing
	local mo = p.mo

	-- karma items!
	if (p.kartstuff[k_itemroulette])
	and (p.kartstuff[k_roulettetype] != 2)	-- don't prevent eggman roulette
		p.kartstuff[k_roulettetype] = 1
	end

	-- burst an item box from the player
	if not (p.bp.et_itemboxtime)
		-- spawn the item
		-- we want to make sure the itembox has at least some momentum
		local angle = P_RandomRange(1, 360)*ANG1
		
		local item = P_SpawnMobj(mo.x, mo.y, mo.z+(mo.height/2), MT_BP_ET_RANDOMITEM)
		item.target = mo	-- so we can't pickup our own items!
		item.color, item.colorized = mo.color, true
		item.momx = FixedMul(item.scale, P_RandomRange(1, 4)*cos(angle))
		item.momy = FixedMul(item.scale, P_RandomRange(1, 4)*sin(angle))
		item.momz = mo.momz + FixedMul(item.scale, P_RandomRange(10, 12)*FRACUNIT)
		S_StartSound(item, sfx_s3k51)
		
		-- set timer before next item
		local reducemultiply = {0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3}	-- there's probably a more optimal way to do this but I cant be fucked tbh
		local reducetime = (TICRATE/2 * reducemultiply[battleplus.GetPlayerCount()])

		p.bp.et_itemboxtime = (5*TICRATE/2) - reducetime
	else
		p.bp.et_itemboxtime = $-1
	end
	
	-- we dont want certain items
	if (p.kartstuff[k_itemtype])
		local illegalItems = {KITEM_ROCKETSNEAKER, KITEM_INVINCIBILITY, KITEM_GROW, KITEM_HYUDORO}
		for i = 1, #illegalItems
			local item = illegalItems[i]
			if (p.kartstuff[k_itemtype] == item)
				K_DropItems(p)
				break
			end
		end
	end
	
	-- store countdown here
	local countdown = p.bp.et_countdown
	if (countdown[2])
		-- drop emerald while respawning
		local respawn_tics = 35
		if (JUICEBOX and (JUICEBOX.value or JUICEBOX_forcefastrespawn.value))
			-- Juicebox fast respawn support
			respawn_tics = 20
		end
		
		if (p.kartstuff[k_respawn] == respawn_tics)
			emeraldthief.BurstEmerald(p, 5, 13)
			battleplus.SmartSetColor(p, false, p.skincolor)
		end
		mo.friction = $ - ($/176)	-- go slower
		-- set colour
		local flash = (leveltime%6 >= 3)
		battleplus.SmartSetColor(p, flash, p.skincolor, flash, flash)
		-- spawn sparks on player
		local rad = (mo.radius*2)>>FRACBITS	-- player radius
		local newx = mo.x + (P_RandomRange(-rad, rad)<<FRACBITS)
		local newy = mo.y + (P_RandomRange(-rad, rad)<<FRACBITS)
		local newz = mo.z + (P_RandomRange(0, mo.height>>FRACBITS)<<FRACBITS)

		local spark = P_SpawnMobj(newx, newy, newz, MT_SIGNSPARKLE)
		spark.scale = $*3/4
		spark.color = p.skincolor
		spark.colorized = true
		-- emerald overlays (yes, overlay*s*, we're spawning 2 for splitscreen hacks)
		local drawFlags = {MFE_DRAWONLYFORP1, MFE_DRAWONLYFORP2, MFE_DRAWONLYFORP3, MFE_DRAWONLYFORP4}
		for o = 1, 2
			local objects = {}	-- easy way to set this for multiple things
			local transparent = (o == 1)	-- first one is our transparent one

			objects[1] = P_SpawnMobj(mo.x, mo.y, mo.z+(P_GetPlayerHeight(p)*3/2), MT_THOK)
			objects[1].sprite = SPR_ETEM
			objects[1].color = p.skincolor
			objects[1].frame = ((leveltime%12)/4)|((transparent) and FF_TRANS50 or 0)
			objects[1].tics = 1
			if not (transparent)	-- spawn sparks on emerald
				local emerald = objects[1]
				local radius = (mobjinfo[MT_BP_ET_EMERALD].radius)>>FRACBITS
				local height = (mobjinfo[MT_BP_ET_EMERALD].height)>>FRACBITS

				objects[2] = P_SpawnMobj(emerald.x+P_RandomRange(-radius, radius)*(mapobjectscale), emerald.y+P_RandomRange(-radius, radius)*(mapobjectscale), emerald.z+P_RandomRange(0, height)*(mapobjectscale), MT_SIGNSPARKLE)
				objects[2].scale = $*3/4
				objects[2].color = p.skincolor
				objects[2].colorized = true
			end

			for k, v in ipairs(objects)
				for i = 0, (splitscreen) -- go through displayplayers
					if not displayplayers[i] then break end

					if (emeraldthief.PlayerHasEmerald(displayplayers[i]) == (transparent))
						v.eflags = $|drawFlags[i+1]
					end
				end
				-- somehow this isnt supposed to be displayed at ALL for any displayplayers, so dont draw it
				if not (v.eflags & (MFE_DRAWONLYFORP1|MFE_DRAWONLYFORP2|MFE_DRAWONLYFORP3|MFE_DRAWONLYFORP4))
					v.flags2 = $|MF2_DONTDRAW
				end
			end
		end
	end
	-- do stuff with countdown
	if (countdown[1])
		if (mo and mo.health) and not p.kartstuff[k_respawn]
			-- reduce timer if you're not cheesing!
			countdown[1] = $-1
		end
	else
		p.bp.et_score = $+1	-- 1 point for surviving a second
		
		-- check gamestate
		if (server.bp_server and server.bp_server.teamgamemode)
			local timers = server.bp_server.et_teamtimers
			if not (timers) then countdown[1] = TICRATE return end	-- uh... the fuck?
			
			timers[p.bp.team] = $-1
			if (timers[p.bp.team])	-- still surviving
				countdown[1] = TICRATE
				for p2 in players.iterate
					if (battleplus.IsDisplayPlayer(p2) ~= 0) then continue end	-- reduce rendering as much as possible
					S_StartSound(nil, timers[p.bp.team] <= 10 and sfx_bptly or sfx_ptally, p2)
				end
			else
				local teams = {server.bp_server.teams[1], server.bp_server.teams[2]}
				local losing_team = (p.bp.team == 1) and 2 or 1
		
				emeraldthief.TallyTeams(true, true)
				-- additional points (we should have server.bp_teamtally stored here)
				local points = (server.bp_server.teamtally[losing_team]/#teams[losing_team])
				for k, v in ipairs(teams[p.bp.team])
					v.bp.et_score = $+points
					server.bp_server.teamtally[p.bp.team] = $+points
				end
				local c = {
					{battleplus.ColorToMap(battleplus.cv_t1c.value), battleplus.ColorToString(battleplus.cv_t1c.value)}, 
					{battleplus.ColorToMap(battleplus.cv_t2c.value), battleplus.ColorToString(battleplus.cv_t2c.value)}
				}
				local points_string = (points == 1) and "point" or "points"
				print("The "..(c[p.bp.team][1])..(c[p.bp.team][2]).." Team".."\x80".." received "..points.." "..points_string.." each for winning!")
				
				-- emerald thief stuff
				for p2 in players.iterate
					p2.marescore = p2.bp.et_score or 0
				end
				emeraldthief.BurstEmeraldSparks(mo, P_GetPlayerHeight(p)*3/2, p.skincolor)
				S_StartSound(mo, sfx_etwin)
				S_StartSound(mo, sfx_kwin)
				battleplus.SmartSetColor(p, false, p.skincolor)
				
				battleplus.EndBattle()
			end
		else
			countdown[2] = $-1
			if (countdown[2])	-- still surviving
				countdown[1] = TICRATE
				for p2 in players.iterate
					if (battleplus.IsDisplayPlayer(p2) ~= 0) then continue end	-- reduce rendering as much as possible
					S_StartSound(nil, countdown[2] <= 10 and sfx_bptly or sfx_ptally, p2)
				end
			else	-- just won
				local tally = 0
				for p2 in players.iterate
					if (p2 ~= p) then tally = $ + p2.bp.et_score end
				end
				local points_string = (tally == 1) and "point" or "points"
				p.bp.et_score = $+tally
				p.marescore = p.bp.et_score
				print(p.name.." received "..tally.." "..points_string.." for winning!")
				emeraldthief.BurstEmeraldSparks(mo, P_GetPlayerHeight(p)*3/2, p.skincolor)
				S_StartSound(mo, sfx_etwin)
				S_StartSound(mo, sfx_kwin)
				battleplus.SmartSetColor(p, false, p.skincolor)
				battleplus.EndBattle()
			end
		end
	end
end

emeraldthief.Func_HandleTeamHUDShift = function()
	if not (server.bp_server and server.bp_server.teamgamemode) then return end	-- only in the team gamemode version
	
	server.bp_server.et_hxshift = $ or 0
	
	local emerald_team = 0
	-- check if they have the emerald
	for team = 1, 2
		local has_emerald = false
		for k, v in ipairs(server.bp_server.teams[team])
			if not (emeraldthief.PlayerHasEmerald(v)) then continue end
			-- if we reach here, they have the emerald
			has_emerald = true
			break	-- no need to continue
		end
		if not (has_emerald) then continue end
		
		emerald_team = team
	end
	
	local hx_teamposition = 19
	if (emerald_team == 1 and server.bp_server.et_hxshift > -(hx_teamposition))
		server.bp_server.et_hxshift = max(-(hx_teamposition), $-3)
	elseif (emerald_team == 2 and server.bp_server.et_hxshift < hx_teamposition)
		server.bp_server.et_hxshift = min(hx_teamposition, $+3)
	elseif not (emerald_team)	-- neutral position
		if (server.bp_server.et_hxshift < 0)
			server.bp_server.et_hxshift = min(0, $+3)
		elseif (server.bp_server.et_hxshift > 0)
			server.bp_server.et_hxshift = max(0, $-3)
		end
	end
end

-- build the function
emeraldthief.Func = function()
	emeraldthief.Func_SetupTeamScores()
	emeraldthief.Func_HandleTeamHUDShift()
	emeraldthief.Func_HandleSpawning()
	for p in players.iterate
		p.bp.et_score = $ or 0	-- score validity check
		emeraldthief.Func_SetJawzTarget(p)	-- make jawz target the thief if it can
		emeraldthief.Func_PlayerThief(p)	-- emerald thief code
		if not (p.exiting) then p.marescore = p.bp.et_score end	-- set score
		-- handle spinout hack
		if (p.bp.et_bursthack)
			emeraldthief.BurstEmerald(p, 5, 13)
			battleplus.SmartSetColor(p, false, p.skincolor)
			p.bp.et_bursthack = nil
		end
	end
end

-- damage functions
emeraldthief.PlayerSpin = function(p, inflictor, source)
	if not (emeraldthief.PlayerHasEmerald(p)) then return end
	if (p.exiting) then return end	-- dont lose the emerald during results

	-- a thief stole it from a thief!?!?!
	if ((source and source.player) and (source.player.kartstuff[k_sneakertimer] or source.player.kartstuff[k_pogospring])) -- got rammed with a sneaker/stomped with a pogospring?
		source.player.bp.et_countdown = $ or {}	-- init table
		source.player.bp.et_countdown = {TICRATE, max(10, $[2] or emeraldthief.cv_countdown.value)}
		emeraldthief.BurstEmeraldSparks(source, 0, source.player.skincolor)
		print(source.player.name.." stole the Emerald from "..p.name.."!")
		S_StartSound(source, sfx_etgrab)
		S_StartSound(source, sfx_kslow)
		for p2 in players.iterate
			if (p2 == source.player) then continue end
			if (battleplus.IsDisplayPlayer(p2) ~= 0) then continue end	-- reduce rendering as much as possible
			S_StartSound(nil, sfx_itrolm, p2)
		end
		server.bp_server.et_thief = source.player	-- update info

		battleplus.SmartSetColor(p, false, p.skincolor)
		
		-- hitfeed support
		if (hitfeed)
			local s = {{source.player.name, source.hf_overridenamecolor or source.player.skincolor}, {skins[source.skin].facemmap, source.player.skincolor}, source.player}
			HF_SendHitMessage(s, nil, "K_HMETE1", server.bp_server.et_color)
		end
	else	-- otherwise burst it out
		p.bp.et_bursthack = true	-- we'll check 1 tic later, then we can see our momentum after being hit
	end
	-- play sound
	if (p.mo and p.mo.valid)	-- justified paranoia
		S_StartSound(p.mo, sfx_8blose)
	end
end
emeraldthief.PlayerSquish = function(p, inflictor, source)
	if not (emeraldthief.PlayerHasEmerald(p)) then return end
	if (p.exiting) then return end	-- dont lose the emerald during results

	-- burst the emerald
	emeraldthief.BurstEmerald(p, 5, 13)
	battleplus.SmartSetColor(p, false, p.skincolor)
	if (p.mo and p.mo.valid)	-- justified paranoia
		S_StartSound(p.mo, sfx_8blose)
	end
end
emeraldthief.PlayerExplode = function(p, inflictor, source)
	if not (emeraldthief.PlayerHasEmerald(p)) then return end
	if (p.exiting) then return end	-- dont lose the emerald during results

	--- burst the emerald
	emeraldthief.BurstEmerald(p, 4, 22)
	battleplus.SmartSetColor(p, false, p.skincolor)
	if (p.mo and p.mo.valid)	-- justified paranoia
		S_StartSound(p.mo, sfx_8blose)
	end
end

-- hud function
emeraldthief.HUDFunc_Minimap = function(v, p)
	if (battleplus.IsDisplayPlayer(p) ~= 0) then return end	-- reduce rendering as much as possible
	if (p.exiting) then return end

	local emerald = server.bp_server.et_emerald
	local thief = server.bp_server.et_thief

	if (thief and thief.valid)	-- there's a thief, mark them!
		if not (thief.mo and thief.mo.valid) then return end
		v.drawOnMinimap(thief.mo.x, thief.mo.y, FRACUNIT, v.cachePatch("MMAPWANT"))
	else	-- mark the emerald on the map instead
		-- case 1: emerald is fucking DEAD
		if not (emerald and emerald.valid) then return end

		-- case 2: spawning in
		if (emerald.state == S_BP_ET_EMERALD_SPAWN)
		or ((emerald.state == S_BP_ET_EMERALD_FLING) and emerald.tics > (states[S_BP_ET_EMERALD_FLING].tics - 40))
			if not (leveltime%10 >= 5) then return end
		end
		
		-- case 3: despawning
		if (emerald.state == S_BP_ET_EMERALD_FLING)
		and (emerald.tics < 4*TICRATE)
			if not (leveltime%2) then return end
		end

		v.drawOnMinimap(emerald.x, emerald.y, FRACUNIT, v.cachePatch("K_ETMMAP"), v.getColormap(-1, emerald.color))
	end
end

emeraldthief.HUDFunc_EmeraldRadar = function(v, p, c)
	if (p.exiting) then return end

	local mo = p.mo
	local emerald = server.bp_server.et_emerald
	local thief = server.bp_server.et_thief

	-- we want to disable check, unless we have the emerald!
	if (thief and thief.valid)
	and (thief == p)
		hud.enable("check")
		return	-- we don't want to render anything else
	end

	-- there's nothing to track?
	if not (emerald and emerald.valid)
	and not (thief and thief.valid)
		return
	end

	-- we have no mo to work with
	if not (mo and mo.valid) then return end

	-- get what xyz coords we'll be tracking
	local trackx, tracky, trackz
	if (thief and thief.valid)
		if not (thief.mo and thief.mo.valid) then return end	-- no mo? fuck no
		trackx, tracky, trackz = thief.mo.x, thief.mo.y, thief.mo.z
	else
		trackx, tracky, trackz = emerald.x, emerald.y, emerald.z
	end
	if (trackx == nil or tracky == nil or trackz == nil) then return end	-- failsafe

	-- get position of the emerald
	local x, y, behind, scale = battleplus.MapToScreen(p, c, trackx, tracky, trackz)
	-- offset y position
	if (thief and thief.valid)
		local thiefheight = FixedMul(scale, thief.mo.height)*P_MobjFlip(thief.mo)
		local emeraldheight = FixedMul(scale, (mobjinfo[MT_BP_ET_EMERALD].height))*P_MobjFlip(thief.mo)
		local offset = FixedMul(scale, 25*FRACUNIT)*P_MobjFlip(thief.mo)
		
		y = $ - thiefheight - emeraldheight - offset
	else
		local emeraldheight = FixedMul(scale, emerald.height)*P_MobjFlip(emerald)
		local offset = FixedMul(scale, 10*FRACUNIT)*P_MobjFlip(emerald)
		
		y = $ - emeraldheight - offset
	end

	x = $/FRACUNIT
	y = $/FRACUNIT

	-- define patch vars
	local patch_prefix = "FR"
	local flags = 0
	-- set patch vars based on conditions
	if (behind)
		x = min(306, max(14, $))
		y = 200
		patch_prefix = "BH"
		flags = V_SNAPTOBOTTOM
	elseif (x < 28 or x > 292)
		local right = (x > 292)

		x = (right) and 320 or 0
		y = 100
		patch_prefix = "SI"
		flags = (right) and V_SNAPTORIGHT|V_FLIP or V_SNAPTOLEFT
		x = $ + ((right) and 25 or 0)
	end
	
	x = min(320, max(0, $))
	y = min(200, max(0, $))

	-- make splitscreen work. :oh:
	local gscale = FRACUNIT
	if splitscreen
		flags = battleplus.ReturnSplitFlags(flags, p.splitscreenindex)

		if splitscreen
			y = $/2
		end
		if splitscreen > 1
			x = $/2
			gscale = FRACUNIT/2
		end
	end
	
	local color = (thief and thief.valid) and thief.skincolor or server.bp_server.et_color
	v.drawScaled(x<<FRACBITS, y<<FRACBITS, gscale, v.cachePatch("K_EMY"..patch_prefix..(leveltime & 1 and 2 or 1)), V_HUDTRANS|flags, v.getColormap(-1, color))
end

emeraldthief.HUDFunc_EmeraldThief = function(v, p)
	if (battleplus.IsDisplayPlayer(p) ~= 0) then return end	-- reduce rendering as much as possible
	if (p.exiting) then return end

	-- get vars
	local thief = server.bp_server.et_thief
	if not (thief and thief.valid) then return end	-- no thief? dont draw
	local timer = (server.bp_server and server.bp_server.teamgamemode) and server.bp_server.et_teamtimers[thief.bp.team] or thief.bp.et_countdown[2]

	-- get drawing vars
	local hx, hy, flags = 265, 129, V_SNAPTOBOTTOM|V_SNAPTORIGHT|V_HUDTRANS
	if (splitscreen) > 1
		hx, hy, flags = 259, 145, V_SNAPTOBOTTOM|V_SNAPTORIGHT|V_HUDTRANS
		if (splitscreen == 3)
			hx, flags = 131, V_SNAPTOBOTTOM|V_HUDTRANS
		end
	end

	-- retrieve the 1st and 2nd integers as separate variables
	local s = tostring(timer)
	local nums = {}
	if (timer >= 10)	-- more than 10, use 2 numbers
		nums = {tonumber(s:sub(1, 1)), tonumber(s:sub(2, 2))}
	else	-- only 1 number
		nums = {0, timer}
	end

	-- poster + character icon
	v.draw(hx, hy, v.cachePatch((splitscreen) > 1 and "K_4MYTHF" or "K_EMYTHF"), flags, v.getColormap(-1, thief.skincolor))
	if (thief.mo and thief.mo.valid)	-- justified paranoia
		v.draw(hx+((splitscreen) > 1 and 13 or 8), hy+((splitscreen) > 1 and 16 or 21), v.cachePatch(skins[thief.mo.skin].facewant), flags, v.getColormap(TC_RAINBOW, thief.skincolor))
	end
	-- numbers
	if not (timer <= 10 and leveltime%10 >= 5)
		v.draw(hx+((splitscreen) > 1 and 28 or 24), hy+((splitscreen) > 1 and 3 or 5), v.cachePatch(((splitscreen) > 1 and "WTSONT" or "WTFONT")..nums[1]), flags, v.getColormap(-1, thief.skincolor))
		v.draw(hx+((splitscreen) > 1 and 34 or 30), hy+((splitscreen) > 1 and 3 or 5), v.cachePatch(((splitscreen) > 1 and "WTSONT" or "WTFONT")..nums[2]), flags, v.getColormap(-1, thief.skincolor))
	end
end

emeraldthief.HUDFunc_TeamsTimers = function(v, p)
	-- dont draw conditions
	if not (server.bp_server and server.bp_server.teamgamemode) then return end	-- only in the team gamemode version
	if (battleplus.IsDisplayPlayer(p) ~= 0) then return end	-- reduce rendering as much as possible
	if (p.exiting) then return end	-- it's finished
	
	-- set drawing vars
	local hx, hy, flags = 237, (splitscreen) and 3 or 9, V_SNAPTOTOP|V_SNAPTORIGHT|V_HUDTRANS
	if (splitscreen > 1)
		hx, hy, flags = 154, 3, V_SNAPTOTOP|V_HUDTRANS
	end
	
	-- offset hy depending on if timelimit is drawn
	local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
	if (not (splitscreen) and hud.enabled("time"))	-- timelimit is drawn (unless in splitscreen, timelimit isnt drawn there)
	or ((gamemode.rules & BP_GMR_TIMELIMIT) and server.bp_server.timelimit)
		hy = $+21
	end
	
	-- get team score numbers!
	local timers
	if (server.bp_server.et_teamtimers)
		timers = {server.bp_server.et_teamtimers[1], server.bp_server.et_teamtimers[2]}
	else
		local default = emeraldthief.cv_countdown.value
		timers = {default*3/2, default*3/2}
	end
	
	local nums = {{}, {}}
	for team = 1, 2
		local s = tostring(timers[team])
		
		-- separate sums into individual numbers
		local t = nums[team]
		for i = 1, s:len()
			t[#t+1] = tonumber(s:sub(i, i))
		end
	end
	
	-- get colors
	local colors = {battleplus.cv_t1c.value, battleplus.cv_t2c.value}
	-- last 10 seconds? start flashing!
	for team = 1, 2
		if not (timers[team] <= 10) then continue end	-- not low enough
		
		-- check if they have the emerald
		local has_emerald = false
		for k, v in ipairs(server.bp_server.teams[team])
			if not (emeraldthief.PlayerHasEmerald(v)) then continue end
			-- if we reach here, they have the emerald
			has_emerald = true
			break	-- no need to continue
		end
		if not (has_emerald) then continue end
		
		local colormap_sets = {
			{SKINCOLOR_GREY, SKINCOLOR_SILVER, SKINCOLOR_SILVER, SKINCOLOR_WHITE, SKINCOLOR_SILVER, SKINCOLOR_SILVER}, -- 1: No colour mapping (light white/black shades are so limited....)
			{SKINCOLOR_BLACK, SKINCOLOR_NICKEL, SKINCOLOR_SILVER, SKINCOLOR_WHITE, SKINCOLOR_SILVER, SKINCOLOR_NICKEL}, -- 2: V_GREYMAP
			{SKINCOLOR_LEATHER, SKINCOLOR_BROWN, SKINCOLOR_BEIGE, SKINCOLOR_WHITE, SKINCOLOR_BEIGE, SKINCOLOR_BROWN}, -- 3: V_BROWNMAP
			{SKINCOLOR_ROSE, SKINCOLOR_PINK, SKINCOLOR_LILAC, SKINCOLOR_WHITE, SKINCOLOR_LILAC, SKINCOLOR_PINK}, -- 4: V_PINKMAP
			{SKINCOLOR_RED, SKINCOLOR_RASPBERRY, SKINCOLOR_SALMON, SKINCOLOR_WHITE, SKINCOLOR_SALMON, SKINCOLOR_RASPBERRY}, -- 5: V_REDMAP
			{SKINCOLOR_PUMPKIN, SKINCOLOR_ORANGE, SKINCOLOR_TANGERINE, SKINCOLOR_WHITE, SKINCOLOR_TANGERINE, SKINCOLOR_ORANGE}, -- 6: V_ORANGEMAP
			{SKINCOLOR_BRICK, SKINCOLOR_PINK, SKINCOLOR_CARAMEL, SKINCOLOR_WHITE, SKINCOLOR_CARAMEL, SKINCOLOR_PINK}, -- 7: V_PEACHMAP
			{SKINCOLOR_BRONZE, SKINCOLOR_ROYAL, SKINCOLOR_GOLD, SKINCOLOR_WHITE, SKINCOLOR_GOLD, SKINCOLOR_ROYAL}, -- 8: V_GOLDMAP
			{SKINCOLOR_OLIVE, SKINCOLOR_MUSTARD, SKINCOLOR_YELLOW, SKINCOLOR_WHITE, SKINCOLOR_YELLOW, SKINCOLOR_MUSTARD}, -- 9: V_YELLOWMAP
			{SKINCOLOR_MOSS, SKINCOLOR_MINT, SKINCOLOR_PISTACHIO, SKINCOLOR_WHITE, SKINCOLOR_PISTACHIO, SKINCOLOR_MINT}, -- 10: V_TEAMAP
			{SKINCOLOR_PINETREE, SKINCOLOR_GREEN, SKINCOLOR_MINT, SKINCOLOR_WHITE, SKINCOLOR_MINT, SKINCOLOR_GREEN}, -- 11: V_GREENMAP
			{SKINCOLOR_NAVY, SKINCOLOR_CERULEAN, SKINCOLOR_JAWZ, SKINCOLOR_WHITE, SKINCOLOR_JAWZ, SKINCOLOR_CERULEAN}, -- 12: V_SKYMAP
			{SKINCOLOR_PLATINUM, SKINCOLOR_STEEL, SKINCOLOR_SLATE, SKINCOLOR_WHITE, SKINCOLOR_SLATE, SKINCOLOR_STEEL}, -- 13: V_STEELMAP
			{SKINCOLOR_BLUE, SKINCOLOR_PERIWINKLE, SKINCOLOR_JAWZ, SKINCOLOR_WHITE, SKINCOLOR_JAWZ, SKINCOLOR_PERIWINKLE}, -- 14: V_BLUEMAP
			{SKINCOLOR_PURPLE, SKINCOLOR_LAVENDER, SKINCOLOR_BUBBLEGUM, SKINCOLOR_WHITE, SKINCOLOR_BUBBLEGUM, SKINCOLOR_LAVENDER}, -- 15: V_PURPLEMAP
			{SKINCOLOR_LAVENDER, SKINCOLOR_PASTEL, SKINCOLOR_TOXIC, SKINCOLOR_WHITE, SKINCOLOR_TOXIC, SKINCOLOR_PASTEL} -- 16: V_LAVENDERMAP
		}
				
		local set = colormap_sets[battleplus.ColorToMap(colors[team], false, true)]
		colors[team] = set[((leveltime%12)/2)+1]
	end
	
	-- draw team scores
	for team = 1, 2
		local hx2 = {hx-13 - (12*(#nums[team]-1)), hx+13}
		
		-- divider
		v.draw(hx, hy, v.cachePatch("TMFNTDT"..team), flags, v.getColormap(-1, colors[team]))
		-- numbers
		for i = 1, #nums[team]
			local n = nums[team][i]
			v.draw(hx2[team], hy, v.cachePatch("TMFNT"..n), flags, v.getColormap(-1, colors[team]))
			hx2[team] = $+12
		end
	end
	
	-- draw the emerald below it!
	if not (server.bp_server.et_color) then return end
	local hx_shift = server.bp_server.et_hxshift or 0
	
	-- check what team has the emerald so we can colour it
	local emerald_team = 0
	for team = 1, 2
		local has_emerald = false
		for k, v in ipairs(server.bp_server.teams[team])
			if not (emeraldthief.PlayerHasEmerald(v)) then continue end
			-- if we reach here, they have the emerald
			has_emerald = true
			break	-- no need to continue
		end
		if not (has_emerald) then continue end
		
		emerald_team = team
	end
	local c = {battleplus.cv_t1c.value, battleplus.cv_t2c.value}
	local color = c[emerald_team] or server.bp_server.et_color
	
	v.drawScaled(((hx-5)<<FRACBITS) - FRACUNIT/2 + (hx_shift<<FRACBITS), (hy+18)<<FRACBITS, FRACUNIT*3/2, v.cachePatch("K_ETMMAP"), flags, v.getColormap(-1, color))
end

emeraldthief.HUDFunc = function(v, p, c)
	emeraldthief.HUDFunc_Minimap(v, p)
	emeraldthief.HUDFunc_EmeraldRadar(v, p, c)
	emeraldthief.HUDFunc_EmeraldThief(v, p)
	emeraldthief.HUDFunc_TeamsTimers(v, p)
end

local emeraldthief_info = {
	name = "Emerald Thief",
	teamname = "Team Emerald Thief",
	objective = "Steal the Emerald!",
	icon = "K_ETHIEF",
	func = emeraldthief.Func,
	playerspin = emeraldthief.PlayerSpin,
	playersquish = emeraldthief.PlayerSquish,
	playerexplode = emeraldthief.PlayerExplode,
	hudfunc = emeraldthief.HUDFunc,
	disabledhuds = {"minirankings", "check"},
	rules = BP_GMR_TEAMS|BP_GMR_REMOVEWANTED|BP_GMR_TIMELIMIT|BP_GMR_REMOVEPOINTS|BP_GMR_HIDETEAMSCORES|BP_GMR_OVERRIDETEAMJAWZ,
	timelimit = emeraldthief.cv_timelimit,
	itemodds = emeraldthief.ItemOdds,
}
battleplus.AddGamemode(BP_GM_EMERALDTHIEF, emeraldthief_info)