battleplus.emeraldthief = {}
local emeraldthief = battleplus.emeraldthief

-- sounds
freeslot("sfx_etgrab", "sfx_etwin")

-- emerald item box
freeslot("S_BP_ET_RANDOMITEM", "MT_BP_ET_RANDOMITEM")
states[S_BP_ET_RANDOMITEM] = {SPR_RNDM, A|FF_ANIMATE|FF_FULLBRIGHT, 8*TICRATE, nil, 23, 3, S_NULL}
mobjinfo[MT_BP_ET_RANDOMITEM] = {
	spawnstate = S_BP_ET_RANDOMITEM,
	deathstate = S_NULL,
	radius = 48*FRACUNIT,
	height = 48*FRACUNIT,
	flags = MF_BOUNCE|MF_SPECIAL|MF_DONTENCOREMAP,
}
addHook("MobjThinker", function(item)
	if not (item and item.health) then return end
	
	-- create ghosts while flinging
	if P_IsObjectOnGround(item) then item.et_grounded = true end
	if not (item.et_grounded) and (item.tics%2)
		local g = P_SpawnGhostMobj(item)
		g.color, g.colorized = item.color, true
		g.tics = 5
	end
	
	-- preserve momentum, make it slide
	item.friction = 65535
	
	-- flash when low on tics
	if (item.tics < 4*TICRATE)
		if not (leveltime%2)
			item.flags2 = $|MF2_DONTDRAW
		else
			item.flags2 = $ & ~(MF2_DONTDRAW)
		end
	end
end, MT_BP_ET_RANDOMITEM)
addHook("TouchSpecial", function(item, mo)
	if not battleplus.IsGamemode(BP_GM_EMERALDTHIEF) then return true end	-- not emerald thief
	if not P_CanPickupItem(mo.player, 1) then return true end
	if (item.target and (item.target == mo)) then return true end -- dont pickup our own items!
	
	-- spawn a new itembox immediately instead of waiting for battle conditions
	local pop = P_SpawnMobj(item.x, item.y, item.z, MT_THOK)
	pop.state = S_XPLD1
	pop.scale = item.scale
	S_StartSound(pop, sfx_pop)
		
	-- give item
	mo.player.kartstuff[k_itemroulette] = 1
	P_RemoveMobj(item)
end, MT_BP_ET_RANDOMITEM)

-- emerald sparks
freeslot("S_BP_ET_EMERALDSPARK", "MT_BP_ET_EMERALDSPARK", "SPR_ETFS")
states[S_BP_ET_EMERALDSPARK] = {SPR_ETFS, A|FF_ANIMATE|FF_FULLBRIGHT, -1, nil, 1, 1, S_BP_ET_EMERALDSPARK}
mobjinfo[MT_BP_ET_EMERALDSPARK] = {
	spawnstate = S_BP_ET_EMERALDSPARK,
	deathstate = S_NULL,
	radius = 2*FRACUNIT,
	width = 2*FRACUNIT,
	flags = MF_NOBLOCKMAP,
}
addHook("MobjThinker", function(spark)
	if (P_IsObjectOnGround(spark)) then P_RemoveMobj(spark) end
end, MT_BP_ET_EMERALDSPARK)

-- emerald
freeslot("S_BP_ET_EMERALD_SPAWN", "S_BP_ET_EMERALD", "S_BP_ET_EMERALD_FLING", "MT_BP_ET_EMERALD", "SPR_ETEM")
states[S_BP_ET_EMERALD_SPAWN] = {SPR_ETEM, A|FF_ANIMATE|FF_FULLBRIGHT, 2*TICRATE, nil, 2, 2, S_BP_ET_EMERALD}
states[S_BP_ET_EMERALD] = {SPR_ETEM, A|FF_ANIMATE|FF_FULLBRIGHT, -1, nil, 2, 4, S_BP_ET_EMERALD}
states[S_BP_ET_EMERALD_FLING] = {SPR_ETEM, A|FF_ANIMATE|FF_FULLBRIGHT, (13*TICRATE)+1, nil, 2, 2, S_NULL}
mobjinfo[MT_BP_ET_EMERALD] = {
    spawnstate = S_BP_ET_EMERALD_SPAWN,
	deathstate = S_NULL,
	deathsound = sfx_etgrab,
    radius = 36*FRACUNIT,
    height = 54*FRACUNIT,
	activesound = sfx_tink,
    flags = MF_BOUNCE
}
addHook("MobjThinker", function(emerald)
	if not (emerald and emerald.valid) then return end
	if not (emerald.health) then return end
	
	-- vars we (could) use throughout
	local radius = (emerald.radius)>>FRACBITS
	local height = (emerald.height)>>FRACBITS
	
	-- spawn sparks 
	local newx = emerald.x + (P_RandomRange(-radius, radius)<<FRACBITS)
	local newy = emerald.y + (P_RandomRange(-radius, radius)<<FRACBITS)
	local newz = emerald.z + (P_RandomRange(0, height)<<FRACBITS)
	
	local spark = P_SpawnMobj(newx, newy, newz, MT_SIGNSPARKLE)
	spark.scale = $*3/4
	spark.color = emerald.color
	spark.colorized = true
	
	-- thinkers
	if (emerald.state == S_BP_ET_EMERALD_SPAWN)
		emerald.flags = ($ & ~(MF_SPECIAL))|(MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT)
		-- fade in
		local fadetime = (emerald.tics)-1
		if (fadetime)
			emerald.flags2 = $ & ~(MF2_DONTDRAW)
			emerald.frame = ($ & ~(FF_TRANSMASK))|TR_TRANS10*(fadetime/7)|FF_FULLBRIGHT
		end
	elseif (emerald.state == S_BP_ET_EMERALD)
		-- set flags
		emerald.flags = $|MF_SPECIAL|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT
	elseif (emerald.state == S_BP_ET_EMERALD_FLING)
		-- burst the emerald on pits
		if (P_CheckDeathPitCollide(emerald)) then P_RemoveMobj(emerald) return end
		
		-- burst the emerald when tics run out
		if (emerald.tics == 1)
			emeraldthief.BurstEmeraldSparks(emerald, 0, emerald.color)
			P_KillMobj(emerald)
			return 
		end
		
		-- set flags
		emerald.flags = ($ & ~(MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT))|MF_SPECIAL
		
		-- visibility
		if (emerald.tics < 4*TICRATE)
		and (leveltime%2)
			emerald.flags2 = $|MF2_DONTDRAW
		else
			emerald.flags2 = $ & ~(MF2_DONTDRAW)
		end
		
		-- ghost trail
		local speed = FixedHypot(emerald.momz, FixedHypot(emerald.momx, emerald.momy))
		
		if not (emerald.flags2 & MF2_DONTDRAW)
		and (leveltime%2)
		and (speed > FixedMul(emerald.scale, FRACUNIT*2))
			local g = P_SpawnGhostMobj(emerald)
			g.colorized = true
			g.tics = 5
		end
		
		-- make it bounce
		local oldmomz = emerald.bp_oldmomz or emerald.momz	-- fallback
		if (P_IsObjectOnGround(emerald))
		and (abs(oldmomz) > 3*emerald.scale)
			-- bounce off the floor
			emerald.momx = FixedMul(FRACUNIT*6/9, $)
			emerald.momy = FixedMul(FRACUNIT*6/9, $)
			emerald.momz = FixedMul(FRACUNIT/2, -oldmomz)
			S_StartSound(emerald, emerald.info.activesound)
		end
		emerald.bp_oldmomz = emerald.momz
	end
end, MT_BP_ET_EMERALD)
addHook("TouchSpecial", function(e, mo)
	local p = mo.player
	
	if p.kartstuff[k_spinouttimer] 
	or p.kartstuff[k_squishedtimer]
	or ((e.target and e.target == mo) and p.powers[pw_flashing])	-- anti respawn cheese 
		return true
	end
	
	p.bp.et_countdown = $ or {}	-- init table
	p.bp.et_countdown = {TICRATE, max(10, $[2] or emeraldthief.cv_countdown.value)}
	emeraldthief.BurstEmeraldSparks(e, 0, p.skincolor, mo)
	-- shit they stole it
	print(p.name.." stole the Emerald!")
	S_StartSound(mo, sfx_kslow)
	for p2 in players.iterate
		if (p2 == p) then continue end
		if (battleplus.IsDisplayPlayer(p2) ~= 0) then continue end	-- no earrape
		S_StartSound(nil, sfx_itrolm, p2)
	end
	-- update powers
	p.kartstuff[k_hyudorotimer] = min(2, $)
	p.kartstuff[k_growshrinktimer] = min(2, $)
	p.kartstuff[k_invincibilitytimer] = min(2, $)
	p.powers[pw_flashing] = 0
	
	server.bp_server.et_thief = p	-- update info
	
	-- hitfeed support
	if (hitfeed)
		local s = {{p.name, mo.hf_overridenamecolor or p.skincolor}, {skins[mo.skin].facemmap, p.skincolor}, p}
		HF_SendHitMessage(s, nil, "K_HMETE1", server.bp_server.et_color)
	end
end, MT_BP_ET_EMERALD)

-- functions
emeraldthief.BurstEmeraldSparks = function(mo, zoffset, color, mommo)
	local radius = (mobjinfo[MT_BP_ET_EMERALD].radius)>>FRACBITS
	local height = (mobjinfo[MT_BP_ET_EMERALD].height)>>FRACBITS
	
	local added_mom = (mommo) and {mommo.momx, mommo.momy, mommo.momz} or {0, 0, 0}
	for i = 1, 64
		local spark = P_SpawnMobj(mo.x+P_RandomRange(-radius, radius)*(mapobjectscale), mo.y+P_RandomRange(-radius, radius)*(mapobjectscale), (mo.z+zoffset)+P_RandomRange(0, height)*(mapobjectscale), MT_BP_ET_EMERALDSPARK)
		spark.color = color
		spark.momx = P_RandomRange(-7, 7)*spark.scale + (added_mom[1]*3/4)
		spark.momy = P_RandomRange(-7, 7)*spark.scale + (added_mom[2]*3/4)
		spark.momz = P_RandomRange(8, 14)*spark.scale + (added_mom[3]*3/4)
	end
end
emeraldthief.BurstEmerald = function(p, momxy, momz)
	if not (emeraldthief.PlayerHasEmerald(p)) then return end
	local mo = p.mo
	
	local angle = p.speed and R_PointToAngle2(0, 0, -mo.momx, -mo.momy) or ANG1*P_RandomRange(1, 360)
	
	local emerald = P_SpawnMobj(mo.x, mo.y, mo.z+(P_GetPlayerHeight(p)*3/2), MT_BP_ET_EMERALD)
	emerald.color = server.bp_server.et_color
	emerald.state = S_BP_ET_EMERALD_FLING
	P_InstaThrust(emerald, angle, momxy*emerald.scale)
	emerald.momz = momz*emerald.scale
	emeraldthief.BurstEmeraldSparks(emerald, 0, server.bp_server.et_color)
	-- anti-respawn cheese
	if (p.kartstuff[k_respawn]) then emerald.target = mo end
	
	-- player stuff
	print(p.name.." dropped the Emerald!")
	S_StartSound(mo, sfx_8blose)
	
	-- set server vars
	server.bp_server.et_emerald = emerald
	server.bp_server.et_thief = nil
	
	-- hitfeed support
	if (hitfeed)
		if not (HF_animIcon["K_HMETE"]) then HF_animIcon["K_HMETE"] = {2, 2} end
		
		local s = {{p.name, mo.hf_overridenamecolor or p.skincolor}, {skins[mo.skin].facemmap, p.skincolor}, p}
		HF_SendHitMessage(s, nil, "K_HMETE", server.bp_server.et_color)
	end
end
emeraldthief.PlayerHasEmerald = function(p)
	if not (server.bp_server.et_thief and server.bp_server.et_thief.valid) then return end
	return (server.bp_server.et_thief == p)
end