-- handle this separately for code clarity
local emeraldthief = battleplus.emeraldthief

-- Recreate race item boxes for emerald thief
-- touchspecial to spawn new items
addHook("TouchSpecial", function(item, mo)
	if not battleplus.IsGamemode(BP_GM_EMERALDTHIEF) then return end	-- not emerald thief
	if not P_CanPickupItem(mo.player, 1) then return end
	
	-- spawn a new itembox immediately instead of waiting for battle conditions
	local pop = P_SpawnMobj(item.x, item.y, item.z, MT_THOK)
	pop.state = S_XPLD1
	pop.scale = item.scale
	S_StartSound(pop, sfx_pop)
	
	-- spawn a new box
	local box = P_SpawnMobj(item.x, item.y, item.z, MT_RANDOMITEM)
	box.scale = item.scale
	box.bp_fuse = 3*TICRATE -- fake fuse
		
	-- give item
	mo.player.kartstuff[k_itemroulette] = 1
	P_RemoveMobj(item)
end, MT_RANDOMITEM)

-- make sure we cant pickup certain items as the thief
addHook("TouchSpecial", function(mcitem, mo)
	if not battleplus.IsGamemode(BP_GM_EMERALDTHIEF) then return end	-- not emerald thief
	if not P_CanPickupItem(mo.player, 3) then return end
	if not (emeraldthief.PlayerHasEmerald(mo.player)) then return end
	
	local illegalItems = {KITEM_ROCKETSNEAKER, KITEM_INVINCIBILITY, KITEM_GROW, KITEM_HYUDORO}
	for i = 1, #illegalItems
		local item = illegalItems[i]
		if (mcitem.threshold == item)
			return true
		end
	end
end, MT_FLOATINGITEM)

-- make jawz targetting prioritize the emerald thief!
emeraldthief.Func_FindJawzTarget = function(mo)
	if not battleplus.IsGamemode(BP_GM_EMERALDTHIEF) then return end	-- not emerald thief
	if not (mo and mo.valid) then return end
	
	-- go through players to get a target
	local targ
	local best = -1
	for p in players.iterate
		-- check player stuff first
		if not (p and p.mo and p.mo.health) then continue end -- needs to be alive
		if (p.spectator) then continue end -- cant be a spectator
		if (p == mo.player) then continue end -- dont target ourselves
		if (p.kartstuff[k_bumper] <= 0) then continue end	-- they're dead
		if (p.kartstuff[k_hyudorotimer]) then continue end -- dont target invisible players
		if (mo.player.bp.team == p.bp.team) then continue end -- dont target teammates

		-- get distance & angle
		local d = P_AproxDistance(p.mo.x - (mo.x + mo.momx), p.mo.y - (mo.y + mo.momy))
		local a = mo.angle - R_PointToAngle2(mo.x, mo.y, p.mo.x, p.mo.y)
		if (battleplus.AngToInt(a) > 180) then a = InvAngle(a) end

		-- now let's find a target!
		if (battleplus.AngToInt(a) > 45) then continue end -- dont target players behind you
		if (abs(p.mo.z - (mo.z + mo.momz)) > (1280<<FRACBITS)/8) then continue end -- too high/low compared to player
		if (d > 2*(1280<<FRACBITS)) then continue end -- too far
		
		-- found an emerald holder? prioritize!
		if (emeraldthief.PlayerHasEmerald(p))
			targ = p
			break
		end
		
		-- otherwise do normal targetting
		local avg = (AngleFixed(a) + d)/2
		if (best == -1 or avg < best)
			targ = p
			best = avg
		end
	end
	
	return targ
end

-- function to set target
emeraldthief.Func_SetJawzTarget = function(p)
	if not (p.mo and p.mo.valid) then return end
	
	-- jawz targeting override to always lock onto the emerald holder
	if (p.kartstuff[k_itemtype] == KITEM_JAWZ and p.kartstuff[k_itemheld])
		local targ = emeraldthief.Func_FindJawzTarget(p.mo)
		
		if (targ) -- we found the emerald holder in range, override
			if (p.kartstuff[k_lastjawztarget] ~= #targ) then S_StartSound(targ.mo, sfx_s3k89) end
			p.kartstuff[k_lastjawztarget] = #targ
			p.kartstuff[k_jawztargetdelay] = 5
		end
	end
end

-- make sure we cant pickup certain items as the thief

-- item odds

-- you get forced karma items with the emerald
-- generally emerald thief:
	-- removes held items from non carriers
	-- removes invuln from carriers
	
emeraldthief.ItemOdds = {
				//P-Odds	 1  2
			   /*Sneaker*/ { 2, 1 }, // Sneaker
		/*Rocket Sneaker*/ { 0, 0 }, // Rocket Sneaker
		 /*Invincibility*/ { 2, 0 }, // Invincibility
				/*Banana*/ { 0, 3 }, // Banana
		/*Eggman Monitor*/ { 0, 1 }, // Eggman Monitor
			  /*Orbinaut*/ { 8, 2 }, // Orbinaut
				  /*Jawz*/ { 8, 2 }, // Jawz
				  /*Mine*/ { 2, 3 }, // Mine
			   /*Ballhog*/ { 2, 1 }, // Ballhog
   /*Self-Propelled Bomb*/ { 0, 0 }, // Self-Propelled Bomb
				  /*Grow*/ { 2, 0 }, // Grow
				/*Shrink*/ { 0, 0 }, // Shrink
		/*Thunder Shield*/ { 0, 0 }, // Thunder Shield
			   /*Hyudoro*/ { 2, 0 }, // Hyudoro
		   /*Pogo Spring*/ { 0, 0 }, // Pogo Spring
		  /*Kitchen Sink*/ { 0, 0 }, // Kitchen Sink
			/*Sneaker x3*/ { 1, 0 }, // Sneaker x3
			 /*Banana x3*/ { 0, 2 }, // Banana x3
			/*Banana x10*/ { 0, 1 }, // Banana x10
		   /*Orbinaut x3*/ { 2, 0 }, // Orbinaut x3
		   /*Orbinaut x4*/ { 1, 1 }, // Orbinaut x4
			   /*Jawz x2*/ { 2, 1 }  // Jawz x2
}