battleplus.hooliganroundup = {}
local hooliganroundup = battleplus.hooliganroundup

-- the cop's choppers
freeslot("S_BP_HR_CHOPPER_IDLE", "S_BP_HR_CHOPPER_CLOSE", 
"S_BP_HR_CHOPPER_CHASE", "S_BP_HR_CHOPPER_NIPPED",
"S_BP_HR_CHOPPER_DEAD", "S_BP_HR_CHOPPER_EXPLODE",
"MT_BP_HR_CHOPPER", "SPR_BPCH")
freeslot("S_BP_HR_CHOPPER_FIN", "MT_BP_HR_CHOPPER_FIN")
states[S_BP_HR_CHOPPER_IDLE] = {SPR_BPCH, A, -1, nil, 0, 0, S_BP_HR_CHOPPER_IDLE}
states[S_BP_HR_CHOPPER_CLOSE] = {SPR_BPCH, A|FF_ANIMATE, -1, nil, 1, 4, S_BP_HR_CHOPPER_CLOSE}
states[S_BP_HR_CHOPPER_CHASE] = {SPR_BPCH, A|FF_ANIMATE, -1, nil, 1, 3, S_BP_HR_CHOPPER_CHASE}
states[S_BP_HR_CHOPPER_NIPPED] = {SPR_BPCH, A|FF_ANIMATE, -1, nil, 1, 2, S_BP_HR_CHOPPER_NIPPED}
states[S_BP_HR_CHOPPER_DEAD] = {SPR_BPCH, A, 175, nil, 0, 0, S_BP_HR_CHOPPER_EXPLODE}
states[S_BP_HR_CHOPPER_EXPLODE] = {SPR_NULL, A, 1, A_JawzExplode, 0, 0, S_NULL}
mobjinfo[MT_BP_HR_CHOPPER] = {
	spawnstate = S_BP_HR_CHOPPER_IDLE,
	deathstate = S_BP_HR_CHOPPER_DEAD,
	deathsound = sfx_s3k5d,
	radius = 16*FRACUNIT,
	height = 24*FRACUNIT,
	speed = 44*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_SPECIAL
}
addHook("MobjDeath", function(chopper)
	chopper.flags = $ & ~(MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT)
	P_SetObjectMomZ(chopper, 8*FRACUNIT)
	P_InstaThrust(chopper, P_RandomRange(1, 360)*ANG1, 16*FRACUNIT)
end, MT_BP_HR_CHOPPER)

addHook("MobjThinker", function(chopper)
	if not ((chopper and chopper.valid) and chopper.health)
		if (chopper and chopper.valid)	-- must not be alive
		and (chopper.state == S_BP_HR_CHOPPER_DEAD)
			-- flash
			if (leveltime%2)
				chopper.flags2 = $|MF2_DONTDRAW
			else
				chopper.flags2 = $ & ~(MF2_DONTDRAW)
			end
			
			-- go straight to the second state if hitting the ground
			P_SetObjectMomZ(chopper, -gravity, true)	-- increased gravity
			if (P_IsObjectOnGround(chopper))
				chopper.state = S_BP_HR_CHOPPER_EXPLODE
				chopper.flags2 = $ & ~(MF2_DONTDRAW)
			end
		end
		
		-- don't do anything else
		return 
	end
	
	-- we'll use this at the bottom of the thinker
	local chopper_state = S_BP_HR_CHOPPER_IDLE
	
	-- unique behaviour for nipped players!
	local nipinfo = chopper.bp_nipinfo
	if (nipinfo)
		local nipped = nipinfo.target
		if not (nipped and nipped.valid)
			P_RemoveMobj(chopper)
			return
		end
		
		-- set state
		chopper_state = S_BP_HR_CHOPPER_NIPPED
		if (chopper.state != state) then chopper.state = chopper_state end
		
		-- stay stuck to the nipped player
		local angle = nipped.player.frameangle + nipinfo.anglediff
		local newx = nipped.x + (nipinfo.dist * cos(angle))
		local newy = nipped.y + (nipinfo.dist * sin(angle))
		local newz = nipped.z + nipinfo.zdiff
		
		chopper.angle = R_PointToAngle2(chopper.x, chopper.y, nipped.x, nipped.y)
		P_TeleportMove(chopper, newx, newy, newz)
		return
	end
	
	-- no target? kill the chopper
	if not (chopper.target and chopper.target.valid)
		P_KillMobj(chopper)
		return
	end
	local mo = chopper.target	-- laziness
	local p = chopper.target.player	-- double laziness, not writing this over and over
	
	-- player conditions to destroy the chopper
	if not (mo.health)	-- dying
	or (p.kartstuff[k_spinouttimer])	-- spinning out
	or (p.kartstuff[k_squishedtimer])	-- squished
	or (p.kartstuff[k_respawn])	-- respawning
	or (p.bp.hr_chopperdelay)	-- on a delay
		P_KillMobj(chopper)
		return
	end
	
	-- mimic our target
	chopper.scale = mo.scale
	chopper.color = mo.color
	
	local function chopperCheck(p2)
		if (p2.spectator) then return end
		if not (p2.bp) then return end
		local mo2 = p2.mo
		
		if not ((mo2 and mo2.valid) and mo2.health) then return end
		if (p2.bp.team != 1) then return end
		if (p2.bp.hr_caught) then return end
		if (p2.bp.hr_grace) then return end
		
		-- protection
		if (p2.kartstuff[k_hyudorotimer]) then return end
		if (p2.kartstuff[k_growshrinktimer] > 0) then return end
		if (p2.kartstuff[k_invincibilitytimer]) then return end
		
		-- get angle & distance
		local d = P_AproxDistance(mo2.x - (mo.x + mo.momx), mo2.y - (mo.y + mo.momy))
		local a = mo.angle - R_PointToAngle2(mo.x, mo.y, mo2.x, mo2.y)
		if (battleplus.AngToInt(a) > 180) then a = InvAngle(a) end
		
		-- distance checks
		if (battleplus.AngToInt(a) > 30) then return end -- dont target players behind you
		if (abs(mo2.z - (mo.z + mo.momz)) > FixedMul(mo.scale, 64*FRACUNIT)) then return end -- too high/low compared to player
		if (d > FixedMul(mo.scale, 152*FRACUNIT)) then return end -- too far
		
		-- if we're here, we good
		return true
	end
	
	-- find players to chase
	local starttime = 6*TICRATE + (3*TICRATE/4)
	if not (chopper.bp_chased)
	and (leveltime > starttime)
		for p2 in players.iterate
			if not (chopperCheck(p2)) then continue end
			chopper.tracer = p2.mo
			break
		end
	end
	
	if (chopper.tracer and chopper.tracer.valid)
	and (chopperCheck(chopper.tracer.player))
		chopper_state = S_BP_HR_CHOPPER_CHASE
		
		local playerspeed = (FixedHypot(mo.momx, mo.momy) / mo.scale) * FRACUNIT
		A_HomingChase(chopper, max(playerspeed, chopper.info.speed), 1)	-- 1 makes it chase the tracer
		chopper.angle = R_PointToAngle2(chopper.x, chopper.y, chopper.tracer.x, chopper.tracer.y)
		chopper.bp_chased = true
	else
		if (chopper.bp_chased)
			chopper_state = S_BP_HR_CHOPPER_CHASE
			
			-- spawn a destination for A_HomingChase
			local newx = mo.x + FixedMul(mo.scale, 64*cos(mo.angle))
			local newy = mo.y + FixedMul(mo.scale, 64*sin(mo.angle))
			
			local targmo = P_SpawnMobj(newx, newy, mo.z, MT_THOK)
			targmo.state = S_INVISIBLE
			
			local speed = chopper.info.speed*7/4
			if (R_PointToDist2(chopper.x, chopper.y, targmo.x, targmo.y) > FixedMul(chopper.scale, speed))
				chopper.tracer = targmo
				A_HomingChase(chopper, speed, 1)	-- 1 makes it chase the tracer
			else
				-- just go straight to our location, we reached the destination
				chopper.momx, chopper.momy, chopper.momz = 0, 0, 0
				P_TeleportMove(chopper, newx, newy, mo.z)
				chopper.bp_chased = nil	-- remove this too
			end
			
			-- make sure these are removed/reset
			P_RemoveMobj(targmo)
			chopper.tracer = nil
		else
			-- change state when close
			if (mo.player.bp.hr_sirensounded)
				chopper_state = S_BP_HR_CHOPPER_CLOSE
			end
			
			-- follow our target instead
			local newx = mo.x + FixedMul(mo.scale, 64*cos(mo.angle))
			local newy = mo.y + FixedMul(mo.scale, 64*sin(mo.angle))
			
			chopper.momx, chopper.momy, chopper.momz = 0, 0, 0
			chopper.angle = mo.angle
			P_TeleportMove(chopper, newx, newy, mo.z)
		end
	end
	
	-- state setting
	if (chopper.state != chopper_state) then chopper.state = chopper_state end
end, MT_BP_HR_CHOPPER)
-- nip players to jail them!
addHook("TouchSpecial", function(chopper, mo)
	if not (chopper.health) then return true end
	if (chopper.bp_nipinfo) then return true end	-- no.
	local p = mo.player
	
	-- validity checks on the player
	if not ((mo and mo.valid) and mo.health)	-- not alive
	or (p.bp.team != 1)	-- not a hooligan
	or (p.bp.hr_caught)	-- shouldnt be possible, but have a fallback regardless
	or (p.bp.hr_grace)	-- grace period for the hooligan
	or (p.kartstuff[k_hyudorotimer])
	or (p.kartstuff[k_growshrinktimer] > 0)
	or (p.kartstuff[k_invincibilitytimer])
		return true
	end
	
	-- set the chopper to "nipped" mode
	chopper.momx, chopper.momy, chopper.momz = 0, 0, 0
	chopper.bp_nipinfo = {
		-- this all doesnt really work right now
		target = mo,
		dist = R_PointToDist2(chopper.x, chopper.y, mo.x, mo.y) / FRACUNIT,
		zdiff = mo.z - chopper.z,
		anglediff = mo.angle - R_PointToAngle2(chopper.x, chopper.y, mo.x, mo.y),
	}
	
	-- nip the player, jailing them!
	local source = chopper.target
	
	p.bp.hr_caught = true
	source.player.marescore = $+2
	battleplus.SpawnBattlePoints(source.player, mo.player, 2)
	P_DamageMobj(mo, nil, nil, 10000)
	print(source.player.name.." caught "..p.name.."!")
	
	-- hitfeed support
	if (hitfeed)
		local s = {{source.player.name, source.hf_overridenamecolor or source.player.skincolor}, {skins[source.skin].facemmap, source.player.skincolor}, source.player}
		local v = {{p.name, mo.hf_overridenamecolor or p.skincolor}, {skins[mo.skin].facemmap, p.skincolor}, p}
		HF_SendHitMessage(s, v, "K_HMCHOP")
	end
	return true
end, MT_BP_HR_CHOPPER)

-- hooligan jails
freeslot("S_BP_HR_JAIL", "MT_BP_HR_JAIL")
states[S_BP_HR_JAIL] = {SPR_THOK, A, -1, nil, 0, 0, S_BP_HR_JAIL}
mobjinfo[MT_BP_HR_JAIL] = {
	spawnstate = S_BP_HR_JAIL,
	radius = 32*FRACUNIT,
	height = 24*FRACUNIT,
	flags = MF_SPECIAL
}
addHook("TouchSpecial", function(jail, mo)
	local p = mo.player
	if not ((mo and mo.valid) and mo.health)	-- not alive
	or (p.bp.team != 1)	-- not a hooligan
	or (p.bp.hr_caught)	-- nvm, it's a caught hooligan!
	or (p.bp.hr_grace)	-- this'd be cheesy
	or (p.kartstuff[k_spinouttimer])	-- spinning out
	or (p.kartstuff[k_squishedtimer])	-- squished
	or (p.kartstuff[k_respawn])	-- respawning
		return true
	end
	
	-- release all hooligans inside this jail!
	local names = {}	-- we'll print who we released!
	jail.bp_players = $ or {}	-- anxiety
	for k, v in pairs(jail.bp_players)
		-- free them!
		v.kartstuff[k_spinouttimer] = 0
		v.bp.hr_caught = nil
		v.bp.hr_grace = TICRATE*5/2
		v.mo.flags = $ & ~(MF_NOGRAVITY)
		table.insert(names, v.name)
	end
	
	-- construct the string, then print it
	local string = p.name.." freed "
	if (#names > 1)
		for i = 1, #names
			-- append player name
			string = $ + names[i]
			
			-- append connecting phrase
			if (i != #names)
				local phrase = (i == #names - 1) and " and " or ", "
				string = $ + phrase
			end
		end
		-- append punctuation at the end
		string = $ + "!"
	else	-- nvm this is easy lol
		string = p.name.." freed "..names[1].."!"
	end
	print(string)
	
	-- give a point for each hooligan freed!
	p.marescore = $ + (#names)
	battleplus.SpawnBattlePoints(p, nil, #names)
	S_StartSound(nil, sfx_cdfm73)
	
	-- spawn fireworks for breaking out!
	for i = 1, 4
		local firework = P_SpawnMobj(jail.x, jail.y, jail.z, MT_KARMAFIREWORK)
		firework.color = mo.color
		firework.momx = mo.momx / 2
		firework.momy = mo.momy / 2
		firework.momz = mo.momz / 2
		P_Thrust(firework, FixedAngle((72*i)<<FRACBITS), FixedMul(jail.scale, P_RandomRange(1, 8)*FRACUNIT))
		P_SetObjectMomZ(firework, P_RandomRange(1, 8)*FRACUNIT, false)
	end
	
	-- hitfeed support
	if (hitfeed)
		if not (HF_animIcon["K_HMJAI"]) then HF_animIcon["K_HMJAI"] = {2, 2} end
		
		local s = {{p.name, mo.hf_overridenamecolor or p.skincolor}, {skins[mo.skin].facemmap, p.skincolor}, p}
		HF_SendHitMessage(s, nil, "K_HMJAI")
	end
end, MT_BP_HR_JAIL)