local hooliganroundup = battleplus.hooliganroundup

hooliganroundup.Func_Gamestate = function()
	-- dont do this if we've already finished
	local finished = false
	for p in players.iterate
		if (p.exiting)
			finished = true
			break
		end
	end
	if (finished) then return end
	
	-- end the round if all the hooligans are caught!
	local teams = {server.bp_server.teams[1], server.bp_server.teams[2]}
	
	local caught, total = 0, 0
	for p in players.iterate
		if (p.bp.team != 1) then continue end
		
		total = $+1
		if not (p.bp.hr_caught) then continue end
		caught = $+1
	end
	
	-- all are caught? end the round
	if (total >= 1)	-- we need more than one
	and (caught == total)	-- they all need to be caught
		-- additional points for the cops
		battleplus.TallyTeams(true, true)
		
		local points = (server.bp_server.teamtally[1]/#teams[1])
		for k, v in ipairs(teams[2])
			v.marescore = $+points
			server.bp_server.teamtally[2] = $+points
		end
		local c = {battleplus.ColorToMap(battleplus.cv_t2c.value), battleplus.ColorToString(battleplus.cv_t2c.value)}
		local points_string = (points == 1) and "point" or "points"
		print("The "..(c[1])..(c[2]).." Team".."\x80".." received "..points.." "..points_string.." each for winning!")
			
		-- end the battle
		battleplus.EndBattle()
	end
end

hooliganroundup.Func_HandleJails = function()
	server.bp_server.hr_jails = $ or {}
	local jails = server.bp_server.hr_jails
	if not (#jails) then return end	-- no jails
	
	for k, v in pairs(jails)
		-- jail isnt valid anymore, remove it
		if not (v and v.valid)
			table.remove(jails, k)
			continue
		end
		
		-- jail doesnt have players? remove it.
		if not (#v.bp_players)
			P_RemoveMobj(v)
			table.remove(jails, k)
			continue
		end
		
		-- set colour
		v.color = battleplus.cv_t2c.value
		
		-- make sure players are valid in the jail
		for index, p in pairs(v.bp_players)
			if not (p.valid)
			or (p.valid and p.spectator)
				-- not valid? remove them from the jail
				table.remove(v.bp_players, index)
			end
		end
	end
end

hooliganroundup.Func_PlayerHooligans = function(p)
	-- not a hooligan
	if (p.bp.team != 1)
		p.bp.hr_grace = nil
		p.bp.hr_caught = nil
		p.bp.hr_jail = nil
		return 
	end
	local mo = p.mo
	
	-- starting grace
	local starttime = 6*TICRATE + (3*TICRATE/4)
	if not (leveltime > starttime)
		p.bp.hr_grace = 3*TICRATE
	end
	
	-- grace before being able to be caught again
	if (p.bp.hr_grace)
		p.bp.hr_grace = $ and $-1
		mo.flags2 = (p.bp.hr_grace%2) and $|MF2_DONTDRAW or $ & ~(MF2_DONTDRAW)
		
		-- make sure this is reset
		if not (p.bp.hr_grace)
			mo.flags2 = $ & ~(MF2_DONTDRAW)
		end
	end
	
	-- are we caught?
	if (p.bp.hr_caught)
		if ((mo and mo.valid) and mo.health)	-- make sure we're alive
			local jail = p.bp.hr_jail
			if not (jail and jail.valid)	-- uh oh, we gotta have a jail!
				local jails = server.bp_server.hr_jails	-- how many jails currently exist
				local maxjails = max(1, #server.bp_server.teams[1] / 2)	-- the max amount of jails we'll need

				-- see if we should make a new jail
				if not (#jails) -- no jails yet
				or ((#jails < maxjails) and not P_RandomRange(0, 2))	-- randomly put them in a new jail instead!
					-- get potential spawns
					local spawns = {}
					for m in thinkers.iterate("mobj")
						if (m.type != MT_RANDOMITEM) then continue end	-- not a random item
						-- dont get items too high up from the ground
						local abs_floorz = battleplus.GetAbsoluteFloorZ(m)
						if (abs(m.z - abs_floorz) > 48*(mapobjectscale)) then continue end	-- too high, we want itemboxes near the floor

						spawns[#spawns+1] = m
					end
					
					-- select spawn and spawn the jail!
					local spawn = spawns[P_RandomRange(1, #spawns)]
					jail = P_SpawnMobj(spawn.x, spawn.y, spawn.z, MT_BP_HR_JAIL)
					table.insert(server.bp_server.hr_jails, jail)	-- add the jail to a table
					S_StartSound(nil, sfx_cdfm44)
					
					-- remove the itembox it spawned on, spawn a new one in
					local box = P_SpawnMobj(spawn.x, spawn.y, spawn.z, MT_RANDOMITEM)
					box.scale = spawn.scale
					box.bp_fuse = 3*TICRATE -- fake fuse
					box.bp_fusetarget = jail

					local pop = P_SpawnMobj(spawn.x, spawn.y, spawn.z, MT_THOK)
					pop.state = S_XPLD1
					pop.scale = spawn.scale
					S_StartSound(pop, sfx_pop)
					P_RemoveMobj(spawn)
				else	-- no need, just select a random one
					jail = jails[P_RandomRange(1, #jails)]
				end
				
				-- add us to the jail's player table!
				p.bp.hr_jail = jail	-- also make sure this is saved so we don't have to do this again
				jail.bp_players = $ or {}
				table.insert(jail.bp_players, p)
				
				-- get a jail position!
				P_TeleportMove(mo, jail.x, jail.y, jail.z + FixedMul(jail.scale, P_RandomRange(192, 256)*FRACUNIT))
				
				local dist = P_RandomRange(0, 256)
				local angle = P_RandomRange(1, 360)*ANG1
				local failsafe = 1
				while (not P_TryMove(mo, mo.x + (dist*cos(angle)), mo.y + (dist*sin(angle)), true))
					if (failsafe >= 16) then break end	-- man, fuck this
					dist = P_RandomRange(64, 256)
					angle = P_RandomRange(1, 360)*ANG1
					failsafe = $+1
				end
				
				p.bp.hr_jailpos = {mo.x, mo.y, mo.z}
			end
			
			local jailpos = p.bp.hr_jailpos
			p.kartstuff[k_spinouttimer] = 24
			p.kartstuff[k_respawn] = 2	-- make sure we stay in respawn
			mo.flags = $|MF_NOGRAVITY
			P_TeleportMove(mo, jailpos[1], jailpos[2], jailpos[3])
			
			-- thoks going towards the jail
			if not (leveltime%8)
				local thok = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
				thok.frame = ($|FF_FULLBRIGHT) & ~FF_TRANSMASK
				thok.scale = $/2
				thok.color = SKINCOLOR_WHITE
				P_TeleportMove(thok, thok.x, thok.y, thok.z + FixedMul(16*FRACUNIT, thok.scale))
				
				-- propel it towards the next waypoint
				local dist = max(1, P_AproxDistance(P_AproxDistance(jail.x - mo.x, jail.y - mo.y), jail.z - mo.z))
				local speed = FixedMul(64*FRACUNIT, thok.scale)
				
				thok.momx = FixedMul(FixedDiv(jail.x - mo.x, dist), speed)
				thok.momy = FixedMul(FixedDiv(jail.y - mo.y, dist), speed)
				thok.momz = FixedMul(FixedDiv(jail.z - mo.z, dist), speed)
				thok.tics = (dist/FRACUNIT) / 30
			end
		elseif (p.bp.hr_jail and not mo.health)	-- ...guess you're going in a new jail!
			-- remove current jail info
			for k, v in pairs(p.bp.hr_jail)
				if (v != p) then continue end
				
				-- if we're here, we found ourself in the table
				table.remove(p.bp.hr_jail, k)
				break
			end
			p.bp.hr_jail = nil
		end
	else
		p.bp.hr_jail = nil	-- make sure this is reset
		if ((mo and mo.valid) and not mo.health)	-- are we dead? we should get caught!
			p.bp.hr_caught = true
		end
	end
end

hooliganroundup.Func_PlayerCops = function(p)
	-- we'll need this
	local mo = p.mo
	
	-- not a cop?
	if (p.bp.team != 2)
		if (mo and mo.valid)
		and (mo.bp_chopper and mo.bp_chopper.valid)
			P_RemoveMobj(mo.bp_chopper)
			mo.bp_chopper = nil
		end
		
		p.bp.hr_chopperdelay = nil
		p.bp.hr_sirentimer = nil
		p.bp.hr_sirensounded = nil
		return 
	end	
	local c = {battleplus.cv_t1c.value, battleplus.cv_t2c.value}
	
	-- introduce a delay before respawning a chopper
	if (p.kartstuff[k_spinouttimer])
	or (p.kartstuff[k_squishedtimer])
	or (p.kartstuff[k_respawn])
		p.bp.hr_chopperdelay = TICRATE*3/2
	else
		p.bp.hr_chopperdelay = $ and $-1
	end
	
	-- we need a mo past this point
	if not (mo and mo.valid) then return end
	
	-- what we'll catch hooligans with
	if not (mo.bp_chopper and mo.bp_chopper.valid)
	and (leveltime >= 130)
	and (mo.health)	-- not dying
	and not (p.bp.hr_chopperdelay)	-- not on a delay
		-- spawn in front of the player
		local newx = mo.x + FixedMul(mo.scale, 64*cos(mo.angle))
		local newy = mo.y + FixedMul(mo.scale, 64*sin(mo.angle))
		
		local chopper = P_SpawnMobj(newx, newy, mo.z, MT_BP_HR_CHOPPER)
		chopper.target = mo
		chopper.color = c[p.bp.team]
		chopper.tics = -1
		
		-- fancy effect
		local pop = P_SpawnMobj(chopper.x, chopper.y, chopper.z, MT_THOK)
		pop.state = S_XPLD1
		pop.scale = chopper.scale/2
		S_StartSound(chopper, sfx_pop)
		
		-- link it to the player so we know we have one
		mo.bp_chopper = chopper
	end
	
	-- start a siren if we're close to a hooligan!
	local starttime = 6*TICRATE + (3*TICRATE/4)
	if (leveltime > starttime)
		local closeby
		if (mo.bp_chopper and mo.bp_chopper.valid)	-- we need something to catch them with!
			for p2 in players.iterate
				local mo2 = p2.mo
				if not ((mo2 and mo2.valid) and mo2.health) then continue end
				if (p2.bp.team != 1) then continue end
				if (p2.bp.hr_caught) then continue end
				
				if (P_AproxDistance(P_AproxDistance(mo2.x - mo.x, mo2.y - mo.y), mo2.z - mo.z) <= FixedMul(mo.scale, 1536*FRACUNIT))
				and (P_CheckSight(mo, mo2))
					closeby = true
					break
				end
			end
		end
		
		p.bp.hr_sirentimer = $ or 0
		if (closeby)
			if not (p.bp.hr_sirensounded)
				S_StartSound(mo, sfx_s3k86)
				p.bp.hr_sirensounded = true
			else
				if not (S_SoundPlaying(mo, sfx_s3k86))
					if not (p.bp.hr_sirentimer%10)
						S_StartSoundAtVolume(mo, sfx_kc3e, 128)
					end
					p.bp.hr_sirentimer = $+1
				end
			end
		else
			p.bp.hr_sirentimer = 0
			p.bp.hr_sirensounded = nil
		end
	end
end

hooliganroundup.Func = function()
	hooliganroundup.Func_Gamestate()
	hooliganroundup.Func_HandleJails()
	for p in players.iterate
		hooliganroundup.Func_PlayerHooligans(p)
		hooliganroundup.Func_PlayerCops(p)
	end
end

hooliganroundup.FinishHook = function()
	local teams = {server.bp_server.teams[1], server.bp_server.teams[2]}
	
	-- additional points for the hooligans
	battleplus.TallyTeams(true, true)
		
	local points = (server.bp_server.teamtally[2]/#teams[2])
	for k, v in ipairs(teams[1])
		v.marescore = $+points
		server.bp_server.teamtally[1] = $+points
	end
	local c = {battleplus.ColorToMap(battleplus.cv_t1c.value), battleplus.ColorToString(battleplus.cv_t1c.value)}
	local points_string = (points == 1) and "point" or "points"
	print("The "..(c[1])..(c[2]).." Team".."\x80".." received "..points.." "..points_string.." each for winning!")
end

hooliganroundup.HUDFunc_Hooligans = function(v)
	if not (#server.bp_server.teams[1]) then return end	-- nobody to display?
	
	-- calculate width first (yes I know this sucks!)
	local total_width = 6	-- left border
	for index, p in pairs(server.bp_server.teams[1])
		if not (p and p.valid) then continue end
		if not (p.mo and p.mo.valid) then continue end
		
		local pp = v.cachePatch(skins[p.mo.skin].facemmap)
		total_width = $+pp.width
		
		-- space it out
		if (index != #server.bp_server.teams[1])
			total_width = $+1
		end
	end
	total_width = $+6	-- right border
	
	-- now ACTUALLY draw it
	local hx, hy, flags = 241, (splitscreen) and 7 or 13, V_SNAPTOTOP|V_SNAPTORIGHT|V_HUDTRANS
	if (splitscreen > 1)
		hx, hy, flags = 160, 7, V_SNAPTOTOP|V_HUDTRANS
	end
	hx = $ - (total_width/2)	-- centre it
	
	-- offset hy depending on if timelimit is drawn
	local gamemode = battleplus.gamemodes[server.bp_server.gamemode]
	if (not (splitscreen) and hud.enabled("time"))	-- timelimit is drawn (unless in splitscreen, timelimit isnt drawn there)
	or ((gamemode.rules & BP_GMR_TIMELIMIT) and server.bp_server.timelimit)
		hy = $+21
	end
	
	-- left border
	v.draw(hx, hy, v.cachePatch("K_STJAIL"), flags)
	hx = $+6
	
	-- players
	v.drawFill(hx, hy, total_width - 12, 6, 23|flags)	-- thankfully this is solid colour
	for index, p in pairs(server.bp_server.teams[1])
		if not (p and p.valid) then continue end
		if not (p.mo and p.mo.valid) then continue end
		
		local pp = v.cachePatch(skins[p.mo.skin].facemmap)
		local skin = (p.bp.hr_caught) and TC_RAINBOW or -1
		local colour = (p.bp.hr_caught) and SKINCOLOR_GREY or battleplus.cv_t1c.value
		v.draw(hx + pp.leftoffset, hy + (12 - pp.height + pp.topoffset) - 4, pp, flags, v.getColormap(skin, colour))
		hx = $+pp.width
		
		if (index != #server.bp_server.teams[1])
			hx = $+1
		end
	end
	
	-- right border
	v.draw(hx, hy, v.cachePatch("K_STJAIR"), flags)
end

hooliganroundup.HUDFunc_Jails = function(v)
	if not (server.bp_server.hr_jails) then return end
	
	for k, jail in ipairs(server.bp_server.hr_jails)
		if not (jail and jail.valid) then continue end	-- jail fucking died, what???
		
		local caught_players = min(9, #jail.bp_players)	-- anxiety
		local num_patch = "STCFN0"..string.byte(tostring(caught_players))
		v.drawOnMinimap(jail.x, jail.y, FRACUNIT*4/5, v.cachePatch("JAILMMAP"), v.getColormap(-1, battleplus.cv_t2c.value))
		v.drawOnMinimap(jail.x, jail.y, FRACUNIT, v.cachePatch(num_patch))
	end
end

hooliganroundup.HUDFunc = function(v, p, c)
	if (battleplus.IsDisplayPlayer(p) != 0) then return end	-- reduce rendering as much as possible
	if (p.exiting) then return end	-- not after we've finished
	
	hooliganroundup.HUDFunc_Hooligans(v)
	hooliganroundup.HUDFunc_Jails(v)
end

hooliganroundup.ItemOddsModifier = function(p, item, newodds)
	if (item == KITEM_HYUDORO) then return 0 end	-- no.
	
	local item_found = false
	local items = {KITEM_INVINCIBILITY, KITEM_GROW}
	for i = 1, #items
		if (item != items[i]) then continue end
		
		-- we found a item to modify
		item_found = true
		break
	end
	if not (item_found) then return newodds end	-- just return what we already had
	
	-- get how many are in the gamestate
	local caught, total = 0, 0
	for p in players.iterate
		if (p.bp.team != 1) then continue end
		
		total = $+1
		if not (p.bp.hr_caught) then continue end
		caught = $+1
	end
	
	-- remove the item under this
	if (p.bp.team == 2)	-- cops cant have this
	or (p.bp.team == 1 and (caught != (total - 1)))	-- we need one hooligan left
		return 0
	end
	
	-- we can have the item now
	return newodds * 2
end

local hooliganroundup_info = {
	name = "Hooligan Roundup",
	objective = "You shouldn't be seeing this.",
	teamobjectives = {"Escape the law together!", "Capture all the Hooligans!"},
	teamicons = {"K_HOLRN1", "K_HOLRN2"},
	func = hooliganroundup.Func,
	finishhook = hooliganroundup.FinishHook,
	hudfunc = hooliganroundup.HUDFunc,
	rules = BP_GMR_FORCETEAMS|BP_GMR_ELIMINATION|BP_GMR_REMOVEWANTED|BP_GMR_TIMELIMIT|BP_GMR_HIDETEAMSCORES,
	timelimit = hooliganroundup.cv_timelimit,
	itemoddsmodifier = hooliganroundup.ItemOddsModifier,
}
battleplus.AddGamemode(BP_GM_HOOLIGANROUNDUP, hooliganroundup_info)