local itemrain = battleplus.itemrain

-- handle items
itemrain.Func_ItemHandler = function()
	if not (server.bp_server.ir_itemlist)
		server.bp_server.ir_itemlist = {}
		if not (#server.bp_server.ir_itemlist) then return end
	end
	
	local i = #server.bp_server.ir_itemlist
	while i
		local item = server.bp_server.ir_itemlist[i]
			
		-- check if item is still valid
		if not (item and item.valid) or not item.health
			table.remove(server.bp_server.ir_itemlist, i)
			i = $-1
			continue
		end
		
		-- check if they've got a fuse
		if (item.ir_fuse)
			item.ir_fuse = $-1
			-- flash
			if (item.ir_fuse < 4*TICRATE)
				if (leveltime%2)
					item.flags2 = $|MF2_DONTDRAW
				else
					item.flags2 = $ & ~MF2_DONTDRAW
				end
			end
			-- remove them when the fuse runs out
			if not (item.ir_fuse)
				P_KillMobj(item) 
				table.remove(server.bp_server.ir_itemlist, i)
				i = $-1
				continue
			end
		else
			-- spawn additional ghosts while falling
			if (item.type == MT_ORBINAUT or item.type == MT_JAWZ_DUD or item.type == MT_FLOATINGITEM) and (leveltime%2) then P_SpawnGhostMobj(item) end
		end
			
		-- give them a fuse when they land
		if (P_IsObjectOnGround(item) and not item.ir_fuse)
			-- they're on the floor so we start the fuse
			item.ir_fuse = 15*TICRATE
			item.momx, item.momy = 0, 0
			i = $-1
			continue
		end
		
		-- go to next item in the table
		i = $-1
	end
end

-- raining functions
itemrain.Func_RainHazards = function()
	-- spawn hazards
	local dropped_mine = false	-- used so we dont spawn more than 1 mine
	for i = 1, P_RandomRange(1, 3)
		-- find valid targets to drop shit on
		local possiblePlayers = {}
		for p in players.iterate
			if (p and p.mo) and not p.spectator then table.insert(possiblePlayers, p.mo) end
		end
		
		if (#possiblePlayers) -- we have atleast 1 target
			-- our target player
			local tmo = possiblePlayers[P_RandomRange(1, #possiblePlayers)]
			-- possible hazards to spawn
			local possibleHazards = {MT_BANANA, MT_EGGMANITEM, MT_ORBINAUT, MT_JAWZ_DUD}
					
			-- the actual hazard
			local starttime = 6*TICRATE + (3*TICRATE/4)
			local sink = P_RandomRange(1, max(4, 16-((leveltime-starttime-(TICRATE/2))/(15*TICRATE)))) == 1
			local mine = (P_RandomRange(1, 5) == 1 and not dropped_mine)
					
			local hazard = P_SpawnMobj(tmo.x+P_RandomRange(-1250, 1250)<<FRACBITS, tmo.y+P_RandomRange(-1250, 1250)<<FRACBITS, tmo.ceilingz-P_RandomRange(0, 150)<<FRACBITS, (spawnsink) and MT_SINK or ((mine) and MT_SSMINE_SHIELD or possibleHazards[P_RandomRange(1, #possibleHazards)]))
			hazard.state = (hazard.type == MT_ORBINAUT or hazard.type == MT_JAWZ_DUD) and (hazard.type == MT_ORBINAUT and S_ORBINAUT_DEAD or S_JAWZ_DEAD1) or $
			hazard.color = hazard.type == MT_ORBINAUT and 3 or $
			hazard.flags2 = (hazard.type == MT_ORBINAUT or hazard.type == MT_JAWZ_DUD) and $|MF2_AMBUSH or $
			hazard.tics = (hazard.type == MT_ORBINAUT or hazard.type == MT_JAWZ_DUD) and -1 or $
			hazard.extravalue1 = (hazard.type == MT_SSMINE_SHIELD) and 49 or $
			hazard.momx, hazard.momy = P_RandomRange(-8, 8)*hazard.scale, P_RandomRange(-8, 8)*hazard.scale
			hazard.momz = P_MobjFlip(hazard)*(P_RandomRange(0, 8)*hazard.scale)
			server.bp_server.ir_itemlist[#server.bp_server.ir_itemlist+1] = hazard -- just so they dont start sliding without slopes
						
			-- pop into existence with style
			local pop = P_SpawnMobj(hazard.x, hazard.y, hazard.z, MT_THOK)
			pop.state = S_XPLD1
					
			-- make sure more than 1 mine per set doesnt drop
			if (hazard.type == MT_SSMINE_SHIELD) then dropped_mine = true end
		end
	end
end
itemrain.Func_RainItems = function()
	-- spawn items
	for i = 1, P_RandomRange(1, 3)
		-- find valid targets to drop shit on
		local possiblePlayers = {}
		for p in players.iterate
			if (p and p.mo) and not p.spectator then table.insert(possiblePlayers, p.mo) end
		end
		
		if (#possiblePlayers) -- we have atleast 1 target
			-- our target player
			local tmo = possiblePlayers[P_RandomRange(1, #possiblePlayers)]
			-- possible items to spawn
			local possibleItems = {KITEM_SNEAKER, KITEM_BANANA, KITEM_EGGMAN, KITEM_ORBINAUT, KITEM_JAWZ, KITEM_MINE, KITEM_BALLHOG, KITEM_POGOSPRING}
					
			-- the actual item
			local item = P_SpawnMobj(tmo.x+P_RandomRange(-1500, 1500)<<FRACBITS, tmo.y+P_RandomRange(-1500, 1500)<<FRACBITS, tmo.ceilingz-P_RandomRange(0, 250)<<FRACBITS, MT_FLOATINGITEM)
			item.scale = $/4
			item.destscale = 3*(mapobjectscale)/2
			item.threshold = (P_RandomRange(1, 64) == 1) and KITEM_KITCHENSINK or possibleItems[P_RandomRange(1, #possibleItems)]
			item.movecount = 1
			item.momx, item.momy = P_RandomRange(-8, 8)*item.scale, P_RandomRange(-8, 8)*item.scale
			item.momz = P_MobjFlip(item)*(P_RandomRange(0, 8)*item.scale)
			item.flags = $|MF_NOCLIPTHING
			server.bp_server.ir_itemlist[#server.bp_server.ir_itemlist+1] = item -- just so they dont start sliding without slopes
						
			-- pop into existence with style
			local pop = P_SpawnMobj(item.x, item.y, item.z, MT_THOK)
			pop.state = S_XPLD1
		end
	end
end

itemrain.Func = function()
	itemrain.Func_ItemHandler()
	-- rain items
	local starttime = 6*TICRATE + (3*TICRATE/4)
	
	if (leveltime > starttime+(TICRATE/2)) -- only after the music starts
	and not (leveltime%35)
		-- increase player score
		for p in players.iterate
			if (p and p.mo) and not p.spectator
			and p.kartstuff[k_bumper]
				P_AddPlayerScore(p, 1)
			end
		end
		-- now it's rain time
		S_StartSound(nil, sfx_s3k51)
		itemrain.Func_RainHazards()
		itemrain.Func_RainItems()
	end
end

local itemrain_info = {
	name = "Item Rain",
	objective = "Survive! ...somehow.",
	icon = "K_ITRAIN",
	func = itemrain.Func,
	rules = BP_GMR_HELL|BP_GMR_BUMPERS,
	bumpers = itemrain.cv_bumpers,
}
battleplus.AddGamemode(BP_GM_ITEMRAIN, itemrain_info)