local randomizer = battleplus.randomizer
local dprint = battleplus.dprint

randomizer.Func = function()
	for p in players.iterate
		-- randomize stats
		p.bp.rn_timer = $ and min(TICRATE, $+1) or 1
		p.bp.rn_kartspeed = $ or P_RandomRange(1, 9)
		p.bp.rn_kartweight = $ or P_RandomRange(1, 9)
		-- set speed & weight
		p.kartspeed = p.bp.rn_kartspeed
		p.kartweight = p.bp.rn_kartweight
		
		-- randomize items
		p.bp.rn_needitem = $ and $-1
		if p.kartstuff[k_itemroulette] then p.bp.rn_needitem = 5 end
		
		if p.kartstuff[k_itemtype] -- player has an item
		and p.bp.rn_needitem
			if not p.bp.rn_randomizeditem
				local possibleItems = {{KITEM_SNEAKER, {1, 3}}, {KITEM_ROCKETSNEAKER}, {KITEM_INVINCIBILITY}, {KITEM_BANANA, {1, 3, 10}}, {KITEM_EGGMAN}, {KITEM_ORBINAUT, {1, 3, 4}}, {KITEM_JAWZ, {1, 2}}, {KITEM_MINE}, {KITEM_BALLHOG}, {KITEM_SPB}, {KITEM_GROW}, {KITEM_SHRINK}, {KITEM_THUNDERSHIELD}, {KITEM_HYUDORO}, {KITEM_POGOSPRING}, {KITEM_KITCHENSINK}}
				local item = P_RandomRange(1, #possibleItems)
				
				p.kartstuff[k_itemtype] = possibleItems[item][1]
				p.kartstuff[k_itemamount] = possibleItems[item][2] and possibleItems[item][2][P_RandomRange(1, #possibleItems[item][2])] or 1
				p.bp.rn_randomizeditem = true
				p.bp.rn_needitem = nil
			end
		else
			p.bp.rn_randomizeditem = nil
		end
	end
end

randomizer.HUDFunc = function(v, p, c)
	if (p.exiting or p.spectator) then return end
	if (p.kartstuff[k_comebacktimer] and not p.kartstuff[k_bumper]) then return end
	if not (p.bp.rn_timer and p.bp.rn_timer%2) then return end
	
	local dp = battleplus.IsDisplayPlayer(p)
	
	if (splitscreen) > 1
		local hx = (dp == 1 or dp == 3) and 194 or 3
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_SNAPTORIGHT|V_HUDTRANS, dp)
					
		v.drawString(hx, 64, "Speed: "..p.bp.rn_kartspeed.."\nWeight: "..p.bp.rn_kartweight, flags, (dp == 1 or dp == 3) and "thin-right" or "thin")
	else
		local hy = (splitscreen) and 50 or 145
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_HUDTRANS, dp)
		v.drawString(10, hy, "Speed: "..p.bp.rn_kartspeed.."\nWeight: "..p.bp.rn_kartweight, flags, "left")
	end
end

randomizer.Unload = function()
	for p in players.iterate
		if (p.mo and p.mo.valid)
			p.kartspeed = skins[p.mo.skin].kartspeed
			p.kartweight = skins[p.mo.skin].kartweight
			dprint(4, "Reset "..p.name.."'s stats to default")
		end
	end
end

local randomizer_info = {
	name = "Randomizer",
	teamname = "Team Randomizer",
	objective = "Last one standing wins!",
	teamobjective = "Last team standing wins!",
	icon = "K_RDMIZR",
	func = randomizer.Func,
	hudfunc = randomizer.HUDFunc,
	unload = randomizer.Unload,
	rules = BP_GMR_TEAMS|BP_GMR_HELL|BP_GMR_TIMELIMIT|BP_GMR_BUMPERS,
	timelimit = randomizer.cv_timelimit,
	bumpers = randomizer.cv_bumpers,
}
battleplus.AddGamemode(BP_GM_RANDOMIZER, randomizer_info)