battleplus.ringrunners = {}
local ringrunners = battleplus.ringrunners

-- sounds
freeslot("sfx_rrring")

-- ring spawnpoint
ringrunners.spawnpoint_stopspawningtime = 5*TICRATE
ringrunners.spawnpoint_spawngracetime = 3*TICRATE/2

freeslot("S_BP_RR_SPAWNPOINT", "MT_BP_RR_SPAWNPOINT")
states[S_BP_RR_SPAWNPOINT] = {SPR_NULL, A, (ringrunners.spawnpoint_stopspawningtime), nil, 0, 0, S_NULL}	-- tics are the default beacon time
mobjinfo[MT_BP_RR_SPAWNPOINT] = {
	spawnstate = S_BP_RR_SPAWNPOINT,
	deathstate = S_NULL,
	height = 16*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT,
}

addHook("MobjThinker", function(mo)
	-- spawn thoks like a beacon
	if not (mo.tics%3)
		local thok = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		thok.frame = ($|FF_FULLBRIGHT) & ~FF_TRANSMASK	-- give fullbright, remove transparency
		thok.color = SKINCOLOR_WHITE
		thok.flags = $|(MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT)
		thok.tics = TICRATE
		
		thok.scale = $*2
		thok.momz = 60*thok.scale
		thok.destscale = 1
		thok.scalespeed = FRACUNIT/18
	end
	mo.bp_spawngrace = $ and min((ringrunners.spawnpoint_spawngracetime), $+1) or 1
	
	if not (mo.bp_spawngrace >= (ringrunners.spawnpoint_spawngracetime))
	or (mo.tics <= (ringrunners.spawnpoint_stopspawningtime))
		-- don't spawn rings
		return
	end
	
	if not ((mo.tics+1)%TICRATE)
		-- spawn rings every second
		ringrunners.RingBurst(mo, 7, {-6, 6}, {-6, 6}, {14, 19})
		S_StartSound(mo, sfx_rrring)
	end
end, MT_BP_RR_SPAWNPOINT)

-- ring hitbox
freeslot("MT_BP_RR_RINGHITBOX")
mobjinfo[MT_BP_RR_RINGHITBOX] = {
	spawnstate = S_INVISIBLE,
	radius = 128*FRACUNIT,
	height = 128*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_SPECIAL
}

-- ring
freeslot("S_BP_RR_RING", "MT_BP_RR_RING", "SPR_BPRI", "SPR_BPRF")
states[S_BP_RR_RING] = {SPR_BPRI, A, (13*TICRATE)+1, nil, 23, 1, S_BP_RR_RING}
mobjinfo[MT_BP_RR_RING] = {
    spawnstate = S_BP_RR_RING,
	deathstate = S_SPRK1,
    radius = 16*FRACUNIT,
    height = 24*FRACUNIT,
    flags = MF_BOUNCE
}

-- dragged rings
freeslot("S_BP_RR_CHASERING", "MT_BP_RR_CHASERING")
states[S_BP_RR_CHASERING] = {SPR_BPRF, A|FF_ANIMATE, -1, nil, 12, 1, S_BP_RR_CHASERING}
mobjinfo[MT_BP_RR_CHASERING] = {
    spawnstate = S_BP_RR_CHASERING,
	deathstate = S_SPRK1,
	deathsound = sfx_itemup,
    radius = 16*FRACUNIT,
    height = 24*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_SPECIAL
}

-- ring handler
addHook("MobjSpawn", function(ring) 
	if not (ring and ring.valid) then return end
	if not G_BattleGametype() then return end
	if not battleplus.IsGamemode(BP_GM_RINGRUNNERS) then return end	-- not ring runners

	-- add mo to the table of stuff to do
	server.bp_server.rr_rings = $ or {}	-- initialize
	server.bp_server.rr_rings[#server.bp_server.rr_rings+1] = ring
end, MT_BP_RR_RING)
ringrunners.HandleRings = function()
	server.bp_server.rr_rings = $ or {}
	if not (#server.bp_server.rr_rings) then return end
	
	local i = #server.bp_server.rr_rings
	while i
		local ring = server.bp_server.rr_rings[i]
		
		-- check if ring is still valid
		if not (ring and ring.valid) or not ring.health
			table.remove(server.bp_server.rr_rings, i)
			i = $-1
			continue
		end
		
		if (P_CheckDeathPitCollide(ring))
			P_RemoveMobj(ring)
			table.remove(server.bp_server.rr_rings, i)
			i = $-1
			continue 
		end
		
		-- make the ring flash before being killed
		if (ring.tics < 4*TICRATE)
			-- make the ring flash!
			ring.flags2 = (leveltime%2) and $|MF2_DONTDRAW or $ & ~(MF2_DONTDRAW)
		end
		
		-- kill the ring at 1 tic left
		if (ring.tics == 1)
			ring.flags2 = $ & ~(MF2_DONTDRAW)
			P_KillMobj(ring)
			table.remove(server.bp_server.rr_rings, i)
			i = $-1
			continue
		end
		
		-- lower threshold, instant grabs shouldnt be possible this way
		ring.threshold = $ and $-1
		
		-- spawn a hitbox for the ring so we don't have to use a player iterate
		if not (ring.bp_hitbox and ring.bp_hitbox.valid)
			local hitbox = P_SpawnMobj(ring.x, ring.y, ring.z, MT_BP_RR_RINGHITBOX)
			hitbox.target = ring
			hitbox.scale = ring.scale
			P_TeleportMove(hitbox, ring.x, ring.y, ring.z - (hitbox.height/2))
			
			-- link it with the ring
			ring.bp_hitbox = hitbox
		else
			local hitbox = ring.bp_hitbox
			hitbox.target = ring
			hitbox.scale = ring.scale
			P_TeleportMove(hitbox, ring.x, ring.y, ring.z - (hitbox.height/2))
		end
		
		-- set the ring's flags
		ring.flags = $ & ~(MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT)
		
		-- speed stuff
		local speed = FixedHypot(ring.momz, FixedHypot(ring.momx, ring.momy))
		
		-- the ring's frame
		local frames = 23
		local frames_add = (speed > 6*ring.scale) and 2 or 1
		
		ring.frame = $ + frames_add
		while (ring.frame > frames)
			ring.frame = $ - frames
		end
			
		local oldmomz = ring.bp_oldmomz or ring.momz	-- fallback
		if (P_IsObjectOnGround(ring))
		and (abs(oldmomz) > 3*ring.scale)
			-- bounce off the floor
			ring.momx = FixedMul(FRACUNIT*6/9, $)
			ring.momy = FixedMul(FRACUNIT*6/9, $)
			ring.momz = FixedMul(FRACUNIT*6/9, -oldmomz)
		end
		ring.bp_oldmomz = ring.momz
		//print(FixedHypot(ring.momz, FixedHypot(ring.momx, ring.momy)) / ring.scale)
		
		-- go to next ring in the table
		i = $-1
	end
end

addHook("TouchSpecial", function(hitbox, mo)
	if not (hitbox.target and hitbox.target.valid) then return end
	local ring = hitbox.target
	
	if (ring.threshold) then return true end
	if not (P_CheckSight(ring, mo)) then return true end
	
	-- dont be collected by hurt players
	if mo.player.kartstuff[k_spinouttimer] 
	or mo.player.kartstuff[k_squishedtimer]
		return true
	end
	
	-- spawn a fake ring that chases the player
	local cring = P_SpawnMobj(ring.x, ring.y, ring.z, MT_BP_RR_CHASERING)
	cring.target = mo
	P_RemoveMobj(ring)	-- remove this
end, MT_BP_RR_RINGHITBOX)

-- chase ring handler
addHook("MobjSpawn", function(cring) 
	if not (cring and cring.valid) then return end
	if not G_BattleGametype() then return end
	if not battleplus.IsGamemode(BP_GM_RINGRUNNERS) then return end	-- not ring runners

	-- add mo to the table of stuff to do
	server.bp_server.rr_chaserings = $ or {}	-- initialize
	server.bp_server.rr_chaserings[#server.bp_server.rr_chaserings+1] = cring
end, MT_BP_RR_CHASERING)
ringrunners.HandleChaseRings = function()
	server.bp_server.rr_chaserings = $ or {}
	if not (#server.bp_server.rr_chaserings) then return end
	
	local i = #server.bp_server.rr_chaserings
	while i
		local cring = server.bp_server.rr_chaserings[i]
		
		-- check if ring is still valid
		if not (cring and cring.valid) or not cring.health
			table.remove(server.bp_server.rr_chaserings, i)
			i = $-1
			continue
		end
		
		-- do we still have a target?
		if not (cring.target and cring.target.valid)
			-- spawn a new ring
			local ring = P_SpawnMobj(cring.x, cring.y, cring.z, MT_BP_RR_RING)
			ring.momx, ring.momy, ring.momz = cring.momx, cring.momy, cring.momz
			ring.threshold = 10
	
			-- remove ring
			P_RemoveMobj(cring)
			table.remove(server.bp_server.rr_chaserings, i)
			i = $-1
			continue
		end
		
		-- handle drag
		cring.bp_dragtimer = $ and min(30, $+1) or 1
		A_HomingChase(cring, (7*FRACUNIT)*cring.bp_dragtimer)
		
		-- go to next ring in the table
		i = $-1
	end
end
addHook("TouchSpecial", function(cring, mo)
	if (cring.target != mo) then return true end
	
	-- give the player rings
	ringrunners.GiveRings(mo.player, 1)
	mo.player.kartstuff[k_wanted] = $-(TICRATE/2)
	cring.momx, cring.momy, cring.momz = 0, 0, 0
end, MT_BP_RR_CHASERING)

-- ring debt overlay
freeslot("MT_BP_RR_DEBT", "SPR_BPDB")
local debtTics = {5, 1, 3, 1, 3, 1, 3}
for i = 1, 7
	freeslot("S_BP_RR_DEBT"..i)
	states[S_BP_RR_DEBT1+(i-1)] = {SPR_BPDB, FF_FULLBRIGHT|0+(i-1), debtTics[i], nil, 0, 0, i<7 and S_BP_RR_DEBT1+i or S_BP_RR_DEBT1}
end
mobjinfo[MT_BP_RR_DEBT] = {
	spawnstate = S_BP_RR_DEBT1,
	deathstate = S_NULL,
	radius = 36*FRACUNIT,
	height = 37*FRACUNIT,
	dispoffset = -2,
	flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY|MF_SCENERY|MF_DONTENCOREMAP
}
addHook("MobjThinker", function(debt)
	if (debt and debt.valid)
		-- now for remove situations
		if not (debt.target and debt.target.player) then P_RemoveMobj(debt) return end -- they dont have a target, or it's not a player
		if not (debt.target.player.health) then P_RemoveMobj(debt) return end -- target is dead
		if not (ringrunners.IsPlayerInDebt(debt.target.player)) then P_RemoveMobj(debt) return end -- target isnt in debt, so why should this be displayed?

		local s = 3*debt.target.scale -- base scale

		-- don't draw if the target isnt drawn
		if (debt.target.flags2 & MF2_DONTDRAW)
			debt.flags2 = $|MF2_DONTDRAW
		else
			debt.flags2 = $ & ~(MF2_DONTDRAW)
		end

		-- set scale
		if not (splitscreen)
			if ((displayplayers[0] and displayplayers[0].valid) and (displayplayers[0].mo and displayplayers[0].mo.valid))
				s = debt.target.scale + FixedMul(FixedDiv(abs(P_AproxDistance(displayplayers[0].mo.x - debt.target.x, displayplayers[0].mo.y - debt.target.y)), 1280<<FRACBITS), debt.target.scale)
				if (s > 16*debt.target.scale)
					s = 16*debt.target.scale
				end
			end
		end
		debt.destscale = s
		
		-- hide if this guy ain't our rival (with splitscreen support, BLESS 1.2)
		local drawFlags = {MFE_DRAWONLYFORP1, MFE_DRAWONLYFORP2, MFE_DRAWONLYFORP3, MFE_DRAWONLYFORP4}
		for i = 0, (splitscreen) -- go through displayplayers
			if not displayplayers[i] then break end
			if (debt.target.player == displayplayers[i]) then debt.eflags = $ & ~(drawFlags[i+1]) continue end -- dont show yourself
					
			if not K_IsPlayerWanted(debt.target.player)
				debt.eflags = $|drawFlags[i+1]
			else
				debt.eflags = $ & ~(drawFlags[i+1])
			end
		end
		-- somehow this isnt supposed to be displayed at ALL for any displayplayers, so dont draw it
		if not (debt.eflags & (MFE_DRAWONLYFORP1|MFE_DRAWONLYFORP2|MFE_DRAWONLYFORP3|MFE_DRAWONLYFORP4))
			debt.flags2 = $|MF2_DONTDRAW
		end
		
		-- set position (flip support too)
		if (debt.target.eflags & MFE_VERTICALFLIP)
			P_TeleportMove(debt, debt.target.x, debt.target.y, debt.target.z - P_GetPlayerHeight(debt.target.player) - (16*debt.target.scale) - (64*s))
			debt.eflags = $|MFE_VERTICALFLIP
		else
			P_TeleportMove(debt, debt.target.x, debt.target.y, debt.target.z + P_GetPlayerHeight(debt.target.player) + (16*debt.target.scale) + (64*s))
			debt.eflags = $ & ~(MFE_VERTICALFLIP)
		end
	end
end, MT_BP_RR_DEBT)

-- functions
ringrunners.GiveRings = function(p, amount)
	if (p.exiting) then return end	-- dont give rings after the round!
	p.bp.rr_rings = max(-99, min(999, $+amount))
	p.bp.rr_hudflash = (amount < 0) and -3 or 3
end
ringrunners.RingBurst = function(mo, amount, momx, momy, momz)
	-- momx, momy and momz are table values
	-- {lowerbound, upperbound}
	-- doesnt actually make players lose rings, use in conjunction with ringrunners.GiveRings
	for i = 1, amount
		local ring = P_SpawnMobj(mo.x, mo.y, mo.z+(mo.height/2), MT_BP_RR_RING)
		ring.momx = P_RandomRange(momx[1], momx[2])*ring.scale
		ring.momy = P_RandomRange(momy[1], momy[2])*ring.scale
		ring.momz = P_RandomRange(momz[1], momz[2])*ring.scale
		ring.threshold = 10
	end
end
ringrunners.IsPlayerInDebt = function(p)
	return (p.bp.rr_rings < 0)
end
ringrunners.CalculateRingsOnPlayer = function(p, amount)
	if K_IsPlayerWanted(p) then amount = $*2 end	-- double if wanted
	if ringrunners.IsPlayerInDebt(p) then amount = $/2 end	-- half if in debt
	
	return amount
end
ringrunners.SpawnRingSpawnpoint = function(x, y, z, ringsets)
	local spawnpoint = P_SpawnMobj(x, y, z, MT_BP_RR_SPAWNPOINT)
	spawnpoint.tics = $+((ringsets*TICRATE)+1)+(ringrunners.spawnpoint_spawngracetime)
	
	return spawnpoint
end