battleplus.ringslinger = {}
local ringslinger = battleplus.ringslinger

-- weapon ring constants
local ringslinger_flags = {
	"BP_RW_AUTOMATIC",
	"BP_RW_BOUNCE",
	"BP_RW_SCATTER",
	"BP_RW_GRENADE",
	"BP_RW_EXPLOSION",
	"BP_RW_RAIL",
}
for k, v in ipairs(ringslinger_flags)
    rawset(_G, v, k)
end

-- define the "panel" we'll be using for all our black magic
freeslot("S_BP_RS_WEAPONPICKUP", "MT_BP_RS_WEAPONPICKUP")	-- state/object
freeslot("SPR_PIKA", "SPR_PIKB", "SPR_PIKS", "SPR_PIKG", "SPR_PIKE", "SPR_PIKR")	-- sprites
states[S_BP_RS_WEAPONPICKUP] = {SPR_PIKA, A, -1, nil, 0, 0, S_BP_RS_WEAPONPICKUP}

-- fade states...
local weapons = {"AUTOMATIC", "BOUNCE", "SCATTER", "GRENADE", "EXPLOSION", "RAIL"}
local sprites = {SPR_PIKA, SPR_PIKB, SPR_PIKS, SPR_PIKG, SPR_PIKE, SPR_PIKR}

for i = 1, #weapons
	local statename = "S_BP_RS_WEAPONPICKUP_"..weapons[i].."FADE"
	freeslot(statename)
	states[_G[statename]] = {sprites[i], A, 2*TICRATE, nil, 0, 0, S_NULL}
end

function A_BP_RS_SetFade(actor, var1, var2)
	actor.state = _G["S_BP_RS_WEAPONPICKUP_"..weapons[actor.rs_weaponring].."FADE"]
end
freeslot("S_BP_RS_WEAPONPICKUP_SETFADE")	-- I literally HATE that this exists. all because I want it all in one object...
states[S_BP_RS_WEAPONPICKUP_SETFADE] = {SPR_NULL, A, 0, A_BP_RS_SetFade, 0, 0, S_NULL}	-- if this goes to nextstate idk what I'll do

mobjinfo[MT_BP_RS_WEAPONPICKUP] = {
	spawnstate = S_BP_RS_WEAPONPICKUP,
	deathstate = S_BP_RS_WEAPONPICKUP_SETFADE,
	deathsound = sfx_ncitem,
	radius = 24*FRACUNIT,
	height = 24*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_SPECIAL,
}
addHook("MobjSpawn", function(panel)
	if not (panel and panel.valid) then return end
	if not G_BattleGametype() then return end
	if not battleplus.IsGamemode(BP_GM_RINGSLINGER) then return end	-- not ringslinger

	-- add mo to the table of stuff to do
	server.bp_server.rs_animpanels = $ or {{}, {}}	-- initialize {normal, fading}
	table.insert(server.bp_server.rs_animpanels[1], panel)
end, MT_BP_RS_WEAPONPICKUP)
addHook("TouchSpecial", function(panel, mo)
	local p = mo.player
	local weaponring = panel.rs_weaponring
	
	-- dont stack weapon rings
	if (p.bp.rs_weaponring) then return true end
	
	-- add to anim table
	server.bp_server.rs_animpanels = $ or {{}, {}}	-- initialize {normal, fading}
	panel.rs_threshold = TICRATE	-- for ringslinger.Func_ReplacePanels()
	table.insert(server.bp_server.rs_animpanels[2], panel)
	P_SetObjectMomZ(panel, FRACUNIT)
	
	-- give the player weapon rings!
	local amounts = {6, 5, 1, 2, 1, 1}
	local amount = amounts[weaponring]
	
	-- set these for the player that got it
	p.bp.rs_weaponring = weaponring
	p.bp.rs_amount = amount
	p.bp.rs_forceringtoss = nil	-- remove this at all costs
	p.kartstuff[k_itemblink] = TICRATE	-- this always goes down, we can reuse it!
	S_StartSound(nil, sfx_itrolf, p)
	
	-- increase this to keep track of how many panels have been obtained!
	server.bp_server.rs_numgotpanels = $+1
end, MT_BP_RS_WEAPONPICKUP)

-- animate all panels
ringslinger.AnimatePanels = function()
	server.bp_server.rs_animpanels = $ or {{}, {}}	-- init
	
	-- easier table access
	local panels_normal = server.bp_server.rs_animpanels[1]
	local panels_fading = server.bp_server.rs_animpanels[2]
	
	-- normal panels
	if (#panels_normal)
		local i = #panels_normal
		while i
			local panel = panels_normal[i]
			
			-- check if ring is still valid
			if not ((panel and panel.valid) and panel.health)
				table.remove(panels_normal, i)
				i = $-1
				continue
			end
			
			-- remove from this table, it'll be in the other table! (though really, the health check above should do this, better safe than sorry!)
			if (panel.state >= S_BP_RS_WEAPONPICKUP_AUTOMATICFADE and panel.state <= S_BP_RS_WEAPONPICKUP_SETFADE)
				table.remove(panels_normal, i)
				i = $-1
				continue
			end
			
			-- set frame, and save the frame/sprite too!
			panel.frame = ((leveltime + panel.rs_leveltime)%16)
			
			-- go to next panel in the table
			i = $-1
		end
	end
	
	if (#panels_fading)
		local i = #panels_fading
		while i
			local panel = panels_fading[i]
			
			-- check if ring is still valid
			if not (panel and panel.valid)
				table.remove(panels_fading, i)
				i = $-1
				continue
			end
			
			-- set frame and fadeout opacity
			local fade = 10 - max(1, min(9, panel.tics*10/(2*TICRATE)))
			panel.frame = ((2*TICRATE - panel.tics)%8*2)
			panel.frame = ($ & ~FF_TRANSMASK)
			panel.frame = $|(fade << FF_TRANSSHIFT)
			
			-- reduce threshold
			panel.rs_threshold = $ and $-1
			
			-- go to next panel in the table
			i = $-1
		end
	end
end

-- basic function for spawning the trail of rings
local trailRings = {}
local function trailFunc(ring)
	if not (ring.momx or ring.momy) then return end
	
	if not (leveltime%(TICRATE/7))
		if (ring.type == MT_BP_RS_AUTOMATICRING)
			P_SpawnGhostMobj(ring)
		elseif (ring.type == MT_BP_RS_GRENADERING)
		or (ring.type == MT_BP_RS_EXPLOSIONRING)
			local trail = P_SpawnMobj(ring.x, ring.y, ring.z, MT_SMOKE)
			trail.scale = ring.scale
		else
			local trail = P_SpawnMobj(ring.x, ring.y, ring.z, MT_SPARK)
			trail.scale = ring.scale
		end
	end
end

-- automatic ring
freeslot("S_BP_RS_AUTOMATICRING", "MT_BP_RS_AUTOMATICRING", "SPR_RNGA")
states[S_BP_RS_AUTOMATICRING] = {SPR_RNGA, A|FF_FULLBRIGHT|FF_ANIMATE, -1, nil, 6, 1, S_BP_RS_AUTOMATICRING}
mobjinfo[MT_BP_RS_AUTOMATICRING] = {
	spawnstate = S_BP_RS_AUTOMATICRING,
	seesound = sfx_thok,
	deathstate = S_SPRK1,
	deathsound = sfx_itemup,
	speed = 60*FRACUNIT,
	radius = 16*FRACUNIT,
	height = 24*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_MISSILE|MF_NOGRAVITY,
}
table.insert(trailRings, MT_BP_RS_AUTOMATICRING)

-- bounce rings
freeslot("S_BP_RS_BOUNCERING", "MT_BP_RS_BOUNCERING", "SPR_RNGB")
states[S_BP_RS_BOUNCERING] = {SPR_RNGB, A|FF_FULLBRIGHT|FF_ANIMATE, -1, nil, 6, 1, S_BP_RS_BOUNCERING}
mobjinfo[MT_BP_RS_BOUNCERING] = {
	spawnstate = S_BP_RS_BOUNCERING,
	seesound = sfx_bnce1,
	deathstate = S_SPRK1,
	deathsound = sfx_itemup,
	speed = 60*FRACUNIT,
	radius = 16*FRACUNIT,
	height = 24*FRACUNIT,
	activesound = sfx_bnce1,
	flags = MF_NOBLOCKMAP|MF_MISSILE|MF_NOGRAVITY|MF_BOUNCE,
}
addHook("MobjThinker", function(ring)
	-- spawn the trail like other rings
	trailFunc(ring)
	
	-- just making sure the ring doesn't lose speed
	P_InstaThrust(ring, R_PointToAngle2(0, 0, ring.momx, ring.momy), ring.info.speed)
	
	-- the bounce ring's fuse
	if (ring.rs_fuse == nil) then ring.rs_fuse = 3*TICRATE end
	ring.rs_fuse = $-1
	
	-- start flashing the ring at a certain fuse
	if (ring.rs_fuse < 2*TICRATE and (leveltime%2))
		ring.flags2 = $|MF2_DONTDRAW
	else
		ring.flags2 = $ & ~(MF2_DONTDRAW)
	end
	
	-- remove when fuse dies out
	if not (ring.rs_fuse)
		P_KillMobj(ring)
	end
end, MT_BP_RS_BOUNCERING)

-- scatter ring
freeslot("S_BP_RS_SCATTERRING", "MT_BP_RS_SCATTERRING", "SPR_RNGS")
states[S_BP_RS_SCATTERRING] = {SPR_RNGS, A|FF_FULLBRIGHT, -1, nil, 0, 0, S_BP_RS_SCATTERRING}
mobjinfo[MT_BP_RS_SCATTERRING] = {
	spawnstate = S_BP_RS_SCATTERRING,
	seesound = sfx_bnce2,
	deathstate = S_SPRK1,
	deathsound = sfx_itemup,
	speed = 60*FRACUNIT,
	radius = 16*FRACUNIT,
	height = 24*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_MISSILE|MF_NOGRAVITY,
}
table.insert(trailRings, MT_BP_RS_SCATTERRING)

-- grenade rings
freeslot("S_BP_RS_GRENADERING", "MT_BP_RS_GRENADERING", "SPR_RNGG")
states[S_BP_RS_GRENADERING] = {SPR_RNGG, A|FF_FULLBRIGHT|FF_ANIMATE, -1, nil, 17, 1, S_BP_RS_GRENADERING}
mobjinfo[MT_BP_RS_GRENADERING] = {
	spawnstate = S_BP_RS_GRENADERING,
	seesound = sfx_thok,
	attacksound = sfx_gbeep,
	painchance = 192*FRACUNIT,
	deathstate = S_SSMINE_EXPLODE,	-- we might've gone a little too far
	speed = 30*FRACUNIT,
	radius = 16*FRACUNIT,
	height = 24*FRACUNIT,
	activesound = sfx_s3k5d,
	flags = MF_NOBLOCKMAP|MF_MISSILE|MF_BOUNCE|MF_GRENADEBOUNCE,
}
addHook("MobjThinker", function(ring)
	-- spawn the trail like other rings
	trailFunc(ring)
	
	-- the grenade ring's fuse
	if (ring.rs_fuse == nil) then ring.rs_fuse = 4*TICRATE end
	ring.rs_fuse = $-1
	
	-- play the beeping sound
	if not (ring.rs_fuse%TICRATE) then S_StartSound(ring, ring.info.attacksound) end
	
	-- explode the grenade ring when the fuse is up
	if not (ring.rs_fuse)
		P_KillMobj(ring)
		return
	end
end, MT_BP_RS_GRENADERING)

-- bomb rings
freeslot("S_BP_RS_EXPLOSIONRING", "MT_BP_RS_EXPLOSIONRING", "SPR_RNGE")
states[S_BP_RS_EXPLOSIONRING] = {SPR_RNGE, A|FF_FULLBRIGHT|FF_ANIMATE, -1, nil, 6, 1, S_BP_RS_EXPLOSIONRING}
mobjinfo[MT_BP_RS_EXPLOSIONRING] = {
	spawnstate = S_BP_RS_EXPLOSIONRING,
	seesound = sfx_cannon,
	painchance = 192*FRACUNIT,
	deathstate = S_SSMINE_EXPLODE,	-- we might've gone a little too far
	speed = 60*FRACUNIT,
	radius = 16*FRACUNIT,
	height = 24*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_MISSILE|MF_NOGRAVITY,
}
table.insert(trailRings, MT_BP_RS_EXPLOSIONRING)

-- rail "rings" (they're just REALLY fast regular rings)
freeslot("S_BP_RS_RAILRING", "MT_BP_RS_RAILRING")
states[S_BP_RS_RAILRING] = {SPR_NULL, A, -1, nil, 0, 0, S_BP_RS_RAILRING}
mobjinfo[MT_BP_RS_RAILRING] = {
	spawnstate = S_BP_RS_RAILRING,
	seesound = sfx_rail1,	-- too fast to play from the function, will leave here regardless
	deathstate = S_SPRK1,
	deathsound = sfx_rail2,
	speed = 60*FRACUNIT,
	radius = 16*FRACUNIT,
	height = 24*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_MISSILE|MF_NOGRAVITY,
}

-- for rings that only need their trails
for i = 1, #trailRings
	local ring = trailRings[i]
	addHook("MobjThinker", trailFunc, ring)
end

-- remove this
addHook("MobjThinker", function(arrow)
	if (leveltime < 106) then return end
	if not (battleplus.IsGamemode(BP_GM_RINGSLINGER)) then return end	-- not ringslinger
	
	-- make sure we don't get multiple I guess?
	local p = arrow.target.player
	if not (p.bp.rs_playerarrow and p.bp.rs_playerarrow.valid)
		-- create our own arrow to do stuff with
		local rs_arrow = P_SpawnMobj(arrow.x, arrow.y, arrow.z, MT_BP_RS_PLAYERARROW)
		rs_arrow.target = arrow.target
		rs_arrow.color = arrow.color
		rs_arrow.scale = arrow.scale
		rs_arrow.destscale = arrow.destscale
		rs_arrow.flags, rs_arrow.flags2 = arrow.flags, arrow.flags2
		
		-- link to the player for safety
		p.bp.rs_playerarrow = rs_arrow
	end
	
	-- remove the old one
	P_RemoveMobj(arrow)
end, MT_PLAYERARROW)

-- when in doubt: create your own! fuck source conflicts
freeslot("MT_BP_RS_PLAYERARROW", "SPR_RSIT")
mobjinfo[MT_BP_RS_PLAYERARROW] = {
	spawnstate = S_PLAYERARROW,
	radius = 36*FRACUNIT,
	width = 37*FRACUNIT,
	flags = MF_NOBLOCKMAP|MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY|MF_SCENERY|MF_DONTENCOREMAP
}
addHook("MobjThinker", function(arrow)
	if not (arrow and arrow.valid) then return end
	
	-- recreation of MT_PLAYERARROW's thinker with everything I'd need
	if not ((arrow.target and arrow.target.valid) and arrow.target.player)
	or (arrow.target.player.spectator)
	or not (arrow.target.player.health and arrow.target.player.playerstate != PST_DEAD)
		P_RemoveMobj(arrow)
		return
	end
	
	-- set vars in relation to target
	arrow.angle = R_PointToAngle(arrow.x, arrow.y) + ANGLE_90
	arrow.color = arrow.target.color
	K_MatchGenericExtraFlags(arrow, arrow.target)
	
	-- set scale
	local scale = 3*arrow.target.scale
	if not (splitscreen)
		if (displayplayers[0] and displayplayers[0].valid) and (displayplayers[0].mo and displayplayers[0].mo.valid)
			scale = min(16*arrow.target.scale, arrow.target.scale + FixedMul(FixedDiv(abs(P_AproxDistance(displayplayers[0].mo.x - arrow.target.x, displayplayers[0].mo.y - arrow.target.y)), 1280<<FRACBITS), arrow.target.scale))
		end
	end
	arrow.destscale = scale
	
	-- set position (flip support too)
	if (arrow.target.eflags & MFE_VERTICALFLIP)
		P_TeleportMove(arrow, arrow.target.x, arrow.target.y, arrow.target.z - (P_GetPlayerHeight(arrow.target.player)) - (16*arrow.target.scale))
		arrow.eflags = $|(MFE_VERTICALFLIP)
	else
		P_TeleportMove(arrow, arrow.target.x, arrow.target.y, arrow.target.z + (P_GetPlayerHeight(arrow.target.player)) + (16*arrow.target.scale))
		arrow.eflags = $ & ~(MFE_VERTICALFLIP)
	end
	
	-- make sure the tracer exists
	if not (arrow.tracer and arrow.tracer.valid)
		local overlay = P_SpawnMobj(arrow.x, arrow.y, arrow.z, MT_OVERLAY)
		overlay.target = arrow
		overlay.state = S_PLAYERARROW_ITEM
		overlay.scale = arrow.scale
		
		-- link to the arrow
		arrow.tracer = overlay
	end
	local item = arrow.tracer	-- laziness
	item.destscale = scale
	
	-- display the weapon ring
	if (arrow.target.player.bp.rs_weaponring)
		arrow.state = S_PLAYERARROW_BOX
		item.sprite = SPR_RSIT
		item.frame = FF_FULLBRIGHT|(arrow.target.player.bp.rs_weaponring - 1)
		
		-- flash while using autos (please work)
		if (arrow.target.player.bp.rs_forceringtoss)
			item.flags2 = (leveltime%2) and $|(MF2_DONTDRAW) or $ & ~(MF2_DONTDRAW)
		else
			item.flags2 = $ & ~(MF2_DONTDRAW)
		end
	else
		arrow.state = S_PLAYERARROW
		item.state = S_PLAYERARROW_ITEM
	end
	
	-- display how many weapon rings we have
	local number, numx	-- make sure these are defined outside the block
	if (arrow.target.player.bp.rs_amount)
	and (arrow.target.player.bp.rs_amount > 1 and arrow.target.player.bp.rs_amount < 10)
		number = P_SpawnMobj(arrow.x, arrow.y, arrow.z, MT_OVERLAY)
		number.target = arrow
		number.state = S_PLAYERARROW_NUMBER
		number.scale = arrow.scale
		number.destscale = scale
		number.frame = FF_FULLBRIGHT|(arrow.target.player.bp.rs_amount)
		
		numx = P_SpawnMobj(arrow.x, arrow.y, arrow.z, MT_OVERLAY)
		numx.target = arrow
		numx.state = S_PLAYERARROW_X
		numx.scale = arrow.scale
		numx.destscale = scale
	end
	
	-- set the visibility for the objects
	local mobjs = {arrow, item, number, numx}
	local drawFlags = {MFE_DRAWONLYFORP1, MFE_DRAWONLYFORP2, MFE_DRAWONLYFORP3, MFE_DRAWONLYFORP4}
	
	for index, mobj in pairs(mobjs)
		if not (mobj and mobj.valid) then continue end	-- uh oh
		//if (splitscreen) then mobj.flags2 = $|MF2_DONTDRAW continue end	-- FUCK splitscreen!!!!
		
		mobj.flags2 = $ & ~(MF2_DONTDRAW)
		for i = 0, (splitscreen) -- go through displayplayers
			if not (displayplayers[i]) then break end
			local p = displayplayers[i]	-- laziness
			if not (p and p.valid) then continue end
			
			-- dont draw our own marker
			if (arrow.target.player == p) then mobj.eflags = $ & ~(drawFlags[i+1]) continue end
			mobj.eflags = $|drawFlags[i+1]
		end
		-- somehow this isnt supposed to be displayed at ALL for any displayplayers, so dont draw it at all
		if not (mobj.eflags & (MFE_DRAWONLYFORP1|MFE_DRAWONLYFORP2|MFE_DRAWONLYFORP3|MFE_DRAWONLYFORP4))
			mobj.flags2 = $|MF2_DONTDRAW
		end
	end
end, MT_BP_RS_PLAYERARROW)