local ringslinger = battleplus.ringslinger

ringslinger.Func_ReplaceItemboxes = function()
	-- ok yeah, we're only doing this ONCE, I am *NOT* iterating through every item box every tic :grimacing:
	if (server.bp_server.rs_panels) then return end
	server.bp_server.rs_panels = {}	-- since this table exists, this wont be done again
	local panels = server.bp_server.rs_panels
	
	-- get items to replace
	for m in thinkers.iterate("mobj")
		if (m.type != MT_RANDOMITEM) then continue end	-- not a random item
		local sprites = {SPR_PIKA, SPR_PIKB, SPR_PIKS, SPR_PIKG, SPR_PIKE, SPR_PIKR}
		local weaponring = P_RandomRange(1, 6)
		
		-- if they're right on the ground, make them float!
		local z_add = 0
		local abs_floorz = battleplus.GetAbsoluteFloorZ(m)
		if not (abs(m.z - abs_floorz)) then z_add = 24*(mapobjectscale) end
		
		local panel = P_SpawnMobj(m.x, m.y, m.z + z_add, MT_BP_RS_WEAPONPICKUP)
		panel.scale = FixedMul(FRACUNIT*6/5, $)
		panel.sprite = sprites[weaponring]
		panel.rs_weaponring = weaponring
		panel.rs_leveltime = leveltime
		table.insert(panels, {panel, {panel.x, panel.y, panel.z}})
		-- ^ store panel and panel location, so we can replace them later
		
		local pop = P_SpawnMobj(m.x, m.y, m.z, MT_THOK)
		pop.state = S_XPLD1
		pop.scale = m.scale
		S_StartSound(pop, sfx_pop)
		P_RemoveMobj(m)
	end
	
	-- set this to be used in other functions
	server.bp_server.rs_numpanels = #panels
end

ringslinger.Func_ReplacePanels = function()
	server.bp_server.rs_numgotpanels = $ or 0
	local numpanels = server.bp_server.rs_numpanels
	local numgotpanels = server.bp_server.rs_numgotpanels
	
	//print("numgotpanels: "..numgotpanels)
	//print("numpanels: "..numpanels)
	//print("panelsneeded: "..(4*numpanels/5))
	
	-- replace panels ala vanilla battle boxes
	if (numgotpanels >= (4*numpanels/5))
		local panels = server.bp_server.rs_panels
		
		for i = 1, #panels
			-- the current panel
			local panel = panels[i][1]
			if (panel and panel.valid)	-- the panel is valid to do checks on
				if (panel.health)	-- still well and healthy, stay alive
				or (panel.rs_threshold)	-- recently hit
					continue
				end
			end
			
			-- respawn the panel!
			local sprites = {SPR_PIKA, SPR_PIKB, SPR_PIKS, SPR_PIKG, SPR_PIKE, SPR_PIKR}
			local weaponring = P_RandomRange(1, 6)
			local x, y, z = panels[i][2][1], panels[i][2][2], panels[i][2][3]
			
			local panel = P_SpawnMobj(x, y, z, MT_BP_RS_WEAPONPICKUP)
			panel.scale = FixedMul(FRACUNIT*6/5, $)
			panel.sprite = sprites[weaponring]
			panel.rs_weaponring = weaponring
			panel.rs_leveltime = leveltime
			
			-- update table info
			panels[i][1] = panel
			server.bp_server.rs_numgotpanels = $ - 1
		end
	end
end

ringslinger.Func_PlayerPunishment = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	if not (mo.health)
		if (p.bp.rs_punishment) then return end
		p.marescore = $/2
		p.bp.rs_punishment = true
	else
		if not (p.bp.rs_punishment) then return end
		p.bp.rs_punishment = nil
	end
end

ringslinger.Func_PlayerWeaponRings = function(p)
	-- remove ANY item we may have
	p.kartstuff[k_itemtype] = KITEM_NONE
	p.kartstuff[k_itemamount] = 0
	
	-- make sure these are defined
	p.bp.rs_weaponring = $ or 0
	p.bp.rs_amount = $ or 0
	
	-- recreate PF_ATTACKDOWN functionality
	if not ((p.cmd.buttons & BT_ATTACK) or p.bp.rs_forceringtoss)
	or (p.bp.rs_ringtossdown)
	or (p.bp.rs_ringtosscooldown)
		-- remove this if set
		if not ((p.cmd.buttons & BT_ATTACK) or p.bp.rs_forceringtoss)
			p.bp.rs_ringtossdown = nil
		end
		
		-- reduce cooldown
		p.bp.rs_ringtosscooldown = $ and $-1 or 0
		return
	end
	p.bp.rs_ringtossdown = true	-- set this here
	
	-- validity checks on if we can actually attack
	local mo = p.mo
	if not ((mo and mo.valid) and mo.health)	-- not alive
	or (p.spectator)	-- make sure this is removed
	or (p.exiting)	-- not when we've finished
	or (p.kartstuff[k_spinouttimer] or p.kartstuff[k_squishedtimer] or p.kartstuff[k_respawn])	-- our vital organs!
	or not (p.bp.rs_weaponring and p.bp.rs_amount)	-- no weapon rings
		p.bp.rs_forceringtoss = nil	-- remove this here just incase
		p.bp.rs_forcedir = nil
		return 
	end
	
	-- get the ring we're throwing
	local rings = {"AUTOMATIC", "BOUNCE", "SCATTER", "GRENADE", "EXPLOSION", "RAIL"}
	local mobj = _G["MT_BP_RS_"..rings[p.bp.rs_weaponring].."RING"]
		
	-- throw the ring out!
	if (p.bp.rs_weaponring == BP_RW_SCATTER)	-- we need to throw multiple!
		-- fire 5 instead of 1
		for i = -2, 2
			local angle = mo.angle + ((p.kartstuff[k_throwdir] == -1) and ANGLE_180 or 0)
			local ring = P_SPMAngle(mo, mobj, angle + (ANG2 * i), 0)
			
			if (ring and ring.valid)
				ring.color = p.skincolor
				ring.colorized = true
				P_TeleportMove(ring, ring.x, ring.y, ring.z + (P_GetPlayerHeight(p)/4))
			end
		end
	elseif (p.bp.rs_weaponring == BP_RW_GRENADE)
		local angle = mo.angle + ((p.kartstuff[k_throwdir] == -1) and ANGLE_180 or 0)
		local ring = P_SPMAngle(mo, mobj, angle, 0)
		
		if (ring and ring.valid)
			//ring.threshold = 10	-- for some reason this just wouldnt be detected by the collide hook... bruh?
			ring.color = p.skincolor
			ring.colorized = true
			P_TeleportMove(ring, ring.x, ring.y, ring.z + (P_GetPlayerHeight(p)/4))
			
			-- adjust trajectory
			if (p.kartstuff[k_throwdir] == -1)	-- backwards
				ring.momx = FixedMul(FRACUNIT/6, $)
				ring.momy = FixedMul(FRACUNIT/6, $)
			else	-- forwards
				ring.momx = $ + mo.momx
				ring.momy = $ + mo.momy
			end
		end
	elseif (p.bp.rs_weaponring == BP_RW_RAIL)	-- oh boy
		local angle = mo.angle + ((p.kartstuff[k_throwdir] == -1) and ANGLE_180 or 0)
		local ring = P_SPMAngle(mo, mobj, angle, 0, MF2_RAILRING|MF2_DONTDRAW)
			
		if (ring and ring.valid)
			P_TeleportMove(ring, ring.x, ring.y, ring.z + (P_GetPlayerHeight(p)/4))
			S_StartSound(mo, ring.info.seesound)	-- gotta play this here, object is too fast
			
			-- make it actually behave like a rail
			for i = 0, 255
				-- spark trail
				if (i%2) then P_SpawnMobj(ring.x, ring.y, ring.z, MT_SPARK) end
				
				-- move the ring continuously in the loop to make it act like a rail
				if not (P_TryMove(ring, ring.x + ring.momx, ring.y + ring.momy, true))
					P_KillMobj(ring)
					break
				end
			end
			
			-- if it's still valid here, remove it
			if (ring and ring.valid) then P_KillMobj(ring) end
		end
	else
		local throwdir = p.bp.rs_forcedir or p.kartstuff[k_throwdir]
		local angle = mo.angle + ((throwdir == -1) and ANGLE_180 or 0)
		local ring = P_SPMAngle(mo, mobj, angle, 0)
		
		if (ring and ring.valid)
			ring.color = p.skincolor
			ring.colorized = true
			P_TeleportMove(ring, ring.x, ring.y, ring.z + (P_GetPlayerHeight(p)/4))
		end
		
		-- special handling for automatics
		if (p.bp.rs_weaponring == BP_RW_AUTOMATIC)
			-- continuous fire (and the only ring that has a cooldown rn!)
			p.bp.rs_ringtossdown = nil
			p.bp.rs_ringtosscooldown = 2
			p.bp.rs_forceringtoss = true	-- so it stays continuous
			if (p.bp.rs_forcedir == nil)	-- only if it's not set
				p.bp.rs_forcedir = p.kartstuff[k_throwdir]	-- save direction, too
			end
		end
	end
		
	-- decrease weapon ring amount (remove the weapon ring once we hit 0)
	p.bp.rs_amount = $-1
	if not (p.bp.rs_amount)
		p.bp.rs_weaponring = 0
	end
end

ringslinger.Func = function()
	ringslinger.Func_ReplaceItemboxes()
	ringslinger.Func_ReplacePanels()
	ringslinger.AnimatePanels()
	for p in players.iterate
		ringslinger.Func_PlayerPunishment(p)
		ringslinger.Func_PlayerWeaponRings(p)
	end
end

-- no need to define other functions when this is the only hud related thing
ringslinger.HUDFunc = function(v, p, c)
	-- conditions when we wouldnt draw it normally
	if (p.exiting or p.spectator) then return end
	if (p.kartstuff[k_comebacktimer] and not p.kartstuff[k_bumper]) then return end
	
	-- don't do anything without a weapon ring
	if not (p.bp.rs_weaponring) then return end
	
	-- display constants
	local dp = battleplus.IsDisplayPlayer(p)
	local skin = (p.kartstuff[k_itemblink] and leveltime%2) and -6 or -1
	local colour = (p.kartstuff[k_itemblink] and leveltime%2) and SKINCOLOR_WHITE or SKINCOLOR_NONE
	
	-- weapon ring constants
	local patches = {"ATOR", "BNCR", "SCTR", "GRNR", "BOMR", "RALR"}
	local weaponring = p.bp.rs_weaponring
	local amount = p.bp.rs_amount or 1
	
	if (splitscreen > 1)
		local right = (dp == 1 or dp == 3)
		local hx = (right) and 121 or -9
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOTOP|V_SNAPTOLEFT|V_SNAPTORIGHT|V_HUDTRANS, dp)
		
		-- the itembox background
		v.draw(hx, -8, v.cachePatch("K_ISBG"), flags)
		if (p.kartstuff[k_itemamount] > 1) then v.draw(hx+((right) and 48 or 0), -8, v.cachePatch("K_ISMUL"), ((right) and V_FLIP or 0)|flags) end
		
		-- the weapon ring we have
		local weaponring_patch = "K_IS"..patches[weaponring]
		v.draw(hx, -8, v.cachePatch(weaponring_patch), flags, v.getColormap(skin, colour))
		
		-- how many we have!
		if (amount > 1)
			local hx2 = (right) and 123 or 15
			v.drawString(hx2, 23, "x"..amount, V_ALLOWLOWERCASE|flags, "left")
		end
	else
		local hx = 5
		local hy = (splitscreen) and 3 or 5
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOTOP|V_SNAPTOLEFT|V_HUDTRANS, dp)
		
		-- the itembox background
		v.draw(hx, hy, v.cachePatch("K_ITBG"), flags)
		if (amount > 1) then v.draw(hx, hy, v.cachePatch("K_ITMUL"), flags) end
		
		-- the weapon ring we have
		local weaponring_patch = "K_IT"..patches[weaponring]
		v.draw(hx, hy, v.cachePatch(weaponring_patch), flags, v.getColormap(skin, colour))
		
		-- how many we have!
		if (amount > 1)
			v.draw(hx+28 - (splitscreen and 2 or 0), hy+41, v.cachePatch("K_ITX"), flags)
			v.drawKartString(hx+38, hy+36, amount, flags)
		end
	end
end

-- damage function for rings
ringslinger.MobjDamage = function(mo, inflictor, source)
	-- should be a player
	if not ((mo and mo.valid) and mo.player) then return end
	local p = mo.player
	
	-- we need an inflictor here
	if not (inflictor and inflictor.valid) then return end
	
	-- make sure it's a ring from this gametype
	if not (inflictor.type == MT_BP_RS_AUTOMATICRING)
	and not (inflictor.type == MT_BP_RS_BOUNCERING)
	and not (inflictor.type == MT_BP_RS_SCATTERRING)
	and not (inflictor.type == MT_BP_RS_GRENADERING)
	//and not (inflictor.type == MT_BP_RS_EXPLOSIONRING)
	and not (inflictor.type == MT_BP_RS_RAILRING)
		if (inflictor.type == MT_BP_RS_EXPLOSIONRING)
			-- hitfeed support
			if (hitfeed)
				if not (HF_animIcon["K_HMBOM"]) then HF_animIcon["K_HMBOM"] = {2, 2} end
				
				local s = {{source.player.name, source.hf_overridenamecolor or source.player.skincolor}, {skins[source.skin].facemmap, source.player.skincolor}, source.player}
				local v = {{p.name, mo.hf_overridenamecolor or p.skincolor}, {skins[mo.skin].facemmap, p.skincolor}, p}
				if (s[3] == v[3]) then s = nil end	-- source == victim, remove source
				
				HF_SendHitMessage(s, v, "K_HMBOM")
			end
			
			-- make sure we don't spin out (they explode on impact!)
			return true
		end
		
		-- I PROMISE there are no more checks like this
		return
	end
	
	-- get base knockback
	local pushx = FixedMul(FRACUNIT/7, inflictor.momx)
	local pushy = FixedMul(FRACUNIT/7, inflictor.momy)
	
	-- certain ringtypes have different knockback
	if (inflictor.type == MT_BP_RS_SCATTERRING)
		-- these will stack up close for a devastating effect, since multiple can hit
		pushx = FixedMul(FRACUNIT*9/20, inflictor.momx)
		pushy = FixedMul(FRACUNIT*9/20, inflictor.momy)
	elseif (inflictor.type == MT_BP_RS_GRENADERING)
		if (inflictor.health)	-- hasn't exploded yet
			-- hitfeed support
			if (hitfeed)
				if not (HF_animIcon["K_HMGRN"]) then HF_animIcon["K_HMGRN"] = {2, 2} end
				
				local s = {{source.player.name, source.hf_overridenamecolor or source.player.skincolor}, {skins[source.skin].facemmap, source.player.skincolor}, source.player}
				local v = {{p.name, mo.hf_overridenamecolor or p.skincolor}, {skins[mo.skin].facemmap, p.skincolor}, p}
				if (s[3] == v[3]) then s = nil end	-- source == victim, remove source
				
				HF_SendHitMessage(s, v, "K_HMGRN")
			end
			
			-- make them explode on contact while moving!
			P_KillMobj(inflictor)
			return true
		end
	elseif (inflictor.type == MT_BP_RS_RAILRING)
		-- these are a little powerful with full knockback LOL
		pushx = FixedMul(FRACUNIT*9/10, inflictor.momx)
		pushy = FixedMul(FRACUNIT*9/10, inflictor.momy)
	end
	
	-- apply knockback
	mo.momx = $ + pushx
	mo.momy = $ + pushy
	S_StartSound(mo, sfx_s3k7b)
	
	-- spinout the player manually so we can wipeout the player instead of spinout
	if (p.kartstuff[k_invincibilitytimer] > 0)
	or (p.kartstuff[k_growshrinktimer] > 0)
	or (p.powers[pw_flashing])
		-- insta shield them instead!
		K_DoInstashield(p)
	else
		K_SpinPlayer(mo.player, source, 1, inflictor)
		P_PlayRinglossSound(mo)
		
		-- hitfeed support
		if (hitfeed)
			local ringslinger_icons = {
				[MT_BP_RS_AUTOMATICRING] = "K_HMATOR",
				[MT_BP_RS_BOUNCERING] = "K_HMBNCR",
				[MT_BP_RS_SCATTERRING] = "K_HMSCTR",
				[MT_BP_RS_GRENADERING] = "K_HMGRN1",	-- guess this is pretty useless, huh
				[MT_BP_RS_EXPLOSIONRING] = "K_HMBOM1",	-- ...this too
				[MT_BP_RS_RAILRING] = "K_HMRALR",
			}
			
			local s = {{source.player.name, source.hf_overridenamecolor or source.player.skincolor}, {skins[source.skin].facemmap, source.player.skincolor}, source.player}
			local v = {{p.name, mo.hf_overridenamecolor or p.skincolor}, {skins[mo.skin].facemmap, p.skincolor}, p}
			if (s[3] == v[3]) then s = nil end	-- source == victim, remove source
			
			HF_SendHitMessage(s, v, ringslinger_icons[inflictor.type])
		end
	end
	
	return true
end

local ringslinger_info = {
	name = "Ringslinger",
	teamname = "Team Ringslinger",
	objective = "Score the most points!",
	icon = "K_ITRALR",
	func = ringslinger.Func,
	hudfunc = ringslinger.HUDFunc,
	mobjdamage = ringslinger.MobjDamage,
	rules = BP_GMR_TEAMS|BP_GMR_TIMELIMIT|BP_GMR_REMOVEWANTED,
	timelimit = ringslinger.cv_timelimit,
}
battleplus.AddGamemode(BP_GM_RINGSLINGER, ringslinger_info)