rawset(_G, "battleplus", {}) -- Battle Plus global namespace
battleplus.version = "v2.0"

-- DEVMODE
/* HUGE list of modes for devmode testing:
	1: adding gamemodes
	2: gamemode selection
	3: itemodds testing
	4: gamemode specifics
*/
battleplus.cv_devmode = CV_RegisterVar({
	name = "bp_devmode",
	defaultvalue = "0",
	PossibleValue = CV_Unsigned,
})
-- devmode print, I'll be using this a lot
battleplus.dprint = function(mode, string)
	if not (battleplus.cv_devmode.value) then return end
	if (battleplus.cv_devmode.value != mode) then return end
	print(string)
end
local dprint = battleplus.dprint	-- oh yeah woo yeah

-- define gamemode constants
battleplus.constants = 0	-- keep track of how many gamemode constants there are
-- used to create and assign a custom constant for use in
battleplus.DefineGamemodeConstant = function(constant)
	-- check if constant is already defined
	if (_G[constant])
		local value = (_G[constant])
		dprint(1, "Constant "..constant.." already defined: "..value)
		return
	end

	local value = (battleplus.constants + 1)
	rawset(_G, constant, value)
	battleplus.constants = $+1
	dprint(1, "Defined gamemode constant "..constant..": "..value)
end
-- define battle plus gamemodes
local battleplus_gamemodes = {
	"BP_GM_BUMPERBATTLE",
	"BP_GM_RINGRUNNERS",
	"BP_GM_EMERALDTHIEF",
	"BP_GM_MINEBLAST",
	"BP_GM_HOTEGGTATO",
	"BP_GM_SINKDEATHMATCH",
	"BP_GM_ITEMRAIN",
	"BP_GM_RANDOMIZER",
	"BP_GM_CHAOSRINGRUNNERS",
	"BP_GM_RINGSLINGER",
	"BP_GM_HOOLIGANROUNDUP",
	"BP_GM_EGGFECTION",
}
for k, v in ipairs(battleplus_gamemodes)
	battleplus.DefineGamemodeConstant(v)
end

-- define gamemode rules
local gamemode_rules = {
	"BP_GMR_TEAMS",	-- this gamemode supports teams
	"BP_GMR_FORCETEAMS",	-- this gamemode NEEDS teams
	"BP_GMR_HELL",	-- gamemode is a hell gamemode
	"BP_GMR_TIMELIMIT",	-- has a timelimit, controlled with .timelimit
	"BP_GMR_OVERTIME",	-- has overtime when timelimit runs out
	"BP_GMR_BUMPERS",	-- .Bumpers sets the cvar that controls this
	"BP_GMR_ELIMINATION",	-- special functionality depending on BP_GMR_BUMPERS
	"BP_GMR_REMOVEWANTED",	-- removes wanted entirely from a gamemode (doesn't work fully with BP_GMR_BUMPERS!)
	"BP_GMR_REMOVEPOINTS",	-- despawns battle points spawned above players
	"BP_GMR_HIDETEAMSCORES",	-- hides team points in team gamemodes, incase you want to do your own thing
	"BP_GMR_OVERRIDETEAMJAWZ",	-- for overriding the teams jawz targetting (currently only used with emerald thief)
}
for k, v in ipairs(gamemode_rules)
	rawset(_G, v, 2^(k-1))
end

-- add gamemodes!
battleplus.gamemodes = {}
-- gamemode is the gamemode number, info is the table info
battleplus.AddGamemode = function(constant, info)
	if not (constant) then dprint(1, "No gamemode constant/number given") return end	-- needs a gamemode constant/number

	-- no overwriting gamemodes
	if (battleplus.gamemodes[constant])
		local name = battleplus.gamemodes[constant].name or "no name?"
		dprint(1, "Gamemode "..constant.." already added: "..name)
		return
	end

	-- not added, lets add the gamemode!
	if not (info) then dprint(1, "No gamemode info table given") return end	-- we need info for the gamemode
	battleplus.gamemodes[constant] = info
	dprint(1, "Added gamemode "..constant..": "..(info.name or "no name?"))
end

-- Pseudorandom Numbers Library v1.0 by LJ Sonic
-- Only use for HUD stuff
local t, x, y, z, w = 0, 69, 690, 6969, 65469
battleplus.N_Random = function()
	t = x ^^ (x << 11)
	x, y, z, w = y, z, w, w ^^ (w >> 19) ^^ t ^^ (t >> 8)
	return w
end
battleplus.N_RandomKey = function(n) return abs(battleplus.N_Random() % n) end
battleplus.N_RandomRange = function(a, b) return a + abs(battleplus.N_Random() % (b - a + 1)) end
battleplus.N_RandomDatas = function() return d end

-- general use functions
battleplus.IsGamemode = function(gamemode)
	if not (server.bp_server) then return false end	-- uh oh
	return (server.bp_server.gamemode == gamemode)
end

battleplus.EndBattle = function()
	if not G_BattleGametype() return end -- This is only meant to end battles, not races!

	for p in players.iterate
		if p.exiting then continue end -- player is already exiting
		p.exiting = 281
		K_KartUpdatePosition(p) -- update the postitions so the correct music plays!
	end

	for p in players.iterate
		-- Set the music
        if not (p and p.mo) or p.spectator then continue end

		if p.exiting
			if p.kartstuff[k_position] == 1
				S_ChangeMusic("kbwin", false, p)
				//print(p.name.." was winning, playing WIN music.")
			else
				if K_IsPlayerLosing(p)
					S_ChangeMusic("kblose", false, p)
					//print(p.name.." was losing, playing LOSE music.")
				else
					S_ChangeMusic("kbok", false, p)
					//print(p.name.." was doing alright, playing COOL music.")
				end
			end
		end
	end
end

battleplus.GetPlayerCount = function(spectators)
	local pnum = 0
	for p in players.iterate
		if (p.spectator and not (spectators)) then continue end
		pnum = $+1
	end

	return pnum
end

battleplus.GetDeadCount = function(team)
	local pnum = 0
	for p in players.iterate
		if p.spectator then continue end
		if (team) and (p.bp.team ~= (team)) then continue end
		if p.kartstuff[k_bumper] then continue end

		pnum = $+1
	end

	return pnum
end

battleplus.SpawnBumper = function(p)
	if not (p and p.mo) or p.spectator then return end
	if not G_BattleGametype() then return end -- Only spawn bumpers in Battle.
	local diff = 0

	local newbumper = p.kartstuff[k_bumper]
	if (newbumper <= 1)
		diff = 0
	else
		diff = FixedAngle(360*FRACUNIT/newbumper)
	end

	local newangle = p.mo.angle
	local newx = p.mo.x + P_ReturnThrustX(p.mo, newangle + ANGLE_180, 64*FRACUNIT)
	local newy = p.mo.y + P_ReturnThrustY(p.mo, newangle + ANGLE_180, 64*FRACUNIT)

	local newmo = P_SpawnMobj(newx, newy, p.mo.z, MT_BATTLEBUMPER)
	newmo.threshold = newbumper
	newmo.target = p.mo
	newmo.angle = (diff*(newbumper-1))
	newmo.color = p.skincolor

	if (newbumper+1 < 2)
		newmo.state = S_BATTLEBUMPER3
	elseif (newbumper+1 < 3)
		newmo.state = S_BATTLEBUMPER2
	else
		newmo.state = S_BATTLEBUMPER1
	end

	return newmo
end

battleplus.ColorToMap = function(c, vflag, num)
	local color_sets = {
	{SKINCOLOR_WHITE, SKINCOLOR_SILVER, SKINCOLOR_SLATE}, -- 1: No colour mapping
	{SKINCOLOR_GREY, SKINCOLOR_NICKEL, SKINCOLOR_BLACK, SKINCOLOR_JET}, -- 2: V_GREYMAP
	{SKINCOLOR_SEPIA, SKINCOLOR_BEIGE, SKINCOLOR_BROWN, SKINCOLOR_LEATHER, SKINCOLOR_RUST}, -- 3: V_BROWNMAP
	{SKINCOLOR_FAIRY, SKINCOLOR_SALMON, SKINCOLOR_PINK, SKINCOLOR_ROSE, SKINCOLOR_BRICK, SKINCOLOR_BUBBLEGUM, SKINCOLOR_LILAC}, -- 4: V_PINKMAP
	{SKINCOLOR_RUBY, SKINCOLOR_RASPBERRY, SKINCOLOR_CHERRY, SKINCOLOR_RED, SKINCOLOR_CRIMSON, SKINCOLOR_MAROON, SKINCOLOR_FLAME, SKINCOLOR_SCARLET, SKINCOLOR_KETCHUP}, -- 5: V_REDMAP
	{SKINCOLOR_DAWN, SKINCOLOR_SUNSET, SKINCOLOR_CREAMSICLE, SKINCOLOR_ORANGE, SKINCOLOR_PUMPKIN, SKINCOLOR_ROSEWOOD, SKINCOLOR_BURGUNDY, SKINCOLOR_TANGERINE}, -- 6: V_ORANGEMAP
	{SKINCOLOR_PEACH, SKINCOLOR_CARAMEL, SKINCOLOR_CREAM}, -- 7: V_PEACHMAP
	{SKINCOLOR_GOLD, SKINCOLOR_ROYAL, SKINCOLOR_BRONZE, SKINCOLOR_COPPER}, -- 8: V_GOLDMAP
	{SKINCOLOR_POPCORN, SKINCOLOR_YELLOW, SKINCOLOR_MUSTARD, SKINCOLOR_OLIVE}, -- 9: V_YELLOWMAP
	{SKINCOLOR_VOMIT, SKINCOLOR_GARDEN, SKINCOLOR_TEA, SKINCOLOR_PISTACHIO}, -- 10: V_TEAMAP
	{SKINCOLOR_LIME, SKINCOLOR_HANDHELD, SKINCOLOR_MOSS, SKINCOLOR_CAMOUFLAGE, SKINCOLOR_ROBOHOOD, SKINCOLOR_MINT, SKINCOLOR_GREEN, SKINCOLOR_PINETREE, SKINCOLOR_EMERALD, SKINCOLOR_SWAMP, SKINCOLOR_DREAM, SKINCOLOR_PLAGUE, SKINCOLOR_ALGAE}, -- 11: V_GREENMAP
	{SKINCOLOR_CARIBBEAN, SKINCOLOR_AQUA, SKINCOLOR_TEAL, SKINCOLOR_CYAN, SKINCOLOR_JAWZ, SKINCOLOR_CERULEAN, SKINCOLOR_NAVY, SKINCOLOR_SAPPHIRE}, -- 12: V_SKYMAP
	{SKINCOLOR_PLATINUM, SKINCOLOR_STEEL}, -- 13: V_STEELMAP
	{SKINCOLOR_PERIWINKLE, SKINCOLOR_BLUE, SKINCOLOR_BLUEBERRY, SKINCOLOR_NOVA}, -- 14: V_BLUEMAP
	{SKINCOLOR_ULTRAVIOLET, SKINCOLOR_PURPLE, SKINCOLOR_FUCHSIA}, -- 15: V_PURPLEMAP
	{SKINCOLOR_PASTEL, SKINCOLOR_MOONSLAM, SKINCOLOR_DUSK, SKINCOLOR_TOXIC, SKINCOLOR_MAUVE, SKINCOLOR_LAVENDER, SKINCOLOR_BYZANTIUM, SKINCOLOR_POMEGRANATE} -- 16: V_LAVENDERMAP
	}

	local maps = {"\x80", "\x86", "\x8e", "\x8d", "\x85", "\x87", "\x8f", "\x8A", "\x82", "\x8b", "\x83", "\x88", "\x8c", "\x84", "\x81", "\x89"}
	if (vflag)
		maps = {0, V_GRAYMAP, V_BROWNMAP, V_PINKMAP, V_REDMAP, V_ORANGEMAP, V_PEACHMAP, V_GOLDMAP, V_YELLOWMAP, V_TEAMAP, V_GREENMAP, V_SKYMAP, V_STEELMAP, V_BLUEMAP, V_PURPLEMAP, V_LAVENDERMAP}
	end

	for i = 1, #color_sets
		for j = 1, #color_sets[i]
			local color = color_sets[i][j]

			if (color == c)
				return ((num) and i or maps[i])
			end
		end
	end

	return ((num) and 1 or maps[1])
end

battleplus.ColorToString = function(c)
	local strings = {"White", "Silver", "Grey", "Nickel", "Black", "Skunk", "Fairy", "Popcorn", "Artichoke", "Pigeon", "Sepia", "Beige", "Walnut", "Brown", "Leather", "Salmon", "Pink", "Rose", "Brick", "Cinnamon", "Ruby", "Raspberry", "Cherry", "Red", "Crimson", "Maroon", "Lemonade", "Flame", "Scarlet", "Ketchup", "Dawn", "Sunset", "Creamsicle", "Orange", "Pumpkin", "Rosewood", "Burgundy", "Tangerine", "Peach", "Caramel", "Cream", "Gold", "Royal", "Bronze", "Copper", "Quarry", "Yellow", "Mustard", "Crocodile", "Olive", "Vomit", "Garden", "Lime", "Handheld", "Tea", "Pistachio", "Moss", "Camouflage", "Robo-Hood", "Mint", "Green", "Pinetree", "Emerald", "Swamp", "Dream", "Plague", "Algae", "Caribbean", "Azure", "Aqua", "Teal", "Cyan", "Jawz", "Cerulean", "Navy", "Platinum", "Slate", "Steel", "Thunder", "Rust", "Wristwatch", "Jet", "Sapphire", "Periwinkle", "Blue", "Blueberry", "Nova", "Pastel", "Moonslam", "Ultraviolet", "Dusk", "Bubblegum", "Purple", "Fuchsia", "Toxic", "Mauve", "Lavender", "Byzantium", "Pomegranate", "Lilac"}
	return strings[tonumber(c)]
end

battleplus.AssignTeam = function(p)
	if not (p and (p.mo and p.mo.valid)) then return end
	if (p.spectator) then return end
	if not (G_BattleGametype()) then return end

	local c = {
		{battleplus.ColorToMap(battleplus.cv_t1c.value), battleplus.ColorToString(battleplus.cv_t1c.value)},
		{battleplus.ColorToMap(battleplus.cv_t2c.value), battleplus.ColorToString(battleplus.cv_t2c.value)}
	}
	p.bp.savedskincolor = p.skincolor
	if (#server.bp_server.teams[1] == #server.bp_server.teams[2])	-- teams are evenly stacked
		local team = P_RandomRange(1, 2)

		table.insert(server.bp_server.teams[team], p)
		print(p.name.." joined the "..(c[team][1])..(c[team][2]).." Team".."\x80"..".")
		p.bp.team = team	-- what team we're on
	elseif (#server.bp_server.teams[1] < #server.bp_server.teams[2])	-- more on team 2
		table.insert(server.bp_server.teams[1], p)
		print(p.name.." joined the "..(c[1][1])..(c[1][2]).." Team".."\x80"..".")
		p.bp.team = 1
	else	-- more on team 1
		table.insert(server.bp_server.teams[2], p)
		print(p.name.." joined the "..(c[2][1])..(c[2][2]).." Team".."\x80"..".")
		p.bp.team = 2
	end
end

battleplus.TallyTeams = function(set, finish)
	if not (server.bp_server.teamgamemode) then return end	-- we dont need to tally

	if (server.bp_server.teamtally)	-- already tallied, use cached values
		for t = 1, 2
			for k, v in ipairs(server.bp_server.teams[t])
				if not (v and v.valid) then continue end
				v.marescore = server.bp_server.teamtally[t]
			end
		end
	else
		local sums = {0, 0}
		for t = 1, 2
			-- get points on the team
			for k, v in ipairs(server.bp_server.teams[t])
				if not (v and v.valid) then continue end
				sums[t] = $+v.marescore
			end
			-- give points
			if not (set) then continue end
			for k, v in ipairs(server.bp_server.teams[t])
				if not (v and v.valid) then continue end
				v.marescore = sums[t]
			end
		end
		-- cache values if need be
		if (finish) then server.bp_server.teamtally = sums end
	end
end

battleplus.SmartSetColor = function(p, colorized, color, ninvinc, ngrowth, nshrink, neggman, ndropdash)
	if not (p and p.mo) or p.spectator then return end
	local flashtime = 4<<(p.kartstuff[k_eggmanexplode]/TICRATE)

	if (p.kartstuff[k_invincibilitytimer] and not ninvinc) -- invincibility is an obvious no-no
	or (((p.kartstuff[k_growshrinktimer] ~= 0 and p.kartstuff[k_growshrinktimer] < 0) and p.kartstuff[k_growshrinktimer] % 5 == 0) and not ngrowth) -- player is flashing due to growth
	or (((p.kartstuff[k_growshrinktimer] ~= 0 and p.kartstuff[k_growshrinktimer] > 0) and p.kartstuff[k_growshrinktimer] % 5 == 0) and not nshrink) -- player is flashing due to shrink
	or ((p.kartstuff[k_eggmanexplode] and not ((p.kartstuff[k_eggmanexplode]+1) % (flashtime/2) != 0)) and not neggman) -- player is flashing due to eggman
	or ((p.kartstuff[k_dropdash] >= TICRATE/4 and p.kartstuff[k_dropdash] & 1) and not ndropdash) -- player is flashing due to dropdash
		return
	end
	p.mo.colorized, p.mo.color = colorized, color
end

battleplus.IsDisplayPlayer = function(p)
	for i = 0, (splitscreen)	-- displayplayers.iterate breaks replays rn so we'll work with this
		local dp = displayplayers[i]

		if (p == dp) then return i end	-- return their index if they're a displayplayer too
	end
	-- if we're here, we're not a displayplayer
	return false
end

battleplus.ReturnSplitFlags = function(f, displaynum)
	if not (splitscreen) return f end

	-- exhibit 1: 2p
	if (splitscreen == 1)
		if not (displaynum) -- p1
			return f & ~(V_SNAPTOBOTTOM)
		else	-- p2
			return (f|V_SPLITSCREEN) & ~(V_SNAPTOTOP)
		end
	else
		-- jesus fuck this sucks dicks
		if not (displaynum)	-- p1
			return f & ~(V_SNAPTOBOTTOM|V_SNAPTORIGHT)
		elseif (displaynum == 1)	-- p2
			return (f|V_HORZSCREEN) & ~(V_SNAPTOBOTTOM|V_SNAPTOLEFT)
		elseif (displaynum == 2)	-- p3
			return (f|V_SPLITSCREEN) & ~(V_SNAPTOTOP|V_SNAPTORIGHT)
		elseif (displaynum == 3)	-- p4
			return (f|V_SPLITSCREEN|V_HORZSCREEN) & ~(V_SNAPTOTOP|V_SNAPTOLEFT)
		end
	end
	return f	-- shouldn't happen but don't return nil either way lol
end

-- courtesy of ampersand
local mkfnt, bpfnt = {}, {}
for i = 1, 255
    local s1, s2 = string.format("MKFNT%03d", i), string.format("BPFNT%03d", i)
    mkfnt[i], bpfnt[i] = s1, s2
end
battleplus.DrawKartText	= function(v, x, y, str, flags, align, redfont)
    if str == "" then str = "-" end
    if str == nil then str = "nil" end
    str = tostring(str)
    str = str:upper()
    local dt = {}
    local offset = 0
    for i = 1,#str do
        local by = string.byte(str:sub(i,i))
		if (redfont)
			if bpfnt[by] ~= nil
			and v.patchExists(bpfnt[by])    -- if patch exists, ready it
				local pp = v.cachePatch(bpfnt[by])
				dt[#dt+1] = {pp, offset}
				offset = $ + pp.width
			else
				offset = $ + 16
			end
		else
			if mkfnt[by] ~= nil
			and v.patchExists(mkfnt[by])    -- if patch exists, ready it
				local pp = v.cachePatch(mkfnt[by])
				dt[#dt+1] = {pp, offset}
				offset = $ + pp.width
			else
				offset = $ + 16
			end
		end
    end

    local eo = 0
    if align == "center" then
        eo = offset/2
    end
    if align == "right" then
        eo = offset
    end

    for i = 1,#dt do
        local ix = dt[i]
        v.draw(x+ix[2]-eo, y, ix[1], flags)
    end
end

battleplus.AngToInt = function(a)
	if not a return 0 end
    a = $/ANG1
    if a < 0
        a = $+360
    end
    return a
end

battleplus.SpawnBattlePoints = function(p, v, points)
	-- set some conditions/fallbacks
	if not (p and p.mo) or p.spectator then return end
	if not (points) then return end	-- no
	
	-- setup what points we'll be spawning
	local pointTable = {}
	while (points)
		-- keep adding to the table till we have 0 points
		table.insert(pointTable, min(3, points))
		points = $ - min(3, points)
	end
	
	-- spawn the points, all centered!
	local pointStates = {S_BATTLEPOINT1A, S_BATTLEPOINT2A, S_BATTLEPOINT3A}
	for i = 1, #pointTable
		local point_i = (#pointTable%2) and i or i+1
		
		-- spawn the points on the source player
		local bpoint = P_SpawnMobj(p.mo.x, p.mo.y, p.mo.z, MT_BP_BATTLEPOINT)
		bpoint.state = pointStates[pointTable[i]]
		bpoint.target = p.mo
		bpoint.color = (v and v.skincolor) and v.skincolor or p.skincolor
		
		-- set offset
		local offset = 48	-- for best results, keep this even
		bpoint.bp_xyoffset = (offset * (point_i/2)) * ((i%2) and -1 or 1)
		if not (#pointTable%2)
			-- offset it backwards a bit for even amounts
			bpoint.bp_xyoffset = $ - ((offset/2) * ((i%2) and -1 or 1)) 
		end
	end
end

battleplus.GetAbsoluteFloorZ = function(mo)
	local abs_floorz = mo.floorz
	-- account for FOFs
	if (mo.subsector and mo.subsector.valid)
		for rover in mo.subsector.sector.ffloors()
			if (mo.z < rover.topheight) then continue end	-- random item is below the fof, skip it

			if (rover.topheight > abs_floorz)
				abs_floorz = rover.topheight
			end
		end
	end

	return abs_floorz
end

battleplus.MapToScreen = function(p, cam, mx, my, mz)
    -- Get camera angle
    local camangle = cam.angle	--R_PointToAngle(p.mo.x+FixedMul(cos(p.mo.angle), 64<<FRACBITS), p.mo.y+FixedMul(sin(p.mo.angle), 64<<FRACBITS))
    local camheight = cam.z
    local camaiming = FixedMul(160<<FRACBITS, tan(cam.aiming+ANGLE_180, true))

    local x = camangle -R_PointToAngle2(cam.x, cam.y, mx, my)
	--print((cam.angle/ANG1).." - "..(R_PointToAngle(mx, my)/ANG1).." = "..(x/ANG1))
    local distfact = FixedMul(FRACUNIT, cos(x)) or 1	-- be careful, this could equal 0.
	local behind

	if x > ANG1*75 or x < ANG1*285
		if x > ANG1*130 or x < ANG1*230 then
			behind = true
			x = -x	-- shift x
			--print("behind")
		elseif x > ANG1*90 or x < ANG1*270	-- fixes a weird bug where x is flipped when it shouldn't be.
			x = -x	-- earthbruh
		end
	end

	x = FixedMul(tan(x+ANGLE_180, true), 160<<FRACBITS)+160<<FRACBITS
    local y = camheight-mz
    y = FixedDiv(y, max(1, FixedMul(distfact, max(1, R_PointToDist2(cam.x, cam.y, mx, my)))))
    y = (y*160)+100<<FRACBITS
    y = y+camaiming

	local fovval = (CV_FindVar("fov").value/FRACUNIT)
	if splitscreen == 1
		fovval = ($*1333)/1000	-- fake the fovtan shit.
	end
	local fov = fovval - 90

	if fov
		-- how much do we divide by exactly...?
		local xdist = (x/FRACUNIT)-160

		-- start at 1: increase by 0.02 for each fov unit
		-- this is a magic number found after some maths. Don't question it.
		local inc = FRACUNIT*20 / 1000
		local start = FRACUNIT
		local divider = start + fov*inc
		divider = max(1, $)

		--print("^FOV: "..fov..", xdist: "..xdist)
		xdist = FixedDiv($*FRACUNIT, divider) / FRACUNIT
		--print("Resulting xdist: "..xdist)
		x = (160+xdist)*FRACUNIT
	end
	
    return x, y, behind, FixedDiv(160*FRACUNIT, max(1, FixedMul(distfact, R_PointToDist2(cam.x, cam.y, mx, my))))
end