local mineblast = battleplus.mineblast

-- make random items give us MORE items
addHook("TouchSpecial", function(item, mo)
	if not battleplus.IsGamemode(BP_GM_MINEBLAST) then return end	-- not mine blast
	if not (mo and mo.player) then return end
	if not (mo.player.kartstuff[k_bumper]) then return end

	if (mo.player.kartstuff[k_itemamount] < mineblast.cv_minelimit.value)
		P_KillMobj(item, mo, mo)
		mo.player.bp.mb_amount = $ and min(mineblast.cv_minelimit.value, $+1) or 1
		return true
	end
end, MT_RANDOMITEM)
-- remove floating items entirely in mine blast
addHook("MobjThinker", function(mcitem)
	-- just remove these in this gamemode
	if not (battleplus.IsGamemode(BP_GM_MINEBLAST)) then return end	-- not mine blast
	
	if (mcitem and mcitem.valid)
		P_RemoveMobj(mcitem)
		return
	end
end, MT_FLOATINGITEM)

/*
-- fix floating items as well for this new system
addHook("TouchSpecial", function(mcitem, mo)
	if not battleplus.IsGamemode(BP_GM_MINEBLAST) then return end	-- not mine blast
	if not (mo and mo.player) then return end

	local p = mo.player
	if (not P_CanPickupItem(p, 3) or (p.kartstuff[k_itemamount] and p.kartstuff[k_itemtype] ~= mcitem.threshold)) then return end
	if (p.kartstuff[k_bumper] <= 0) then return end

	p.bp.mb_amount = $ and min(mineblast.cv_minelimit.value, $+mcitem.movecount) or mcitem.movecount
end, MT_FLOATINGITEM)*/

-- get player mine count for hud and blow up forward thrown mines
mineblast.Func_PlayerMineThinker = function(p)
	p.bp.mb_mines = $ or {}	-- init
	for i = 1, 2
		p.bp.mb_mines[i] = $ or {}	-- init

		for k, v in ipairs(p.bp.mb_mines[i])
			-- check if they're still valid
			if not (v.valid)
				table.remove(p.bp.mb_mines[i], k)
				continue
			end

			-- only for forward thrown mines
			if (i ~= 1) then continue end
			v.mb_throwntics = $ or 1	-- make SURE this exists
			
			-- explode forward thrown mines on the ground
			if (P_IsObjectOnGround(v))
				--P_KillMobj(v)
				BP_ExplodeMine(v)
				table.remove(p.bp.mb_mines[i], k)	-- they wont be valid afterwards, anyways
				continue
			end
			
			v.mb_throwntics = $+1
		end
	end
end

-- hook for bomb damage I guess
local function bombdeath(mo)
	if not battleplus.IsGamemode(BP_GM_MINEBLAST) return end
	BP_ExplodeMine(mo)
	return true
end

addHook("MobjDeath", bombdeath, MT_SSMINE)
addHook("MobjDeath", bombdeath, MT_SSMINE_SHIELD)

addHook("MobjDamage", bombdeath, MT_SSMINE)
addHook("MobjDamage", bombdeath, MT_SSMINE_SHIELD)

-- proximity detonation hack for mb
function A_GrenadeRing(actor, var1, var2)
	if not battleplus.IsGamemode(BP_GM_MINEBLAST)
		super(actor, var1, var2)
		return 
	end

	local explodedist = FixedMul(actor.info.painchance, mapobjectscale)

	for p in players.iterate do
		local mo = p.mo
		if not mo continue end	-- bruh!

		if mo == actor.target and actor.treshold continue end	-- this is my friend.

		if (gametype == GT_CTF or gametype == GT_TEAMMATCH)
		and not CV_FindVar("friendlyfire").value and actor.target and actor.target.player
		and actor.target.player.ctfteam == mo.player.ctfteam
			continue
		end	-- this condition sucks

		-- see if it went over / under
		if (actor.z - explodedist > mo.z + mo.height)
			continue -- overhead
		end
		if (actor.z + actor.height + explodedist < mo.z)
			continue -- underneath
		end

		if (P_AproxDistance(P_AproxDistance(mo.x - actor.x, mo.y - actor.y), mo.z - actor.z) > explodedist)
			continue
		end	-- Too far away

		-- Explode!
		BP_ExplodeMine(actor)
		break
	end
end

-- player sidelining handler
mineblast.Func_PlayerSideline = function(p)
	if not (p.kartstuff[k_bumper])
		-- half our points on sideline
		if not (p.bp.mb_sidelined)
			p.bp.mb_sidelined = true
			if not (p.exiting) then p.bp.mb_score = $/2 end
			p.marescore = p.bp.mb_score
		end
		-- respawn them!
		if not (p.kartstuff[k_comebacktimer])
			for i = 1, mineblast.cv_bumpers.value
				battleplus.SpawnBumper(p)
				p.kartstuff[k_bumper] = $+1
			end
			p.bp.mb_sidelined = nil
		else
			p.kartstuff[k_comebacktimer] = min(5*TICRATE, $)	-- 5 second respawn
			if (p.exiting) then p.kartstuff[k_comebacktimer] = max(69, $) end	-- dont respawn after round end
		end
	end
end

-- player mine getting & throwing
mineblast.Func_PlayerMines = function(p)
	if not (p and p.mo) or p.spectator then return end
	local mo = p.mo
	-- speed up roulette
	if (p.kartstuff[k_itemroulette] and p.kartstuff[k_itemroulette] ~= 104) then p.kartstuff[k_itemroulette] = 104 end

	-- we have an item, lets hack it
	if (p.bp.mb_amount)	-- we have mines
	and (p.kartstuff[k_bumper])	-- make sure we're still in the game
		p.bp.mb_amount = $ and min(mineblast.cv_minelimit.value, $) or 1
		-- kartstuff
		p.kartstuff[k_itemtype] = KITEM_MINE	-- make SURE we have mines
		p.kartstuff[k_itemamount] = p.bp.mb_amount	-- mine limit
		-- throw vars
		p.pflags = $|PF_ATTACKDOWN	-- we cant use them, though
		p.bp.mb_throwinfo = $ or {}
		
		-- validity checks to make sure we can throw!
		if not ((mo and mo.valid) and mo.health)	-- not alive
		or (p.spectator)	-- make sure this is removed
		or (p.exiting)	-- not when we've finished
		or (p.kartstuff[k_spinouttimer] or p.kartstuff[k_squishedtimer] or p.kartstuff[k_respawn])	-- our vital organs!
			p.bp.mb_throwinfo = {}
			mo.colorized = false
			return 
		end

		-- let's start throwing it!
		local info = p.bp.mb_throwinfo

		if (p.cmd.buttons & BT_ATTACK)	-- holding item button, charge it up!
			info.strength = $ or 0	-- set strength
			if (p.kartstuff[k_throwdir] == -1)
				-- holding brake? stop strength growth
				info.stop = ($ ~= nil) and min(3*TICRATE/2, $+1) or 0
				if (info.stop == (3*TICRATE/2))
					-- start decreasing strength
					info.strength = max(0, $-1)
				end
			else
				if (info.stop)
					info.stop = min(TICRATE, $-1)
				else
					-- who knows, gamespeed changing in battle may become a reality someday
					local maxstrength = (gamespeed == 0 and 58 or (gamespeed == 2 and 86 or 72))
					info.strength = min(maxstrength, $+2)
				end
			end
				
			-- colorize player based on charge
			local maxstrength = (gamespeed == 0 and 58 or (gamespeed == 2 and 86 or 72))
			local strengthadd = info.strength*4/maxstrength
				
			mo.colorized = not (leveltime%(6-strengthadd)) and true or false
			info.held = true
		else	-- let's try release a mine
			if (info.held)	-- we can release it
				-- first set some variables
				local dir = (p.kartstuff[k_throwdir] == 1 and 2 or (p.kartstuff[k_throwdir] == -1 and -1 or 1))
				local throwheight = (20+(dir*10))*(mapobjectscale) + mo.momz

				-- now we release it
				if (dir > 0) -- throw forward
					if (#p.bp.mb_mines[1] < mineblast.cv_throwlimit.value)
						local mine = P_SpawnMobj(mo.x, mo.y, mo.z+(P_GetPlayerHeight(p)/2), MT_SSMINE_SHIELD)
						mine.momx = mo.momx + FixedMul(((10+info.strength)*cos(mo.angle)), (mapobjectscale))
						mine.momy = mo.momy + FixedMul(((10+info.strength)*sin(mo.angle)), (mapobjectscale))
						mine.momz = P_MobjFlip(mo)*(info.strength < 32 and throwheight/2 or throwheight)
						mine.target = mo
						mine.threshold = 10
						mine.mb_throwntics = 1
						table.insert(p.bp.mb_mines[1], mine)

						-- spawn gfx when you throw it
						local throwmo = P_SpawnMobj(mo.x, mo.y, mo.z + (P_GetPlayerHeight(p)/2), MT_FIREDITEM)
						throwmo.target = p.mo
						throwmo.movecount = 0
						S_StartSound(mine, sfx_tossed)

						K_PlayAttackTaunt(mo)
						p.bp.mb_amount = $-1
						p.kartstuff[k_itemamount] = $-1
					else
						p.bp.mb_releaseflash = 3
					end
				else
					if (#p.bp.mb_mines[2] < mineblast.cv_deploylimit.value) -- make sure they arent trying to nuke the damn server
						-- set some variables...
						local dropradius = FixedHypot(mo.radius, mo.radius) + FixedHypot(mobjinfo[MT_SSMINE_SHIELD].radius, mobjinfo[MT_SSMINE_SHIELD].radius)
						local newx = mo.x + P_ReturnThrustX(mo, mo.angle+ANGLE_180, dropradius)
						local newy = mo.y + P_ReturnThrustY(mo, mo.angle+ANGLE_180, dropradius)

						local mine = P_SpawnMobj(newx, newy, mo.z, MT_SSMINE_SHIELD)
						mine.threshold = 10
						mine.extravalue1 = 49
						mine.target = mo
						table.insert(p.bp.mb_mines[2], mine)

						K_PlayAttackTaunt(mo)
						p.bp.mb_amount = $-1
						p.kartstuff[k_itemamount] = $-1
					else
						p.bp.mb_releaseflash = 3
					end
				end
				
				p.bp.mb_throwinfo = {}
				mo.colorized = false
			end
		end
	else
		p.bp.mb_throwinfo = {}
		p.bp.mb_amount = 0
		mo.colorized = false
		
		-- failsafe???
		if (p.kartstuff[k_itemtype])
		and (p.kartstuff[k_itemamount])
			p.kartstuff[k_itemtype] = KITEM_MINE	-- make SURE we have mines
			p.bp.mb_amount = p.kartstuff[k_itemamount]
		end
	end

	-- reduce flash
	p.bp.mb_releaseflash = $ and $-1 or 0
end

-- build the function
mineblast.Func = function()
	for p in players.iterate
		mineblast.Func_PlayerMineThinker(p)	-- mine count for HUD & blowing up mines
		mineblast.Func_PlayerSideline(p)	-- player sidelining
		mineblast.Func_PlayerMines(p)	-- everything else mine related
		
		-- the player's score (we're gonna do some HACKY shit here)
		p.bp.mb_score = $ or 0
		if not (p.exiting) then p.marescore = p.bp.mb_score end	-- make sure this is set
	end
end

-- should damage function
mineblast.ShouldDamage = function(p, inflictor, source)
    if (source and source.player)
        if (p == source.player)
            K_DoInstashield(p)
            return false
        end
    end
end

-- damage functions, make sure we still get points
mineblast.PlayerDamage = function(p, inflictor, source)
	local mo = p.mo
	
	-- source checks
	if not (source and source.valid)	-- no source
	or not (source.player)	-- it's not a player?
	or (source == mo)	-- same player
		return
	end
	
	-- this'll make it easier (no combos!)
	if (p.powers[pw_flashing] or p.kartstuff[k_squishedtimer] or p.kartstuff[k_spinouttimer])
	or (p.exiting)
		return
	end
	
	-- mine blast damage points
	local minpoints = (p.kartstuff[k_bumper] == 1) and 2 or 1
	if (inflictor.mb_throwntics)	-- if it has this, it's DIRECT. give more points!
		-- get more points for snipes
		local points = max(minpoints, min(3, (inflictor.mb_throwntics * 2)/21))
		
		source.player.bp.mb_score = $ + points
		source.player.marescore = source.player.bp.mb_score
		battleplus.SpawnBattlePoints(source.player, p, points)
		if (points == 3) then S_StartSound(mo, sfx_bsnipe) end
	else	-- just give 1 point for the damage inflicted
		source.player.bp.mb_score = $ + minpoints
		source.player.marescore = source.player.bp.mb_score
		battleplus.SpawnBattlePoints(source.player, p, minpoints)
	end
end

-- make SURE mines explode on contact when thrown
mineblast.MineCollide = function(mine, mo)
	if not (battleplus.IsGamemode(BP_GM_MINEBLAST)) then return end
	
	if (mine.type != MT_SSMINE_SHIELD) then return end
	if (mine.threshold) then return end
	if not (mine.mb_throwntics) then return end	-- this'll be a good way to identify this
	
	-- make sure it's a player
	if not ((mo and mo.valid) and mo.health)
	or not (mo.player)	-- I'll just make it work with players for now
		return
	end
	local p = mo.player
	
	-- kill the mine on collision
	if (abs(mine.z - mo.z) <= P_GetPlayerHeight(p))
		P_KillMobj(mine)
	end
end
addHook("MobjCollide", mineblast.MineCollide, MT_SSMINE_SHIELD)
addHook("MobjMoveCollide", mineblast.MineCollide, MT_SSMINE_SHIELD)

-- hud function
mineblast.HUDFunc_DeployedMines = function(v, p)
	-- conditions when we wouldnt draw it normally
	if (p.exiting or p.spectator) then return end
	if (p.kartstuff[k_comebacktimer] and not p.kartstuff[k_bumper]) then return end

	local mines_deployed = p.bp.mb_mines and #p.bp.mb_mines[2] or 0
	local dp = battleplus.IsDisplayPlayer(p)
	local redfont = ((leveltime%10 >= 5) and mines_deployed == mineblast.cv_deploylimit.value)

	if (splitscreen > 1)
		local hx = (dp == 1 or dp == 3) and 118 or 1
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_SNAPTORIGHT|V_HUDTRANS, dp)
		local map = (redfont) and V_REDMAP or 0

		v.draw(hx, 69, v.cachePatch("K_MNEICO"), flags)
		v.drawString(hx+26, 75, mines_deployed, flags|map, "right")
	else
		local hy = (splitscreen) and 57 or 152
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_HUDTRANS, dp)

		-- get numbers
		local s = tostring(mines_deployed)
		local nums = {}
		if (mines_deployed >= 10)	-- more than 10, use 2 numbers
			nums = {tonumber(s:sub(1, 1)), tonumber(s:sub(2, 2))}
		else	-- only 1 number
			nums = {0, mines_deployed}
		end
		-- get font
		local font = (redfont) and "MBFONT" or "OPPRNK"

		-- draw stuff
		v.draw(10, hy, v.cachePatch("K_STMNED"), flags)
		v.draw(35, hy+5, v.cachePatch(font.."0"..nums[1]), flags)
		v.draw(41, hy+5, v.cachePatch(font.."0"..nums[2]), flags)
	end
end

mineblast.HUDFunc_Mines = function(v, p)
	-- conditions when we wouldnt draw it normally
	if (p.exiting or p.spectator) then return end
	if (p.kartstuff[k_comebacktimer] and not p.kartstuff[k_bumper]) then return end
	if not (p.kartstuff[k_itemtype] and p.kartstuff[k_itemamount]) then return end

	local holding = (p.bp.mb_throwinfo and p.bp.mb_throwinfo.held)
	local dp = battleplus.IsDisplayPlayer(p)

	local skin = (p.bp.mb_releaseflash or (p.kartstuff[k_itemblink] and leveltime%2)) and -6 or -1
	local color = p.bp.mb_releaseflash and SKINCOLOR_RED or ((p.kartstuff[k_itemblink] and leveltime%2) and SKINCOLOR_WHITE or SKINCOLOR_NONE)

	-- who knows, gamespeed changing in battle may become a reality someday
	local maxstrength = (gamespeed == 0 and 58 or (gamespeed == 2 and 86 or 72))
	local strength = p.bp.mb_throwinfo and p.bp.mb_throwinfo.strength or 0

	if (splitscreen > 1)
		-- draw the backdrop + mine
		local right = (dp == 1 or dp == 3)
		local hx = (right) and 121 or -9
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOTOP|V_SNAPTOLEFT|V_SNAPTORIGHT|V_HUDTRANS, dp)

		v.draw(hx, -8, v.cachePatch("K_ISBG"), flags)
		-- we want this drawn before K_ITMUL
		if (p.kartstuff[k_itemamount] > 1) then v.draw(hx+((right) and 48 or 0), -8, v.cachePatch("K_ISMUL"), ((right) and V_FLIP or 0)|flags) end
		-- the mine
		if not ((holding) and leveltime%2) then v.draw(hx, -8, v.cachePatch("K_ISMINE"), flags, v.getColormap(skin, color)) end
		-- amount of items we have
		if (p.kartstuff[k_itemamount] > 1)
			local hx2 = (right) and 123 or 15
			v.drawString(hx2, 23, "x"..p.kartstuff[k_itemamount], V_ALLOWLOWERCASE|flags, "left")
		end

		-- add a strength bar
		if (holding)
			v.draw(hx+17, 19, v.cachePatch("K_ISIMER"), flags)
			-- main bar
			v.drawFill(hx+18, 20, strength*12/maxstrength, 1, 120|flags)
			-- borders
			v.drawFill(hx+18, 20, 1, 1, 12|flags)
			if ((strength*12/maxstrength) > 0) then v.drawFill(hx+17+strength*12/maxstrength, 20, 1, 1, 12|flags) end
		end
	else
		-- draw the backdrop + mine
		local hy = (splitscreen) and 3 or 5
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOTOP|V_SNAPTOLEFT|V_HUDTRANS, dp)

		v.draw(5, hy, v.cachePatch("K_ITBG"), flags)
		-- we want this drawn before K_ITMUL
		if (p.kartstuff[k_itemamount] > 1) then v.draw(5, hy, v.cachePatch("K_ITMUL"), flags) end
		-- the mine
		if not ((holding) and leveltime%2) then v.draw(5, hy, v.cachePatch("K_ITMINE"), flags, v.getColormap(skin, color)) end
		-- amount of items we have
		if (p.kartstuff[k_itemamount] > 1)
			v.draw(33, hy+41, v.cachePatch("K_ITX"), flags)
			battleplus.DrawKartText(v, 43, hy+36, p.kartstuff[k_itemamount], flags, "left")
		end

		-- add a strength bar
		if (holding)
			v.draw(16, hy+35, v.cachePatch("K_ITIMER"), flags)
			-- main bar
			v.drawFill(17, hy+36, strength*26/maxstrength, 1, 120|flags)
			v.drawFill(17, hy+37, strength*26/maxstrength, 1, 8|flags)
			-- borders
			v.drawFill(17, hy+36, 1, 2, 12|flags)
			if ((strength*26/maxstrength) > 0) then v.drawFill(16+strength*26/maxstrength, hy+36, 1, 2, 12|flags) end
		end
	end
end

mineblast.HUDFunc = function(v, p, c)
	mineblast.HUDFunc_DeployedMines(v, p)
	mineblast.HUDFunc_Mines(v, p)
end

local mineblast_info = {
	name = "Mine Blast",
	teamname = "Team Mine Blast",
	objective = "Score the most points!",
	icon = "K_ITMINE",
	teamicons = {"K_TMMNB1", "K_TMMNB2"},
	func = mineblast.Func,
	playerspin = mineblast.PlayerDamage,
	playerexplode = mineblast.PlayerDamage,
	shouldspin = mineblast.ShouldDamage,
	shouldexplode = mineblast.ShouldDamage,
	hudfunc = mineblast.HUDFunc,
	disabledhuds = {"item", "wanted"},
	rules = BP_GMR_TEAMS|BP_GMR_TIMELIMIT|BP_GMR_BUMPERS|BP_GMR_REMOVEPOINTS,
	timelimit = mineblast.cv_timelimit,
	bumpers = mineblast.cv_bumpers,
}
battleplus.AddGamemode(BP_GM_MINEBLAST, mineblast_info)