local ringrunners = battleplus.ringrunners

-- spawn rings
ringrunners.Func_ReturnSpawnpoints = function()
	if not (server.bp_server.rr_cacheditems)	-- this will happen every 15 seconds, lets reduce processing
		server.bp_server.rr_cacheditems = {}
		local t = server.bp_server.rr_cacheditems
			
		for m in thinkers.iterate("mobj")
			if (m.type ~= MT_RANDOMITEM) then continue end	-- not a random item
			-- dont get items too high up from the ground
			local abs_floorz = battleplus.GetAbsoluteFloorZ(m)
			if (abs(m.z - abs_floorz) > 48*(mapobjectscale)) then continue end	-- too high, we want itemboxes near the floor
			
			t[#t+1] = {m.x>>FRACBITS, m.y>>FRACBITS, m.z>>FRACBITS}
		end
	end
	return server.bp_server.rr_cacheditems	-- return spawns
end
ringrunners.Func_GetSpawnpoints = function()
	local spawns = ringrunners.Func_ReturnSpawnpoints()
	//print("Total Spawnpoints: "..#spawns)
	-- get possible spawns
	local possibleSpawns = {}
	for k, v in ipairs(spawns)
		table.insert(possibleSpawns, v)
	end
	
	-- now select the spawns
	local spawnpoints = {}
	for i = #possibleSpawns, 1, -1
		local spawn = P_RandomRange(1, i)
		table.insert(spawnpoints, possibleSpawns[spawn])
		table.remove(possibleSpawns, spawn)
		
		-- check how many spawns we have
		if (#spawnpoints == max(1, #spawns/4))
		or (#spawnpoints == 5)
			break	-- we have enough
		end
	end
	return spawnpoints
end
ringrunners.Func_SpawnRings = function()
	-- check if round is over
	for p in players.iterate
		if (p.exiting) then return end
	end
	
	local starttime = 6*TICRATE + (3*TICRATE/4)
	
	-- handle spawned spawners to show in the hud
	server.bp_server.rr_spawns = $ or {}
	for k, v in ipairs(server.bp_server.rr_spawns)
		if not (v and v.valid)
			table.remove(server.bp_server.rr_spawns, k)
		end
	end
	-- spawn spawners
	if (leveltime >= starttime)	-- only spawn rings after starttime
		local ringtime = (leveltime-starttime)
		if not (ringtime%(TICRATE*15))	-- 4 per minute seems fine
			local spawnpoints = ringrunners.Func_GetSpawnpoints()
			//print("Spawnpoints selected: "..#spawnpoints)
			for i = 1, #spawnpoints
				local spawnpoint = spawnpoints[i]
				local spawn = ringrunners.SpawnRingSpawnpoint(spawnpoint[1]<<FRACBITS, spawnpoint[2]<<FRACBITS, spawnpoint[3]<<FRACBITS, 3)
				table.insert(server.bp_server.rr_spawns, spawn)
			end
		end
	end
end

-- handle player rings
ringrunners.Func_PlayerRings = function(p)
	local mo = p.mo
	-- validity checks on vars used
	p.bp.rr_rings = $ or 0
	p.bp.rr_hudflash = $ or 0
	
	-- lose rings if you fall off like a cunt
	local respawn_tics = 35
	if (JUICEBOX and (JUICEBOX.value or JUICEBOX_forcefastrespawn.value))
		-- Juicebox fast respawn support
		respawn_tics = 20
	end
		
	if (p.kartstuff[k_respawn] == respawn_tics)
	and (p.bp.joincheck)	-- join hack
		local rings = ringrunners.CalculateRingsOnPlayer(p, 15)
		ringrunners.GiveRings(p, -rings)
		if (mo and mo.valid)	-- justified paranoia
			ringrunners.RingBurst(p.mo, rings, {-6, 6}, {-6, 6}, {14, 18})
			S_StartSound(mo, sfx_8blose)
		end
	end
	
	-- set score as your rings
	if not (p.exiting) then p.marescore = max(0, p.bp.rr_rings) end
	
	-- make sure the hudflash is always at 0
	if (p.bp.rr_hudflash ~= 0)
		p.bp.rr_hudflash = $+(p.bp.rr_hudflash < 0 and 1 or -1)
	end
end

-- update player wanted
ringrunners.Func_PlayerWanted = function(p)
	if (battleplus.GetPlayerCount() > 2)
		if (p.bp.rr_rings < 30)
			-- mercy
			p.kartstuff[k_wanted] = 0
		end
		-- the player with the most bumpers become wanted, lets exploit that
		local highest = true
		for i in players.iterate
			if not (i and i.bp.rr_rings) or i.spectator then continue end

			if (p.bp.rr_rings < (i.bp.rr_rings-29))
				highest = false
				break
			end
		end
		p.kartstuff[k_bumper] = (highest) and 3 or 2
	else	-- wanted isnt needed in 1v1s
		p.kartstuff[k_bumper] = 2
		p.kartstuff[k_wanted] = 0
	end
end

-- handle player debt
ringrunners.Func_PlayerRingDebt = function(p)
	if not ringrunners.IsPlayerInDebt(p) then return end
	local mo = p.mo
	
	if (mo and mo.valid)
		if not mo.rr_debt or not mo.rr_debt.valid
			mo.rr_debt = P_SpawnMobj(mo.x, mo.y, mo.z, MT_BP_RR_DEBT)
			mo.rr_debt.target = mo
			mo.rr_debt.flags2 = $|MF2_DONTDRAW
		end
	end
end

-- build the gamemode function
ringrunners.Func = function()
	ringrunners.Func_SpawnRings()
	ringrunners.HandleRings()
	ringrunners.HandleChaseRings()
	for p in players.iterate
		ringrunners.Func_PlayerRings(p)
		ringrunners.Func_PlayerWanted(p)
		ringrunners.Func_PlayerRingDebt(p)
	end
end

-- damage functions
ringrunners.PlayerSpin = function(p, inflictor, source)
	if (p.exiting) then return end	-- dont lose rings during results
	if (p.powers[pw_flashing]) then return end	-- no combos
	-- take away rings
	local rings = ringrunners.CalculateRingsOnPlayer(p, 15)
	ringrunners.GiveRings(p, -rings)
	if (p.mo and p.mo.valid)	-- justified paranoia
		S_StartSound(p.mo, sfx_8blose)
	end
	-- were they stolen? give them to the thief
	if ((source and source.player) and (source.player.kartstuff[k_sneakertimer] or source.player.kartstuff[k_pogospring])) -- got rammed with a sneaker/stomped with a pogospring?
		ringrunners.GiveRings(source.player, rings)
	else	-- otherwise burst them out
		ringrunners.RingBurst(p.mo, rings, {-6, 6}, {-6, 6}, {14, 18})
	end
end
ringrunners.PlayerSquish = function(p, inflictor, source)
	if (p.exiting) then return end	-- dont lose rings during results
	if (p.powers[pw_flashing]) then return end	-- no combos
	-- take away rings
	local rings = ringrunners.CalculateRingsOnPlayer(p, 15)
	ringrunners.GiveRings(p, -rings)
	ringrunners.RingBurst(p.mo, rings, {-6, 6}, {-6, 6}, {14, 18})
	if (p.mo and p.mo.valid)	-- justified paranoia
		S_StartSound(p.mo, sfx_8blose)
	end
end
ringrunners.PlayerExplode = function(p, inflictor, source)
	if (p.exiting) then return end	-- dont lose rings during results
	if (p.powers[pw_flashing]) then return end	-- no combos
	-- take away rings
	local rings = ringrunners.CalculateRingsOnPlayer(p, 15)
	ringrunners.GiveRings(p, -rings)
	ringrunners.RingBurst(p.mo, rings, {-5, 5}, {-5, 5}, {20, 24})
	if (p.mo and p.mo.valid)	-- justified paranoia
		S_StartSound(p.mo, sfx_8blose)
	end
end

-- player bumping
ringrunners.PlayerBump = function(mo, mo2)
	if not battleplus.IsGamemode(BP_GM_RINGRUNNERS) then return end	-- not ring runners
	
	if not ((mo2 and mo2.valid) and mo2.player) then return end
	local p, p2 = mo.player, mo2.player -- lazy
	
	-- I'm too lazy to do all these checks so lets just make it easier
	local kartstuff_checks = {k_sneakertimer, k_pogospring, k_justbumped, k_spinouttimer, k_squishedtimer, k_invincibilitytimer}
	for i = 1, #kartstuff_checks
		local c = kartstuff_checks[i]
		
		if (p.kartstuff[c] or p2.kartstuff[c]) then return end
	end
	-- grow gets a special check since it's 2 functions in 1 timer
	if ((p.kartstuff[k_growshrinktimer] > 0) or (p2.kartstuff[k_growshrinktimer] > 0)) then return end
	
	-- dont affect teammates
	if (server.bp_server.teamgamemode)
	and (p.bp.team == p2.bp.team)
		return
	end
	
	if (abs(mo.z - mo2.z) <= P_GetPlayerHeight(p))
	and not (p.powers[pw_flashing] or p2.powers[pw_flashing])
		local extra_drop = max(0, p2.kartweight - p.kartweight)
		local rings = ringrunners.CalculateRingsOnPlayer(p, 2+(extra_drop))
		-- burst rings from each player
		ringrunners.GiveRings(p, -rings)
		ringrunners.RingBurst(mo, rings, {-6, 6}, {-6, 6}, {14, 18})
		S_StartSound(mo, sfx_rrring)
	end
end
addHook("MobjCollide", ringrunners.PlayerBump, MT_PLAYER)
addHook("MobjMoveCollide", ringrunners.PlayerBump, MT_PLAYER)

-- hud function
ringrunners.HudFunc_MinimapRings = function(v, p)
	if (battleplus.IsDisplayPlayer(p) ~= 0) then return end	-- reduce rendering as much as possible
	if not (server.bp_server.rr_spawns) then return end
	if (p.exiting) then return end
	
	for k, spawn in ipairs(server.bp_server.rr_spawns)
		if not (spawn and spawn.valid) then continue end	-- case 1: spawn fucking died
		
		-- case 2: spawning in
		local spawngrace = spawn.bp_spawngrace or 0
		if (spawngrace < (ringrunners.spawnpoint_spawngracetime))
			if not (leveltime%10 >= 5) then continue end
		end
		-- case 3: despawning
		if (spawn.tics <= 3*TICRATE)
			if not (leveltime%2) then continue end
		end
		
		v.drawOnMinimap(spawn.x, spawn.y, FRACUNIT*3/4, v.cachePatch("K_RRMMAP"), v.getColormap(-1, SKINCOLOR_GOLD))
	end
end

ringrunners.HudFunc_PlayerRings = function(v, p)
	-- conditions when we wouldnt draw it normally
	if (p.exiting or p.spectator) then return end
	
	local rings = p.bp.rr_rings or 0
	local flash = p.bp.rr_hudflash or 0
	
	local dp = battleplus.IsDisplayPlayer(p)
	local critical = (((leveltime%10 >= 5) and (rings < 0)) or flash < 0)	-- critical!
	local flashcheck = (flash > 0)
	
	local patch = (critical) and "K_STDNGN" or ((flashcheck) and "K_STFNGN" or "K_STRNGN")
	
	if (splitscreen) > 1
		local hx = (dp == 1 or dp == 3) and 120 or 3
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_SNAPTORIGHT|V_HUDTRANS, dp)
		local patch = (critical) and "K_DNGICO" or ((flashcheck) and "K_FNGICO" or "K_RNGICO")
		
		v.draw(hx, 86, v.cachePatch(patch), flags)
		v.drawString(hx+37, 88, rings, flags|(critical and V_REDMAP or 0), "right")
	else
		local hy = (splitscreen) and 76 or 171
		local flags = battleplus.ReturnSplitFlags(V_SNAPTOBOTTOM|V_SNAPTOLEFT|V_HUDTRANS, dp)
		
		v.draw(9, hy, v.cachePatch(patch), flags, v.getColormap(-1, p.skincolor))
		battleplus.DrawKartText(v, 92, hy+3, rings, flags, "right", critical)
	end
end

ringrunners.HudFunc_RankingsRings = function(v, p)
	if (splitscreen) then return end
	if (battleplus.IsDisplayPlayer(p) != 0) then return end	-- reduce rendering as much as possible
	if (battleplus.GetPlayerCount() <= 1) then return end	-- not enough players
	if (p.exiting) then return end
	
	-- since battle always displays the top 5... time to cheat.......
	local rankplayers = {}
	for player in players.iterate
		if (player.spectator) then continue end
		table.insert(rankplayers, player)
	end
	table.sort(rankplayers, function(a, b) return a.kartstuff[k_position] < b.kartstuff[k_position] end)
	
	local hudy = 101 - (9 * min(5, #rankplayers))
	for i = 1, min(5, #rankplayers)	-- only 1st to 5th
		-- get the rankplayer
		local rankplayer = rankplayers[i]
		if not (rankplayer and rankplayer.valid) then continue end
		if (rankplayer.spectator) then continue end
		if not (rankplayer.mo) then continue end
		
		local hudx = 26
		if (rankplayer.mo.color)	-- this is a check in the source, I'll keep it here too for consistency???
			local rings = rankplayer.bp.rr_rings or 0
			if (rings > 0)
				-- get how many rings we'll be displaying
				local increment = 5
				local large, medium, small = 0, 0, 0
				while (rings >= (increment ^ 2))	-- first get large rings
					rings = $ - (increment ^ 2)
					large = $ + 1
				end
				while (rings >= increment)	-- next get medium rings
					rings = $ - increment
					medium = $ + 1
				end
				while (rings > 0)	-- rest will be small rings
					rings = $ - 1
					small = $ + 1
				end
				
				-- function to draw rings
				local function drawRing(patch_b)
					-- get patch
					local patch, patch_type = "K_RNGA", (patch_b) and "B" or "A"
					local hudx_add = 4
					if (large > 0)
						patch = "K_RNGC"
						hudx_add = 5
						large = $ - 1
					elseif (medium > 0)
						patch = "K_RNGB"
						medium = $ - 1
					else
						small = $ - 1
					end
					hudx_add = $ - ((patch_b) and 1 or 0)
					
					v.draw(hudx, hudy, v.cachePatch(patch..patch_type), V_HUDTRANS|V_SNAPTOLEFT, v.getColormap(-1, SKINCOLOR_GOLD))
					hudx = $ + hudx_add
				end
				
				-- display them
				if (large or medium or small)
					drawRing()
					
					-- draw subsequent rings
					while (large > 0)
						drawRing(true)
					end
					while (medium > 0)
						drawRing(true)
					end
					while (small > 0)
						drawRing(true)
					end
				end
			end
		end
		
		hudy = $ + 18
	end
end

ringrunners.HudFunc = function(v, p, c)
	ringrunners.HudFunc_MinimapRings(v, p)
	ringrunners.HudFunc_PlayerRings(v, p)
	ringrunners.HudFunc_RankingsRings(v, p)
end

local ringrunners_info = {
	name = "Ring Runners",
	teamname = "Team Ring Runners",
	objective = "Collect the most rings!",
	icon = "K_RNGRNR",
	func = ringrunners.Func,
	playerspin = ringrunners.PlayerSpin,
	playersquish = ringrunners.PlayerSquish,
	playerexplode = ringrunners.PlayerExplode,
	hudfunc = ringrunners.HudFunc,
	rules = BP_GMR_TEAMS|BP_GMR_TIMELIMIT|BP_GMR_REMOVEPOINTS,
	timelimit = ringrunners.cv_timelimit,
}
battleplus.AddGamemode(BP_GM_RINGRUNNERS, ringrunners_info)